//
//  EditEntryPolicyVC.m
//  pwSafe
//
//  Created by Jorge P. Vasquez on 04/10/12.
//
//

#import "EditEntryPolicyVC.h"
#import "AppTheme.h"
#import "SafeFile.h"
#import "Entry.h"
#import "PwGenParsViewController.h"
#import "PWPolicy.h"
#import "PWSprefs.h"
#import "PWScore.h"
#import "NSStringUtils.h"

@interface EditEntryPolicyVC ()
{
  SafeFile *_file;
  Entry *_entry;
  NSString *_pwPolicyName;
  PwGenParsViewController *_pwPolicyVC;
  NSArray *_names;
  void (^_handler)(EditEntryPolicyVC *editVC);
  BOOL _useCustomPolicy;
}
@end

@implementation EditEntryPolicyVC

- (id)initWithEntry:(Entry *)entry handler:(void (^)(EditEntryPolicyVC *editVC))handler
{
  self = [super initWithStyle:UITableViewStyleGrouped];
  if(!self)
    return nil;

  _file = entry.safeFile;
  _entry = entry;
  _pwPolicyName = entry.pwPolicyName;
  _handler = [handler copy];
  _useCustomPolicy = entry.hasCustomPasswordPolicy;
  
  self.navigationItem.title = @"Password Policy";
  
  return self;
}

- (void)dealloc
{
  if(self.isViewLoaded)
  {
    self.tableView.delegate = nil;
    self.tableView.dataSource = nil;
  }
}

- (void)safeWasInternallyModified
{
  _pwPolicyName = _entry.pwPolicyName;
  _names = nil;
  [self.tableView reloadData];
}

static const bool pwPolicyNumericOnly(const PWPolicy &pol)
{
  return (pol.flags & (PWPolicy::UseLowercase |
                       PWPolicy::UseUppercase |
                       PWPolicy::UseDigits |
                       PWPolicy::UseSymbols |
                       PWPolicy::UseHexDigits |
                       PWPolicy::UseEasyVision |
                       PWPolicy::MakePronounceable)) == PWPolicy::UseDigits;
}

- (BOOL)passwordPolicyNumericOnly
{
  if(_pwPolicyName && _pwPolicyName.length)
  {
    PWPolicy pol;
    [_file pwPolicy:pol named:_pwPolicyName];
    return pwPolicyNumericOnly(pol);
  }
  else if(_useCustomPolicy)
  {
    if(_pwPolicyVC && !_pwPolicyVC.canceled)
    {
      return _pwPolicyVC.passwordPolicyNumericOnly;
    }
    else
    {
      NSAssert(_entry.hasCustomPasswordPolicy, @"Must have one");
      PWPolicy pol;
      [_entry getPasswordPolicy:pol];
      return pwPolicyNumericOnly(pol);
    }
  }
  else
  {
    PWPolicy pol;
    [_file defaultPwPolicy:pol];
    return pwPolicyNumericOnly(pol);
  }
}

- (void)updateEntry:(Entry *)entry
{
  PWPolicy empty;
  empty.Empty();
  if(_pwPolicyName && _pwPolicyName.length)
  {
    [entry setPasswordPolicy:empty];
    entry.pwPolicyName = _pwPolicyName;
  }
  else if(_pwPolicyVC && !_pwPolicyVC.canceled)
  {
    entry.pwPolicyName = @"";
    [_pwPolicyVC updateEntry:entry];
  }
  else
  {
    entry.pwPolicyName = @"";
    [entry setPasswordPolicy:empty];
  }
}

- (NSString *)randomPassword
{
  try
  {
    std::auto_ptr<PWSprefs> prefs(new PWSprefs);
    @synchronized(_file) {
      prefs->Load(_file.pwsCore.GetDBPreferences());
    }
  
    if(_pwPolicyName && _pwPolicyName.length)
    {
      PWPolicy pol;
      [_file pwPolicy:pol named:_pwPolicyName];
      return [NSString stringWithwstring:pol.MakeRandomPassword(prefs.get())];
    }
    else if(_useCustomPolicy)
    {
      if(_pwPolicyVC && !_pwPolicyVC.canceled)
      {
        return _pwPolicyVC.randomPassword;
      }
      else
      {
        NSAssert(_entry.hasCustomPasswordPolicy, @"Must have one");
        PWPolicy pol;
        [_entry getPasswordPolicy:pol];
        return [NSString stringWithwstring:pol.MakeRandomPassword(prefs.get())];
      }
    }
    else
    {
      PWPolicy pol;
      [_file defaultPwPolicy:pol];
      return [NSString stringWithwstring:pol.MakeRandomPassword(prefs.get())];
    }
  }
  catch(const std::exception &exception)
  {
    A77Log(@"C++ exception generating password: %s", exception.what());
    return [NSString string];
  }
  catch(...)
  {
    A77Log(@"C++ unknown exception type generating password");
    return [NSString string];
  }
}

#pragma mark - View lifecycle

- (void)viewDidLoad
{
  [super viewDidLoad];
  [self.tableView themeViewDidLoadInIpadPos:TVP_RIGHT];
}

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
  // Return YES for supported orientations
  return (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad || interfaceOrientation == UIInterfaceOrientationPortrait);
}

- (void)didReceiveMemoryWarning
{
  [super didReceiveMemoryWarning];
  _names = nil;
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
  return 2;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
  if(section == 0)
    return 2; // Default + custom policies
  if(!_names)
    _names = _file.namedPwPoliciesNames;
  return _names.count;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
  static NSString *CellIdentifier = @"Cell";
  UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
  if(!cell)
  {
    cell = [[UITableViewCell alloc]initWithStyle:UITableViewCellStyleDefault reuseIdentifier:CellIdentifier];
    cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
  }
  NSUInteger section = indexPath.section;
  NSUInteger row = indexPath.row;
  BOOL isSelected;
  BOOL disclosure = NO;
  BOOL hasNamedPolicy = _pwPolicyName && _pwPolicyName.length;
  
  if(section == 0)
  {
    if(row == 0)
    {
      cell.textLabel.text = @"Default policy";
      isSelected = !_useCustomPolicy && !hasNamedPolicy;
    }
    else
    {
      cell.textLabel.text = @"Custom policy";
      isSelected = _useCustomPolicy && !hasNamedPolicy;
      disclosure = YES;
    }
  }
  else
  {
    if(!_names)
      _names = _file.namedPwPoliciesNames;
    NSString *name = _names[row];
    cell.textLabel.text = name;
    NSAssert(!hasNamedPolicy || [_names indexOfObject:_pwPolicyName] != NSNotFound, @"Must have a valid policy");
    isSelected = [_pwPolicyName isEqualToString:name];
  }

  cell.accessibilityTraits = (isSelected ? UIAccessibilityTraitSelected : UIAccessibilityTraitNone) | (disclosure ? UIAccessibilityTraitButton : UIAccessibilityTraitNone);
  cell.imageView.image = isSelected ? [UIImage imageNamed:@"checkmark"] : nil;
  cell.accessoryType = disclosure ? UITableViewCellAccessoryDisclosureIndicator : UITableViewCellAccessoryNone;
  cell.selectionStyle = disclosure ? UITableViewCellSelectionStyleBlue : UITableViewCellSelectionStyleNone;
  [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];

  return cell;
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
{
  if(section == 0)
    return nil;
  return @"Named policies";
}

#pragma mark - Table view delegate

- (NSIndexPath *)currentlySelectedRow
{
  if(!_useCustomPolicy && (!_pwPolicyName || !_pwPolicyName.length))
    return [NSIndexPath indexPathForRow:0 inSection:0];
  else if(_useCustomPolicy && (!_pwPolicyName || !_pwPolicyName.length))
    return [NSIndexPath indexPathForRow:1 inSection:0];
  else
  {
    NSAssert(_pwPolicyName && _pwPolicyName.length, @"Must use a named policy then");
    if(!_names)
      _names = _file.namedPwPoliciesNames;
    NSUInteger idx = [_names indexOfObject:_pwPolicyName];
    NSAssert(idx != NSNotFound, @"Must find");
    if(idx == NSNotFound)
      return nil;
    return [NSIndexPath indexPathForRow:idx inSection:1];
  }
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
  NSUInteger section = indexPath.section;
  NSUInteger row = indexPath.row;

  NSIndexPath *oldSelectedRow = [self currentlySelectedRow];
  
  if(section == 0)
  {
    if(row == 0) // Default
    {
      _useCustomPolicy = NO;
      _pwPolicyName = @"";
    }
    else if(row == 1) // Custom
    {
      if(!_pwPolicyVC)
      {
        _pwPolicyVC = [[PwGenParsViewController alloc]initWithEntry:_entry inFile:_file];
        [_pwPolicyVC setOnDoneTarget:self action:@selector(onPwPolicyEditComplete:)];
        _pwPolicyVC.navigationItem.title = @"Custom Policy";
      }
      [self.navigationController pushViewController:_pwPolicyVC animated:YES];
    }
  }
  else
  {
    _useCustomPolicy = NO;
    if(!_names)
      _names = _file.namedPwPoliciesNames;
    _pwPolicyName = _names[row];
  }
  
  NSArray *cells = [oldSelectedRow isEqual:indexPath] ? @[indexPath] : @[indexPath, oldSelectedRow];
  
  [self.tableView reloadRowsAtIndexPaths:cells withRowAnimation:UITableViewRowAnimationFade];
}

- (void)onPwPolicyEditComplete:(PwGenParsViewController *)vc
{
  [self.navigationController popViewControllerAnimated:YES];
  if(vc.canceled)
    return;
  _useCustomPolicy = YES;
  _pwPolicyName = @"";
  [self.tableView reloadData];
}

@end
