//
//  EditableSimpleTableViewCell.m
//  pwSafe
//
//  Created by Jorge Peixoto Vasquez on 07/11/11.
//  Copyright (c) 2011 app77. All rights reserved.
//

#import "EditableSimpleTableViewCell.h"
#import "AppTheme.h"

@implementation EditableSimpleTableViewCell
{
  UITextField *m_textField; 
  UILabel *m_detailLabel;
  TextValidator m_textValidator;
  TextChanged m_textChanged;
  NSString *m_originalText;
}


static UIColor * detailColor()
{
  static UIColor * kDetailColor = nil;
  if(!kDetailColor)
    kDetailColor = [UIColor colorWithWhite:.5 alpha:1];

  return kDetailColor;
}


static UIColor * selectedTextColor()
{
#ifdef THEME_USE_CUSTOM_TABLE_VIEW_CELL_SELECTION_BG
  return themeTableViewSelectedTextColor;
#endif
  return [UIColor whiteColor];
}


- (id)initWithStyle:(UITableViewCellStyle)style reuseIdentifier:(NSString *)reuseIdentifier
{
  self = [super initWithStyle:style reuseIdentifier:reuseIdentifier];
  if (!self) 
    return nil;

  BOOL supportsFontStyles = [UIFont respondsToSelector:@selector(preferredFontForTextStyle:)];
  
  // Initialization code
  m_textField = [[UITextField alloc]init];
  m_textField.autoresizingMask = UIViewAutoresizingFlexibleBottomMargin | UIViewAutoresizingFlexibleWidth;
  [self.contentView addSubview:m_textField];
  self.contentView.autoresizesSubviews = YES;
  m_textField.backgroundColor = [UIColor clearColor];
  m_textField.opaque = NO;
  m_textField.delegate = self;
  m_textField.clearButtonMode = UITextFieldViewModeWhileEditing;
  if(supportsFontStyles)
    m_textField.font = [UIFont preferredFontForTextStyle:UIFontTextStyleBody];
  else
    m_textField.font = [UIFont boldSystemFontOfSize:18];

  m_detailLabel = [[UILabel alloc]init];
  if(supportsFontStyles)
    m_detailLabel.font = [UIFont preferredFontForTextStyle:UIFontTextStyleCaption1];
  else
    m_detailLabel.font = [UIFont systemFontOfSize:14.0f];
  
  m_detailLabel.autoresizingMask = UIViewAutoresizingFlexibleTopMargin | UIViewAutoresizingFlexibleWidth;
  m_detailLabel.textColor = detailColor();
  m_detailLabel.backgroundColor = [UIColor clearColor];
  m_detailLabel.opaque = NO;
  [self.contentView addSubview:m_detailLabel];
  
  return self;
}

- (NSString *)detailText
{
  return m_detailLabel.text;
}

- (void)setDetailText:(NSString *)detailText
{
  m_detailLabel.text = detailText;
}

- (UITextField *)textField
{
  return m_textField;
}

#define TEXT_MARGIN (iOS6() ? 10.0f : 15.0f)
#define DETAIL_MARGIN 2.0f

- (void)layoutSubviews
{
  [super layoutSubviews];
  
  CGFloat imgOffset = 0;
  if(self.imageView.image)
  {
    CGRect imgFrame = self.imageView.frame;
    imgOffset = imgFrame.origin.x + imgFrame.size.width;
  }

  NSString *detailText = m_detailLabel.text;
  BOOL hasDetail = detailText && detailText.length;

  CGSize cellSize = self.contentView.frame.size;
  CGFloat textWidth = cellSize.width - imgOffset - 2 * TEXT_MARGIN;
  if(!iOS6() && self.isEditing)
    textWidth -= 22;
  
  CGRect detailFrame = CGRectZero;
  if(hasDetail)
  {
    detailFrame.size.height = [@"17fp" sizeWithFont:m_detailLabel.font].height;
    detailFrame.origin.x = imgOffset + TEXT_MARGIN;
    detailFrame.origin.y = cellSize.height - DETAIL_MARGIN - detailFrame.size.height;
    detailFrame.size.width = textWidth;
  }
  m_detailLabel.frame = detailFrame;
  
  CGRect frame;
  UIFont *font = m_textField.font;
  frame.size.height = font.lineHeight;
  frame.size.width = textWidth;
  frame.origin.y = hasDetail ? DETAIL_MARGIN : rintf((cellSize.height - frame.size.height) / 2);
  frame.origin.x = imgOffset + TEXT_MARGIN;
  
  m_textField.frame = frame;
  if(!self.isEditing)
    [m_textField resignFirstResponder];
  
  m_textField.userInteractionEnabled = self.isEditing;   
}

- (void)textFieldDidBeginEditing:(UITextField *)textField
{
  if(!iOS6())
    textField.textColor = kThemeTintColor;
  m_originalText = [textField.text copy];
}

- (BOOL)textAcceptable:(NSString *)text
{
  text = [text stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
  if(text.length == 0)
    return NO;
  if([text isEqualToString:m_originalText])
    return YES;
  if(!m_textValidator)
    return YES;
  return m_textValidator(text);
}

- (BOOL)textFieldShouldEndEditing:(UITextField *)textField
{
  if(![self textAcceptable:textField.text])
    textField.text = m_originalText;
  return YES;
}

- (BOOL)textFieldShouldReturn:(UITextField *)textField
{
  if(![self textAcceptable:textField.text])
    textField.text = m_originalText;
  [textField resignFirstResponder];
  return YES;
}

- (void)textFieldDidEndEditing:(UITextField *)textField
{
  if(!iOS6())
    textField.textColor = [UIColor blackColor];
 
  if(m_textChanged)
    m_textChanged([textField.text stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]], m_originalText);
  m_originalText = nil;
}

- (void)setTextValidator:(TextValidator)validator
{
  m_textValidator = [validator copy];
}

- (void)setTextChanged:(TextChanged)onTextChanged
{
  m_textChanged = [onTextChanged copy];
}

- (void)setHighlighted:(BOOL)highlighted animated:(BOOL)animated
{
  [super setHighlighted:highlighted animated:animated];
  if(highlighted)
  {
    m_textField.textColor = selectedTextColor();
    m_detailLabel.textColor = selectedTextColor();
  }
  else if(!self.isSelected)
  {
    if(!animated)
    {
      m_textField.textColor = [UIColor blackColor];
      m_detailLabel.textColor = detailColor();
    }
    else
    {
      dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, .2 * NSEC_PER_SEC);
      dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
        m_textField.textColor = [UIColor blackColor];
        m_detailLabel.textColor = detailColor();        
      });
    }
  }
}

- (void)setSelected:(BOOL)selected animated:(BOOL)animated
{
  [super setSelected:selected animated:animated];
  
  if(selected)
  {
    m_textField.textColor = selectedTextColor();    
    m_detailLabel.textColor = selectedTextColor();
  }
  else if(!self.isHighlighted)
  {
    if(!animated)
    {
      m_textField.textColor = [UIColor blackColor];
      m_detailLabel.textColor = detailColor();
    }
    else
    {
      dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, .2 * NSEC_PER_SEC);
      dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
        m_textField.textColor = [UIColor blackColor];
        m_detailLabel.textColor = detailColor();        
      });
    }
  }
}


@end
