//
//  Entry.m
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/7/11.
//  Copyright 2011 app77. All rights reserved.
//

#import "Entry.h"
#include "ItemData.h"
#include "PWSrand.h"
#include "PWPolicy.h"
#include "PWSprefs.h"
#include "PWScore.h"
#import "NSStringUtils.h"
#import "SafeFile.h"
#import "PWHistory.h"
#import "StringX.h"
#import "core.h"

#import <limits>

@implementation Entry
{
  BOOL m_deleteCItemData;
  SafeFile *__weak m_file;
  NSNumber *m_isProtected;
  BOOL m_neverSetPassword;
}

@synthesize title = m_title;
@synthesize username = m_username;
@synthesize password = m_password;
@synthesize group = m_group;
@synthesize url = m_url;
@synthesize notes = m_notes;
@synthesize email = m_email;
@synthesize pwPolicyName = m_pwPolicyName;
@synthesize cItemData = m_cItemData;
@synthesize safeFile = m_file;
@synthesize mTime = m_mTime;
@synthesize cTime = m_cTime;

- (id) initWithFile:(SafeFile *)file
{
  NSParameterAssert(file);
  
  if (!(self = [super init]))
    return nil;

  CItemData *item = new CItemData();
  m_deleteCItemData = YES;
  m_cItemData = item;
  m_group = @"";
  m_file = file;
  m_neverSetPassword = YES;
  [self generateNewUUID];
  [self setCTime:[NSDate date]];

#ifdef PWSPREFS_STATICS
  const PWSprefs *prefs = PWSprefs::GetInstance();
#else
  std::auto_ptr<PWSprefs> prefs (new PWSprefs);
  @synchronized(m_file) {
    NSAssert(&m_file.pwsCore, @"File must have a pwsCore");
    StringX dbPrefs = m_file.pwsCore.GetDBPreferences();
    prefs->Load(dbPrefs);
  }
#endif
  bool keep = prefs->GetPref(PWSprefs::SavePasswordHistory);
  uint maxEntries = prefs->GetPref(PWSprefs::NumPWHistoryDefault);
  if(maxEntries > std::numeric_limits<u_int8_t>::max())
    maxEntries = std::numeric_limits<u_int8_t>::max();
  auto header = ::MakePWHistoryHeader(keep, maxEntries, 0);
  item->SetPWHistory(header);

  return self;
}


- (id) initWithItemData:(const CItemData &)item file:(SafeFile *)file
{
  if (!(self = [super init]))
    return nil;
  
  m_file = file;
  
  [self setCItemData:&item isConst:YES];
  m_neverSetPassword = NO;
  
  return self;
}

- (void) generateNewUUID
{
  uuid_t newUUID;
  PWSrand::GetInstance()->GetRandomData(newUUID, sizeof(uuid_t));
  self.mutableItemData->SetUUID(newUUID);
}

- (void) parsePwHistoryFromItem:(const CItemData &)item keep:(bool &)keep max:(uint &)maxEntries entries:(PWHistList &)entries
{
  auto pwHistoryStr = item.GetPWHistory();
  if (pwHistoryStr.empty())
  {
    // If GetPWHistory() is empty, use preference values!
#ifdef PWSPREFS_STATICS
    const PWSprefs *prefs = PWSprefs::GetInstance();
#else
    std::auto_ptr<PWSprefs> prefs (new PWSprefs);
    @synchronized(m_file) {
      prefs->Load(m_file.pwsCore.GetDBPreferences());
    }
#endif
    keep = prefs->GetPref(PWSprefs::SavePasswordHistory);
    uint max = prefs->GetPref(PWSprefs::NumPWHistoryDefault);
    maxEntries = MIN(max, std::numeric_limits<u_int8_t>::max());
  }
  else
  {
    size_t num_err;
    size_t max;
    keep = CreatePWHistoryList(pwHistoryStr, max, num_err, entries, PWSUtil::TMC_EXPORT_IMPORT);
    maxEntries = MIN(static_cast<uint>(max), std::numeric_limits<u_int8_t>::max());
  }
}

+ (BOOL)automaticallyNotifiesObserversOfPwHistory
{
  return NO;
}

- (void)setCItemData:(const CItemData *)cItemData isConst:(BOOL)isConst
{
  m_cItemData = cItemData;
  const CItemData &item = *cItemData;
  
  m_deleteCItemData = !isConst;
  
  m_group = [NSString stringWithwstring:item.GetGroup()] ?: @"";
  m_title = [NSString stringWithwstring:item.GetTitle()];
  m_username = [NSString stringWithwstring:item.GetUser()];
  m_url = [NSString stringWithwstring:item.GetURL()];
}

- (NSString *)password
{
  if(m_password)
    return m_password;

  m_password = [NSString stringWithwstring:m_cItemData->GetPassword()];
  return m_password;
}

- (NSString *)notes
{
  if(m_notes)
    return m_notes;
  
  m_notes = [NSString stringWithwstring:m_cItemData->GetNotes()];
  return m_notes;
}

- (NSString *)url
{
  if(m_url)
    return m_url;
  
  m_url = [NSString stringWithwstring:m_cItemData->GetURL()];
  return m_url;
}

- (NSString *)email
{
  if(m_email)
    return m_email;
  
  m_email = [NSString stringWithwstring:m_cItemData->GetEmail()];
  return m_email;
}

- (BOOL)isProtected
{
  if(m_isProtected)
    return m_isProtected.boolValue;
  
  BOOL prot = (BOOL)m_cItemData->IsProtected();
  m_isProtected = @(prot);
  return prot;
}

- (NSString *)pwPolicyName
{
  if(m_pwPolicyName)
    return m_pwPolicyName;
  
  m_pwPolicyName = [NSString stringWithwstring:m_cItemData->GetPolicyName()];
  return m_pwPolicyName;
}

- (NSDate *)mTime
{
  if(m_mTime)
    return m_mTime;
  
  time_t mTime = 0;
  m_cItemData->GetRMTime(mTime);
  if(mTime)
    m_mTime = [NSDate dateWithTimeIntervalSince1970:mTime];

  return m_mTime;
}

- (NSDate *)cTime
{
  if(m_cTime)
    return m_cTime;
  
  time_t cTime = 0;
  m_cItemData->GetCTime(cTime);
  if(cTime)
    m_cTime = [NSDate dateWithTimeIntervalSince1970:cTime];
  
  return m_cTime;
}


- (id) initWithEntry:(Entry *)entry file:(SafeFile *)file
{
  if (!(self = [super init]))
    return nil;
  
  m_cItemData = new CItemData(*entry.cItemData);
  m_deleteCItemData = YES;
  m_title = [entry.title copy];
  m_group = [entry.group copy];
  m_username = [entry.username copy];

  m_file = file;
  
  m_neverSetPassword = NO;
  
  return self;
}

- (BOOL)validateValue:(inout __autoreleasing id *)ioValue forKey:(NSString *)inKey error:(out NSError *__autoreleasing *)outError
{
  NSString *valueStr = *ioValue;
  
  if([inKey isEqualToString:@"title"] ||
     [inKey isEqualToString:@"password"])
  {
    if(valueStr && valueStr.length > 0)
      return YES;
    
    if(outError)
    {
      NSString *errDesc = [NSString stringWithFormat:@"The %@ cannot be empty", inKey];
      *outError = [NSError errorWithDomain:@"EntryErrorDomain" code:1 userInfo:@{NSLocalizedDescriptionKey: errDesc}];
    }
    return NO;
  }
  
  return [super validateValue:ioValue forKey:inKey error:outError];
}

- (CItemData *)mutableItemData
{
  NSAssert(m_deleteCItemData, @"Mutable is deletabel");
  return const_cast<CItemData *>(m_cItemData);
}

- (NSComparisonResult) caseInsensitiveCompare:(Entry *)aEntry
{
  return [self.title caseInsensitiveCompare:aEntry.title];
}

-(void)dealloc
{
  if(m_deleteCItemData)
    delete self.mutableItemData;
}

- (NSData *)uuid
{
  uuid_array_t uuid;
  m_cItemData->GetUUID(uuid);
  return [NSData dataWithBytes:&uuid length:sizeof(uuid_array_t)];
}

- (void)effectivePasswordPolicy:(PWPolicy &)policy
{
  bool polSet = false;
  if(m_cItemData->IsPolicyNameSet())
  {
    @synchronized(m_file) {
      polSet = m_file.pwsCore.GetPolicyFromName(m_cItemData->GetPolicyName(), policy);
    }
  }
  
  if(!polSet)
  {
    m_cItemData->GetPWPolicy(policy);
    if(m_cItemData->IsSymbolsSet())
      policy.symbols = m_cItemData->GetSymbols();
  }
}

- (NSString *) randomPassword
{
  PWPolicy pol;
  [self effectivePasswordPolicy:pol];

  try
  {
#ifdef PWSPREFS_STATICS
    return [NSString stringWithwstring:pol.MakeRandomPassword()];
#else
    std::auto_ptr<PWSprefs> prefs (new PWSprefs);
    @synchronized(m_file) {
      prefs->Load(m_file.pwsCore.GetDBPreferences());
    }
    return [NSString stringWithwstring:pol.MakeRandomPassword(prefs.get())];
#endif
  }
  catch(const std::exception &exception)
  {
    A77Log(@"C++ exception generating password: %s", exception.what());
    return [NSString string];
  }
  catch(...)
  {
    A77Log(@"C++ unknown exception type generating password");
    return [NSString string];
  }
}

- (BOOL)passwordPolicyNumericOnly
{
  PWPolicy pol;
  [self effectivePasswordPolicy:pol];
  return ((pol.flags & ( PWPolicy::UseLowercase |
                         PWPolicy::UseUppercase |
                         PWPolicy::UseDigits |
                         PWPolicy::UseSymbols |
                         PWPolicy::UseHexDigits |
                         PWPolicy::UseEasyVision |
                         PWPolicy::MakePronounceable)) == PWPolicy::UseDigits) ? YES : NO;
}

- (BOOL)hasCustomPasswordPolicy
{
  return m_cItemData->IsPasswordPolicySet() ? YES : NO;
}

- (void) getPasswordPolicy:(PWPolicy &)policy;
{
  const CItemData *item = m_cItemData;
  item->GetPWPolicy(policy);
  if(item->IsSymbolsSet())
    policy.symbols = item->GetSymbols();
}

- (void)setPasswordPolicy:(const PWPolicy &)policy
{
  CItemData *item = const_cast<CItemData *>(m_cItemData);
  NSAssert(m_deleteCItemData, @"Must be non-const");
  item->SetPWPolicy(policy);
  item->SetSymbols(policy.symbols);
}

- (NSURL *)parsedURL
{
  NSString *urlStr = self.url;
  
  if(urlStr.length == 0)
    return nil;
  
  static NSString * const prefixes[] = {
    @"[alt]",
    @"[ssh]",
    @"{alt}",
    @"[autotype]",
  };
  
  for(auto prefix : prefixes)
  {
    if([urlStr rangeOfString:prefix].location == 0)
    {
      urlStr = [urlStr substringFromIndex:prefix.length];
      break;
    }
  }
  
  NSURL *url = [NSURL URLWithString:urlStr];
  NSString *scheme = url.scheme;
  
  if((!url || !scheme || !scheme.length)  && [urlStr rangeOfString:@"http://" options:NSCaseInsensitiveSearch].location != 0 &&
     [urlStr rangeOfString:@"https://" options:NSCaseInsensitiveSearch].location != 0)
    url = [NSURL URLWithString:[@"http://" stringByAppendingString:urlStr]];
  if(!url)
    DebugLog(@"Could not create URL from '%@'", urlStr);
  
  return url;
}

- (void)clearPwHistory
{
  [self willChangeValueForKey:@"pwHistory"];

  bool keep;
  uint max;
  PWHistList entries;
  [self parsePwHistoryFromItem:*m_cItemData keep:keep max:max entries:entries];
  max = MIN(max, std::numeric_limits<u_int8_t>::max());
  auto header = ::MakePWHistoryHeader(keep, max, 0);
  CItemData *item = self.mutableItemData;
  item->SetPWHistory(header);
  
  [self didChangeValueForKey:@"pwHistory"];
}

static inline void setter(CItemData *item, __strong NSString *&var, NSString *val, void (CItemData::*setter)(const StringX &title, TCHAR delimiter))
{
  if(!val)
    val = @"";
  var = [val copy];
  (item->*setter)([var wstring], 0);
}

static inline void setter(CItemData *item, __strong NSString *&var, NSString *val, void (CItemData::*setter)(const StringX &title))
{
  if(!val)
    val = @"";
  var = [val copy];
  (item->*setter)([var wstring]);
}

#ifndef PWSPREFS_STATICS      

static inline void setter(CItemData *item, __strong NSString *&var, NSString *val, const PWSprefs *prefs, void (CItemData::*setter)(const StringX &title, const PWSprefs *prefs))
{
  if(!val)
    val = @"";
  var = [val copy];
  (item->*setter)([var wstring], prefs);
}

#endif

- (void)setTitle:(NSString *)title
{
  setter(self.mutableItemData, m_title, title, &CItemData::SetTitle);
}

- (void)setUsername:(NSString *)username
{
  setter(self.mutableItemData, m_username, username, &CItemData::SetUser);
}

- (void)setPassword:(NSString *)password
{
  [self willChangeValueForKey:@"pwHistory"];
  
#ifdef PWSPREFS_STATICS
  setter(self.mutableItemData, m_password, password, &CItemData::UpdatePassword);
#else
  std::auto_ptr<PWSprefs> prefs(new PWSprefs);
  @synchronized(m_file) {
    prefs->Load(m_file.pwsCore.GetDBPreferences());
  }
  if(m_neverSetPassword)
  {
    setter(self.mutableItemData, m_password, password, &CItemData::SetPassword);
    self.mutableItemData->SetPMTime();
    m_neverSetPassword = NO;
  }
  else
  {
    setter(self.mutableItemData, m_password, password, prefs.get(), &CItemData::UpdatePassword);
  }
#endif
  
  [self didChangeValueForKey:@"pwHistory"];
}

- (void)setGroup:(NSString *)group
{
  setter(self.mutableItemData, m_group, group, &CItemData::SetGroup);
}

- (void)setUrl:(NSString *)url
{
  setter(self.mutableItemData, m_url, url, &CItemData::SetURL);
}

- (void)setEmail:(NSString *)email
{
  setter(self.mutableItemData, m_email, email, &CItemData::SetEmail);
}

- (void)setIsProtected:(BOOL)isProtected
{
  m_isProtected = @(isProtected);
  self.mutableItemData->SetProtected((bool)isProtected);
}

- (void)setPwPolicyName:(NSString *)pwPolicyName
{
  setter(self.mutableItemData, m_pwPolicyName, pwPolicyName, &CItemData::SetPolicyName);
}

- (void)setMTime:(NSDate *)mDate
{
  time_t mTime = mDate ? rint(mDate.timeIntervalSince1970) : 0;
  self.mutableItemData->SetRMTime(mTime);
  m_mTime = mDate;
}

- (void)setCTime:(NSDate *)cDate
{
  time_t cTime = cDate ? rint(cDate.timeIntervalSince1970) : 0;
  self.mutableItemData->SetCTime(cTime);
  m_cTime = cDate;
}

- (void)setNotes:(NSString *)notes
{
  if(!notes)
  {
    setter(self.mutableItemData, m_notes, nil, &CItemData::SetNotes);    
    return;
  }
  NSCharacterSet *newLine = [NSCharacterSet characterSetWithCharactersInString:@"\n"];
  NSUInteger idx = 0;
  NSMutableString *newNotes = [NSMutableString stringWithCapacity:notes.length + 20];
  NSString *crlf = @"\r\n";
  NSString *lf = @"\n";
  for(;;)
  {
    NSRange range = [notes rangeOfCharacterFromSet:newLine options:0 range:NSMakeRange(idx, notes.length - idx)];
    NSUInteger loc = range.location;
    if(loc == NSNotFound)
    {
     [newNotes appendString:[notes substringWithRange:NSMakeRange(idx, notes.length - idx)]];
      break;
    }
    [newNotes appendString:[notes substringWithRange:NSMakeRange(idx, loc - idx)]];
    if(loc == 0 || [notes characterAtIndex:loc - 1] != '\r')
      [newNotes appendString:crlf];
    else
      [newNotes appendString:lf];

    idx = loc + 1;
  }
  setter(self.mutableItemData, m_notes, newNotes, &CItemData::SetNotes);
}

- (void)setPwHistoryKeep:(BOOL)pwHistoryKeep
{
  CItemData &ci = *self.mutableItemData;
  StringX cs_tmp = ci.GetPWHistory();
  
  if(pwHistoryKeep)
  {
    if (cs_tmp.length() < 5)
    {
      StringX text;
      Format(text, _T("1%02x00"), m_file.defaultPwHistoryEntries);
      ci.SetPWHistory(text);
    }
    else if (cs_tmp[0] == L'0')
    {
      cs_tmp[0] = L'1';
      ci.SetPWHistory(cs_tmp);
    }
  }
  else
  {
    if (cs_tmp.length() >= 5 && cs_tmp[0] == L'1')
    {
      cs_tmp[0] = L'0';
      ci.SetPWHistory(cs_tmp);
    }
  }
}

- (void)setPwHistoryMaxEntries:(uint)pwHistoryMaxEntries
{
  if(pwHistoryMaxEntries > std::numeric_limits<u_int8_t>::max())
    pwHistoryMaxEntries = std::numeric_limits<u_int8_t>::max();
  
  CItemData &ci = *self.mutableItemData;
  StringX cs_tmp = ci.GetPWHistory();
  
  size_t len = cs_tmp.length();
  if (len < 5)
    return;
  st_PWH_status st_pwhs;
  st_pwhs.pwh = cs_tmp;
  st_pwhs.es = ci.GetStatus();
  int status, old_max, num_saved;
  const wchar_t *lpszPWHistory = cs_tmp.c_str();
#if (_MSC_VER >= 1400)
  int iread = swscanf_s(lpszPWHistory, _T("%01d%02x%02x"),
                        &status, &old_max, &num_saved);
#else
  int iread = swscanf(lpszPWHistory, _T("%01d%02x%02x"),
                      &status, &old_max, &num_saved);
#endif
  if (iread == 3 && status == 1 && num_saved <= pwHistoryMaxEntries)
  {
    StringX prefix;
    Format(prefix, _T("1%02x"), pwHistoryMaxEntries);
    cs_tmp = prefix + cs_tmp.substr(3);
    ci.SetPWHistory(cs_tmp);
  }
}

- (void) addPasswordToHistory:(NSString *)password withPMtime:(time_t)time
{
  bool keep;
  uint maxEntries;
  PWHistList histEntries;
  [self parsePwHistoryFromItem:*m_cItemData keep:keep max:maxEntries entries:histEntries];
  if(!keep)
    return;
  PWHistEntry histEntry;
  histEntry.password = [password wstring];
  histEntry.changetttdate = time;
  histEntry.changedate = PWSUtil::ConvertToDateTimeString(time, PWSUtil::TMC_EXPORT_IMPORT);
  
  if (histEntry.changedate.empty())
  {
    StringX unk;
    LoadAString(unk, IDSC_UNKNOWN);
    histEntry.changedate = unk;
  }
  
  histEntries.push_back(histEntry);
  if(histEntries.size() > maxEntries)
  {
    histEntries = PWHistList(histEntries.begin() + histEntries.size() - maxEntries, histEntries.end());
    NSAssert(histEntries.size() == maxEntries, @"Must max out");
  }
  
  // Now create string version!
  StringX new_PWHistory, buffer;
  
  Format(new_PWHistory, _T("1%02x%02x"), maxEntries, histEntries.size());
  
  PWHistList::iterator iter;
  for (iter = histEntries.begin(); iter != histEntries.end(); iter++)
  {
    Format(buffer, _T("%08x%04x%s"), static_cast<long>(iter->changetttdate), iter->password.length(), iter->password.c_str());
    new_PWHistory += buffer;
    buffer = _T("");
  }
  
  self.mutableItemData->SetPWHistory(new_PWHistory);
}

- (BOOL)matchesSearchTerm:(NSString *)searchText titleOnly:(BOOL)titleOnly
{
  if(!self.title)
    self.title = @"";
  
  NSRange range = [self.title rangeOfString:searchText options:NSCaseInsensitiveSearch | NSDiacriticInsensitiveSearch];
  if(range.location != NSNotFound)
    return YES;
  if(titleOnly)
    return NO;

  // Sanitize other fields
  if(!self.username)
    self.username = @"";
  if(!self.url)
    self.url = @"";
  if(!self.group)
    self.group = @"";
  if(!self.notes)
    self.notes = @"";
  if(!self.email)
    self.email = @"";
  if(!self.password)
    self.password = @"";
  
  NSArray *strings = @[self.username, self.url, self.group, self.notes, self.email, self.password];
  
  for(NSString *string in strings)
  {
    range = [string rangeOfString:searchText options:NSCaseInsensitiveSearch | NSDiacriticInsensitiveSearch];
    if(range.location != NSNotFound)
      return YES;
  }
  
  return NO;
}

- (NSString *)details
{
  NSString *url = self.url;
  NSString *username = self.username;
  
  BOOL hasURL = url != nil && ![url isEqualToString:@""];
  BOOL urlHasMethod = url != nil && [url rangeOfString:@"://"].location != NSNotFound;
  BOOL hasUsername = username != nil && ![username isEqualToString:@""];
  BOOL usernameIsEmail = username != nil && [username rangeOfString:@"@"].location != NSNotFound;
  
  if(hasUsername)
  {
    if(hasURL && !usernameIsEmail && !urlHasMethod)
      return [NSString stringWithFormat:@"%@@%@", username, url];
    else
      return username ? username : @"";
  }
  else 
  {
    return url ? url : @"";
  }
}

- (NSString *)description
{
  return [NSString stringWithFormat:@"Entry [%@/%@]", self.title, self.username];
}

- (NSArray *)pwHistory
{
  bool keep;
  uint max;
  PWHistList entries;
  [self parsePwHistoryFromItem:*m_cItemData keep:keep max:max entries:entries];
  if(entries.empty())
    return @[];
  
  NSMutableArray *ret = [NSMutableArray arrayWithCapacity:entries.size()];
  for(const PWHistEntry &entry : entries)
  {
    PwHistoryEntry *he = [[PwHistoryEntry alloc]init];
    he.password = [NSString stringWithwstring:entry.password];
    he.changeTimestamp = [NSDate dateWithTimeIntervalSince1970:entry.changetttdate];
    [ret addObject:he];
  }
  
  return ret;
}

- (uint)pwHistoryMaxEntries
{
  bool keep;
  uint max;
  PWHistList entries;
  [self parsePwHistoryFromItem:*m_cItemData keep:keep max:max entries:entries];

  return max;
}

- (BOOL)pwHistoryKeep
{
  bool keep;
  uint max;
  PWHistList entries;
  [self parsePwHistoryFromItem:*m_cItemData keep:keep max:max entries:entries];

  return keep;
}

@end

@implementation PwHistoryEntry

@end
