//
//  EntryDetailsViewController.m
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/21/11.
//  Copyright 2011 app77. All rights reserved.
//

#import "EntryDetailsViewController.h"
#import "Entry.h"
#import "SafeFile.h"
#import "MBProgressHUD.h"
#import "SimpleDetailTableViewCell.h"
#import "SelectGroupViewController.h"
#import "EditEntryPolicyVC.h"
#import "PwHistoryVC.h"
#import "SafeCtx.h"
#import "AppTheme.h"
#import "PasteboardPasswordCleaner.h"
#import "OpenInChromeController.h"
#import "PasteboardGravitationalManager.h"

NSString *kNewEntryTitle = @"New entry";

#define TITLE_ROW_IDX 0
#define USERNAME_ROW_IDX 1
#define PASSWORD_ROW_IDX 2
#define URL_ROW_IDX 3
#define EMAIL_ROW_IDX 4
#define GROUP_ROW_IDX 5
#define HISTORY_ROW_IDX 6
#define POLICY_ROW_IDX 7

static NSString * const kSmartPasteboardEnabledKey = @"smartPasteboardEnabled";

@interface EntryDetailsViewController() <UIAlertViewDelegate, PwHistoryVCDelegate>
{
  Entry *m_entry;
  SafeFile *m_file;
  
  BOOL m_showingFields;
  BOOL m_popOnCancel;
  
  NSString *m_newNotes;
  NSString *m_newTitle;
  NSString *m_newUser;
  NSString *m_newPassword;
  NSString *m_newURL;
  NSString *m_newEmail;
  NSString *m_newGroup;
  SelectGroupViewController *m_selectGroupViewController;
  EditEntryPolicyVC *m_editPwPolicyVC;
  PwHistoryVC *m_historyVC;
  SafeCtx *m_safeCtx;
  NSIndexPath *m_indexPathToFocusWhenScrollEnds;
  UIBarButtonItem *m_originalBackButtonItem;
  UIBarButtonItem *m_lockBarButtonItem;
}

@end

@implementation EntryDetailsViewController

@synthesize popOnCancel = m_popOnCancel;

- (id) initWithFile:(SafeFile *)file ctx:(SafeCtx *)ctx entry:(Entry *)entry
{
  self = [super initWithStyle:UITableViewStyleGrouped];
  if (!self) 
    return nil;

  self.navigationItem.title = entry.title;
  m_entry = entry;
  m_file = file;
  m_safeCtx = ctx;

  UIBarButtonItem *showFieldsBtn = [UIBarButtonItem buttonWithImageNamed:imageName(@"Glasses") target:self action:@selector(onShowOrHideFields:)];
  showFieldsBtn.accessibilityLabel = @"Hide/Un-hide";
  showFieldsBtn.accessibilityHint = @"Shows or hides password and notes.";
  UIBarButtonItem *spacer = [[UIBarButtonItem alloc]initWithBarButtonSystemItem:UIBarButtonSystemItemFlexibleSpace target:nil action:nil];
  UIBarButtonItem *goBtn = [UIBarButtonItem buttonWithImageNamed:imageName(@"OpenSafari") target:self action:@selector(onCopyAndGoToURL:)];
  goBtn.accessibilityLabel = @"Open";
  m_lockBarButtonItem = [UIBarButtonItem buttonWithImageNamed:imageName(entry.isProtected ?@"entryLocked" : @"entryUnlocked") target:self action:@selector(protect:)];
  m_lockBarButtonItem.accessibilityLabel = entry.isProtected ? @"Locked" : @"Unlocked";
  m_lockBarButtonItem.accessibilityHint = entry.isProtected ? @"Unlocks entry." : @"Locks entry.";
  
  NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
  NSNumber *showHiddenFieldsObj = [defaults objectForKey:@"revealHiddenFields"];
  m_showingFields = showHiddenFieldsObj && showHiddenFieldsObj.boolValue;
  
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)  
  {
    self.toolbarItems = m_showingFields ? @[m_lockBarButtonItem, spacer, goBtn] : @[m_lockBarButtonItem, spacer, showFieldsBtn, spacer, goBtn];
  }
  else
  {
    UIBarButtonItem *closeBtn = [UIBarButtonItem buttonWithImageNamed:imageName(@"Close") target:self action:@selector(onCloseSafe:)];
    closeBtn.accessibilityLabel = @"Close";
    
    self.toolbarItems = m_showingFields ? @[closeBtn, spacer, m_lockBarButtonItem, spacer, goBtn] : @[closeBtn, spacer, m_lockBarButtonItem, spacer, showFieldsBtn, spacer, goBtn];
  }
  self.navigationItem.rightBarButtonItem = self.editButtonItem;
  self.editButtonItem.enabled = !entry.isProtected;
  UIBarButtonItem *backButton = [[UIBarButtonItem alloc] initWithTitle:@"Entry" style:UIBarButtonItemStyleBordered target:nil action:nil];
  self.navigationItem.backBarButtonItem = backButton;
  
  return self;
}

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter]removeObserver:self];
  if(self.isViewLoaded)
  {
    self.tableView.delegate = nil;
    self.tableView.dataSource = nil;
  }
}

- (UIBarButtonItem *)btnCopyAndGo
{
  return [self.toolbarItems lastObject];
}

NSString * const kOpenInChromeKey;

static BOOL smartPasteboardEnabled()
{
  NSNumber *value = [[NSUserDefaults standardUserDefaults]objectForKey:kSmartPasteboardEnabledKey];
  return value ? value.boolValue : YES;
}

- (void)copyPasswordAndUsername
{
  Entry *entry = m_entry;
  Entry *baseEntry = [m_file baseEntryForEntry:entry];
  NSString *password = (baseEntry ? baseEntry : entry).password;
  NSString *username = smartPasteboardEnabled() ? entry.username : nil;
  
  if(password && password.length)
  {
    MBProgressHUD *hud = [self showHudWithMessage:@"Password copied"];
    
    if(!username.length)
    {
      [UIPasteboard generalPasteboard].string = password;
      [PasteboardPasswordCleaner passwordCopied:password];
    }
    else
    {
      PasteboardGravitationalManager *pgm = [PasteboardGravitationalManager sharedManager];
      [pgm copyNames:@[@"Password", @"Username"] values:@[password, username] passwordIdx:0];
      hud.detailsLabelText = @"Tilt device face down to copy username\nwhen using another app.";
    }
  }
}

- (void)onCopyAndGoToURL:(id)sender
{
  NSURL *url = [m_entry parsedURL];
  if(!url)
    return;
  
  [self copyPasswordAndUsername];
  
  dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, (int64_t)(0.3 * NSEC_PER_SEC));
  dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
    NSString *scheme = url.scheme;
    BOOL web = [scheme isEqualToString:@"http"] || [scheme isEqualToString:@"https"];
    
    if(web && [[NSUserDefaults standardUserDefaults]boolForKey:kOpenInChromeKey] && [OpenInChromeController sharedInstance].isChromeInstalled)
      [[OpenInChromeController sharedInstance]openInChrome:url];
    else
      [[UIApplication sharedApplication] openURL:url];
  });

}

- (void)onShowOrHideFields:(id)sender
{
  m_showingFields = !m_showingFields;
  NSIndexPath *passwordIdx = [NSIndexPath indexPathForRow:PASSWORD_ROW_IDX inSection:0];
  NSIndexPath *notesIdx = [NSIndexPath indexPathForRow:0 inSection:1];
  
  [self.tableView reloadRowsAtIndexPaths:@[passwordIdx, notesIdx] withRowAnimation:UITableViewRowAnimationFade];
}

- (void)onCloseSafe:(id)sender
{
  [m_safeCtx expireNow];
  m_safeCtx.opened = NO;
  [UIPasteboard generalPasteboard].string = @"";
  
  UIViewController *currentVC = self.navigationController.viewControllers.lastObject;
  [currentVC dismissModalViewControllerAnimated:YES];
  [self.navigationController popToRootViewControllerAnimated:YES];
  
  UIView *rootView = [MBProgressHUD rootView];
  MBProgressHUD *hud = [[MBProgressHUD alloc]initWithView:rootView];
  hud.customView = [[UIImageView alloc]initWithImage:[UIImage imageNamed:@"Check.png"]];
  hud.mode = MBProgressHUDModeCustomView;
  hud.labelText = @"Clipboard cleared";
  hud.animationType = MBProgressHUDAnimationZoom;
  hud.yOffset = -50;
  hud.opacity = .8;
  hud.removeFromSuperViewOnHide = YES;
  [rootView addSubview:hud];
  [hud show:YES];
  uint32_t secs = UIAccessibilityIsVoiceOverRunning() ? 3 : 1;
  dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, secs * NSEC_PER_SEC);
  dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
    [hud hide:YES];
  });
}

- (void)protect:(id)sender
{
  NSAssert(!self.isEditing, @"Cannot change protection of an entry being edited");
  BOOL newProtected = !m_entry.isProtected;
  Entry *newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
  newEntry.isProtected = newProtected;
  [m_file editEntry:m_entry newEntry:newEntry];
  m_entry = newEntry;
  self.editButtonItem.enabled = !newProtected;
  UIBarButtonItem *oldLockBtn = m_lockBarButtonItem;
  NSString *imgName = imageName(newProtected ? @"entryLocked" : @"entryUnlocked");
  m_lockBarButtonItem = [UIBarButtonItem buttonWithImageNamed:imgName target:self action:@selector(protect:)];
  m_lockBarButtonItem.accessibilityLabel = newProtected ? @"Locked" : @"Unlocked";
  m_lockBarButtonItem.accessibilityHint = newProtected ? @"Unlocks entry." : @"Locks entry.";

  NSMutableArray *toolbarItems = [self.toolbarItems mutableCopy];
  NSUInteger idx = [toolbarItems indexOfObjectIdenticalTo:oldLockBtn];
  NSAssert(idx != NSNotFound, @"old toolbar button must be there");
  toolbarItems[idx] = m_lockBarButtonItem;
  self.toolbarItems = toolbarItems;
  
  UIView *rootView = [MBProgressHUD rootView];
  MBProgressHUD *hud = [[MBProgressHUD alloc]initWithView:rootView];
  hud.customView = [[UIImageView alloc]initWithImage:[UIImage imageNamed:imgName]];
  hud.mode = MBProgressHUDModeCustomView;
  hud.labelText = newProtected ? @"Entry Protected" : @"Entry unprotected";
  hud.animationType = MBProgressHUDAnimationZoom;
  hud.opacity = .8;
  hud.removeFromSuperViewOnHide = YES;
  [rootView addSubview:hud];
  [hud show:YES];
  uint32_t secs = UIAccessibilityIsVoiceOverRunning() ? 3 : 1;
  dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, secs * NSEC_PER_SEC);
  dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
    [hud hide:YES];
  });
  
}

- (void)safeWasInternallyModified
{
  if(m_newGroup)
  {
    if([m_file.entries.allKeys indexOfObject:m_newGroup] == NSNotFound)
      m_newGroup = @"";
  }
  [self.tableView reloadData];

  [m_selectGroupViewController safeWasInternallyModified];
  [m_historyVC safeWasInternallyModified];
  [m_editPwPolicyVC safeWasInternallyModified];
}

- (void)groupRenamedFrom:(NSString *)oldGroupName to:(NSString *)newGroupName
{
  BOOL reloadGroup;
  if(m_newGroup && [m_newGroup isEqualToString:oldGroupName])
  {
    reloadGroup = YES;
    m_newGroup = [newGroupName copy];
  }
  else
  {
    reloadGroup = [m_entry.group isEqualToString:newGroupName];
  }
  
  if(reloadGroup)
    [self.tableView reloadRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:GROUP_ROW_IDX inSection:0]] withRowAnimation:UITableViewRowAnimationNone];
  
  [m_selectGroupViewController groupRenamedFrom:oldGroupName to:newGroupName];
}

#pragma mark - Forcefully close edition

- (void)alertViewCancel:(UIAlertView *)alertView
{
  [self onCancelEdit:alertView];
}

- (void)alertView:(UIAlertView *)alertView didDismissWithButtonIndex:(NSInteger)buttonIndex
{
  if(self.isEditing && buttonIndex == alertView.cancelButtonIndex)
    [self onCancelEdit:alertView];
  else
    [self setEditing:NO animated:YES];
}

#pragma mark - Text changed handlers

- (void)onNotesChanged:(NSNotification *)notification
{
  UITextView *txtView = notification.object;
  NSString *newNotes = txtView.text;
  if([newNotes isEqualToString:m_entry.notes])
    m_newNotes = nil;
  else
    m_newNotes = [newNotes copy];
}

- (void)onTitleChangedTo:(NSString *)newTitle
{
  if([newTitle isEqualToString:m_entry.title])
  {
    m_newTitle = nil;
  }
  else
  {
    m_newTitle = [newTitle copy];
    self.navigationItem.title = newTitle;
  }
}

- (void)onUserChangedTo:(NSString *)newUser
{
  if([newUser isEqualToString:m_entry.username])
    m_newUser = nil;
  else
    m_newUser = [newUser copy];
}

- (void)onPasswordChangedTo:(NSString *)newPassword
{
  if([newPassword isEqualToString:m_entry.password])
    m_newPassword = nil;
  else
    m_newPassword = [newPassword copy];
}

- (void)onEmailChangedTo:(NSString *)newEmail
{
  if([newEmail isEqualToString:m_entry.email])
    m_newEmail = nil;
  else
    m_newEmail = [newEmail copy];
}

- (void)onURLChangedTo:(NSString *)newURL
{
  if([newURL isEqualToString:m_entry.url])
    m_newURL = nil;
  else
    m_newURL = [newURL copy];
}

#pragma mark - View lifecycle

- (void)viewDidLoad
{
  self.tableView.allowsSelectionDuringEditing = YES;
  [self btnCopyAndGo].enabled = m_entry.parsedURL != nil;
  [super viewDidLoad];
  [self.tableView themeViewDidLoadInIpadPos:TVP_RIGHT];
}

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
  // Return YES for supported orientations
  return (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad || interfaceOrientation == UIInterfaceOrientationPortrait);
}

- (BOOL)passwordIsNumeric
{
  if(m_editPwPolicyVC)
    return m_editPwPolicyVC.passwordPolicyNumericOnly;
  else if(m_entry.hasCustomPasswordPolicy || (m_entry.pwPolicyName && m_entry.pwPolicyName.length))
    return m_entry.passwordPolicyNumericOnly;
  else
    return m_file.defaultPwPolicyNumericOnly;
}

- (void)viewWillAppear:(BOOL)animated
{
  [super viewWillAppear:animated];

  SimpleDetailTableViewCell *cell = (SimpleDetailTableViewCell *)[self.tableView cellForRowAtIndexPath:[NSIndexPath indexPathForRow:PASSWORD_ROW_IDX inSection:0]];
  cell.detailTextField.keyboardType = [self passwordIsNumeric] ? UIKeyboardTypeNumberPad : UIKeyboardTypeDefault;
  
  // Needed here as well for new items creation
  [self.navigationController setToolbarHidden:self.editing animated:YES];
  
  if(!m_selectGroupViewController)
    return;
  // Copy the new group back
  NSString *newGroup = [m_selectGroupViewController selectedGroup];
  NSString *entryGroup = m_entry.group;
  if(!entryGroup)
    entryGroup = @"";
  
  if([newGroup isEqualToString:entryGroup])
    m_newGroup = nil;
  else
    m_newGroup = newGroup;
  
  m_selectGroupViewController = nil;
  [self.tableView reloadRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:GROUP_ROW_IDX inSection:0]] withRowAnimation:UITableViewRowAnimationNone];
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
  return self.isEditing ? 2 : 3;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
  if(!section)
    return self.isEditing ? 8 : 7;
  else if(section == 1)
    return 1;
  else
    return 2;
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
{
  if(section == 1)
    return @"Notes";
  if(section == 2)
    return @"Dates";
  return nil;
}

- (NSString *) notesTextToShow 
{
  NSString *text = m_newNotes ? m_newNotes : m_entry.notes;
  BOOL hide = !m_showingFields && !self.editing;
  if(hide)
  {
    NSUInteger length = 32;
    unichar dots[length];
    for(NSUInteger i = 0; i < length; i++)
      dots[i] = 0x2022; // bullet
    text = [NSString stringWithCharacters:dots length:length];
  }
  return text;
}

CGFloat cellWidth(void) 
{
  CGFloat w = UI_USER_INTERFACE_IDIOM() != UIUserInterfaceIdiomPad ? (iOS6() ? 295.0f : 320.0f) :
  UIInterfaceOrientationIsPortrait([UIApplication sharedApplication].statusBarOrientation)? 360.0 : 595.0;
  return w;
}

- (CGFloat)tableView:(UITableView *)tableView heightForRowAtIndexPath:(NSIndexPath *)indexPath
{
  if(indexPath.section == 0 || indexPath.section == 2)
    return tableView.rowHeight;
  
  if(self.editing)
  {
    if(UI_USER_INTERFACE_IDIOM() != UIUserInterfaceIdiomPad)
      return 160; 
    if(UIInterfaceOrientationIsPortrait([UIApplication sharedApplication].statusBarOrientation))
      return 480;
    else
      return 250;
  }

  NSString *cellText = [self notesTextToShow];
  
  if(!cellText || [cellText isEqualToString:@""])
    return tableView.rowHeight;
  
  UIFont *cellFont = iOS6() ? [UIFont systemFontOfSize:17] : [UIFont preferredFontForTextStyle:UIFontTextStyleBody];
  CGSize constraintSize = CGSizeMake(cellWidth() - 25, MAXFLOAT);
  CGSize labelSize = [cellText sizeWithFont:cellFont constrainedToSize:constraintSize lineBreakMode:UILineBreakModeWordWrap];
  
  if(iOS6())
    return labelSize.height + 20;
  else
    return labelSize.height + 30;
}

- (void) getContentsForIndexPath: (NSIndexPath *) indexPath text_p: (NSString **) text_p title_p: (NSString **) title_p  
{
  NSInteger section = indexPath.section;
  NSInteger row = indexPath.row;
  
  if(section == 1)
  {
    *text_p = m_newNotes ? m_newNotes : m_entry.notes;
    *title_p = @"Notes";
    return;
  }
  else if(section == 2)
  {
    NSDate *date = row ? m_entry.mTime : m_entry.cTime;
    static NSDateFormatter *formatter;
    if(!formatter)
    {
      formatter = [[NSDateFormatter alloc]init];
      formatter.timeStyle = NSDateFormatterShortStyle;
      formatter.dateStyle = NSDateFormatterShortStyle;
    }
    *text_p = [formatter stringFromDate:date];
    *title_p = row ? @"Changed" : @"Created";
    return;
  }

  switch(row)
  {
    case TITLE_ROW_IDX:
      *title_p = @"Title";
      *text_p = m_newTitle ? m_newTitle : m_entry.title;
      break;
    case USERNAME_ROW_IDX:
      *title_p = @"Username";
      *text_p = m_newUser ? m_newUser : m_entry.username;
      break;
    case PASSWORD_ROW_IDX:
      *title_p = @"Password";
      *text_p = m_newPassword ? m_newPassword : m_entry.password;
      break;
    case URL_ROW_IDX:
      *title_p = @"URL";
      *text_p = m_newURL ? m_newURL : m_entry.url;
      break;
    case EMAIL_ROW_IDX:
      *title_p = @"Email";
      *text_p = m_newEmail ? m_newEmail : m_entry.email;
      break;
    case GROUP_ROW_IDX:
      *title_p = @"Group";
      *text_p = [[(m_newGroup ? m_newGroup : m_entry.group) componentsSeparatedByString:@"."]lastObject];
      break;
    case HISTORY_ROW_IDX:
      *title_p = @"Password History";
      *text_p = nil;
      break;
    case POLICY_ROW_IDX:
      *title_p = @"Password Policy";
      *text_p = nil;
      break;
  }
}

- (SEL)onTextChangedForIndexPath:(NSIndexPath *)indexPath
{
  NSUInteger row = indexPath.row;
  
  NSAssert(indexPath.section == 0, @"Notes don't use this");
  NSAssert(row >= TITLE_ROW_IDX && row <= EMAIL_ROW_IDX, @"Unknown row");

  switch(row)
  {
    case TITLE_ROW_IDX:
      return @selector(onTitleChangedTo:);
    case USERNAME_ROW_IDX:
      return @selector(onUserChangedTo:);
    case PASSWORD_ROW_IDX:
      return @selector(onPasswordChangedTo:);
    case URL_ROW_IDX:
      return @selector(onURLChangedTo:);      
    case EMAIL_ROW_IDX:
      return @selector(onEmailChangedTo:);      
    default:
      return nil;
  }
}

- (SEL)onReturnPressedForIndexPath:(NSIndexPath *)indexPath
{
  NSUInteger row = indexPath.row;
  
  NSAssert(indexPath.section == 0, @"Notes don't use this");
  NSAssert(row >= TITLE_ROW_IDX && row <= EMAIL_ROW_IDX, @"Unknown row");
  
  switch(row)
  {
    case TITLE_ROW_IDX:
      return @selector(onTitleReturn);
    case USERNAME_ROW_IDX:
      return @selector(onUserReturn);
    case PASSWORD_ROW_IDX:
      return @selector(onPasswordReturn);
    case URL_ROW_IDX:
      return @selector(onURLReturn);      
    case EMAIL_ROW_IDX:
      return @selector(onEmailReturn);      
    default:
      return nil;
  }
}

- (void)setFocusTo:(NSIndexPath *)indexPath animated:(BOOL)animated
{
  UITableView *tableView = self.tableView;
  [tableView scrollToRowAtIndexPath:indexPath atScrollPosition:UITableViewRowAnimationTop animated:animated];
  SimpleDetailTableViewCell *cell = (SimpleDetailTableViewCell *)[tableView cellForRowAtIndexPath:indexPath];
  if(!cell)
  {
    m_indexPathToFocusWhenScrollEnds = indexPath;
    return;
  }
  NSAssert(cell, @"Must have a cell");
  BOOL ok = [cell.detailTextField becomeFirstResponder];
  NSAssert(ok, @"Must work");
  (void)ok;
}

- (void)scrollViewDidEndScrollingAnimation:(UIScrollView *)scrollView
{
  if(m_indexPathToFocusWhenScrollEnds)
  {
    NSIndexPath *indexPath = m_indexPathToFocusWhenScrollEnds;
    m_indexPathToFocusWhenScrollEnds = nil;
    [self setFocusTo:indexPath animated:NO];
  }
}

- (void)onTitleReturn
{
  [self setFocusTo:[NSIndexPath indexPathForRow:USERNAME_ROW_IDX inSection:0] animated:YES];
}

- (void)onUserReturn
{
  [self setFocusTo:[NSIndexPath indexPathForRow:PASSWORD_ROW_IDX inSection:0] animated:YES];
}

- (void)onPasswordReturn
{
  [self setFocusTo:[NSIndexPath indexPathForRow:URL_ROW_IDX inSection:0] animated:YES];
}

- (void)onURLReturn
{
  [self setFocusTo:[NSIndexPath indexPathForRow:EMAIL_ROW_IDX inSection:0] animated:YES];
}

#define TEXTVIEWTAG 129

- (void)onEmailReturn
{
  NSIndexPath *indexPath = [NSIndexPath indexPathForRow:0 inSection:1];
  UITableView *tableView = self.tableView;
  [tableView scrollToRowAtIndexPath:indexPath atScrollPosition:UITableViewRowAnimationNone animated:YES];
  UITableViewCell *cell = [tableView cellForRowAtIndexPath:indexPath];
  UITextView *txtView = (UITextView *)[cell.contentView viewWithTag:TEXTVIEWTAG];
  [txtView becomeFirstResponder];
}

- (void)setupNotesCell:(UITableViewCell *)cell
{
  UITextView *txtView = (UITextView *)[cell.contentView viewWithTag:TEXTVIEWTAG];
  BOOL editing = self.editing;
  BOOL hasAutolayout = [NSLayoutConstraint class] != nil;
  
  if(!txtView || !txtView.scrollEnabled != !editing)
  {
    if(txtView)
    {
      [txtView removeFromSuperview];
      [[NSNotificationCenter defaultCenter]removeObserver:self
                                                     name:UITextViewTextDidEndEditingNotification
                                                   object:txtView];
      if(hasAutolayout)
        [cell.contentView removeConstraints:cell.contentView.constraints];
    }
    
    txtView = [[UITextView alloc]initWithFrame:CGRectZero];
    txtView.tag = TEXTVIEWTAG;
    [cell.contentView addSubview:txtView];
    
    if(iOS6())
    {
      txtView.font = [UIFont systemFontOfSize:17];
    }
    else
    {
      txtView.font = [UIFont preferredFontForTextStyle:UIFontTextStyleBody];
    }
    txtView.backgroundColor = nil;
    txtView.textColor = [UIColor blackColor];
    txtView.opaque = NO;
    txtView.userInteractionEnabled = editing;
    txtView.scrollEnabled = editing;
    
    [[NSNotificationCenter defaultCenter]addObserver:self
                                            selector:@selector(onNotesChanged:)
                                                name:UITextViewTextDidEndEditingNotification
                                              object:txtView];
    
    if(hasAutolayout)
    {
      txtView.translatesAutoresizingMaskIntoConstraints = NO;
      NSDictionary *views = NSDictionaryOfVariableBindings(txtView);
      NSDictionary *metrics = @{ @"m" : @8, @"l" : @2, @"t" : @2, @"r": @2 };
      [cell.contentView addConstraints:[NSLayoutConstraint constraintsWithVisualFormat:@"V:|-t-[txtView]-m-|" options:0 metrics:metrics views:views]];
      [cell.contentView addConstraints:[NSLayoutConstraint constraintsWithVisualFormat:@"H:|-l-[txtView]-r-|" options:0 metrics:metrics views:views]];
    }
  }
  
  txtView.text = [self notesTextToShow];
  [txtView setContentOffset:CGPointMake(0, editing ? 1 : 0) animated:YES];
  
  if(!hasAutolayout)
  {
    CGFloat w = cellWidth();
    
    DebugLog(@"Created notes cell with editing = %d. w = %f", self.editing, w);
    
    txtView.frame = CGRectMake(2, 2, w, [self tableView:self.tableView heightForRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:1]] - 12);
  }

  txtView.showsVerticalScrollIndicator = YES;
}

- (void)configureAccessibilityForCell:(UITableViewCell *)cell atIndexPath:(NSIndexPath *)indexPath
{
  const NSInteger row = indexPath.row;
  const NSInteger section = indexPath.section;
  BOOL editing = self.isEditing;
  
  if(section == 2)
    return;
  if(section == 1)
  {
    cell.accessibilityTraits = editing ? UIAccessibilityTraitNone : UIAccessibilityTraitButton;
    cell.accessibilityHint = editing ? nil : @"Copies notes";
    return;
  }
  if(row == POLICY_ROW_IDX || row == HISTORY_ROW_IDX)
  {
    cell.accessibilityTraits = UIAccessibilityTraitButton;
    return;
  }
  cell.accessibilityTraits = editing ? UIAccessibilityTraitNone : UIAccessibilityTraitButton;
  if(row == GROUP_ROW_IDX)
    cell.accessibilityHint = editing ? @"Changes group." : [NSString stringWithFormat:@"Copies %@.", cell.textLabel.text];
  else
    cell.accessibilityHint = editing ? nil : [NSString stringWithFormat:@"Copies %@.", cell.textLabel.text];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
  const NSInteger row = indexPath.row;
  const NSInteger section = indexPath.section;
  
  if(section == 2)
  {
    NSString *title;
    NSString *text;
    static NSString *CellId = @"DateCell";
    [self getContentsForIndexPath: indexPath text_p: &text title_p: &title];

    UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:CellId];
    if(!cell)
      cell = [[UITableViewCell alloc]initWithStyle:UITableViewCellStyleValue2 reuseIdentifier:CellId];
    cell.textLabel.text = title;
    cell.detailTextLabel.text = text;
    cell.selectionStyle = UITableViewCellSelectionStyleNone;
    [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];
    [self configureAccessibilityForCell:cell atIndexPath:indexPath];
    return cell;
  }
  
  if(row == GROUP_ROW_IDX)
  {
    // Group 
    NSAssert(section == 0, @"Unknown indexpath");
    static NSString *CellIdentifier = @"GroupCell";
    
    UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
    
    if (cell == nil) 
    {
      cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue2 reuseIdentifier:CellIdentifier];
    }
    
    NSString *title;
    NSString *text;
    [self getContentsForIndexPath: indexPath text_p: &text title_p: &title];
    
    cell.textLabel.text = title;
    cell.detailTextLabel.text = text;
    cell.editingAccessoryType = UITableViewCellAccessoryDisclosureIndicator;
    
    [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];
    [self configureAccessibilityForCell:cell atIndexPath:indexPath];
    
    return cell;
  }
  if(row == POLICY_ROW_IDX || row == HISTORY_ROW_IDX)
  {
    // Password policy
    NSAssert(section == 0, @"Unknown indexpath");
    static NSString *CellIdentifier = @"PwPolicyOrHistCell";
    
    UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
    
    if (cell == nil) 
    {
      cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:CellIdentifier];
      if(iOS6())
      {
        cell.textLabel.font = [UIFont boldSystemFontOfSize:12];
        cell.textLabel.textColor = [UIColor colorWithRed:0.32 green:0.4 blue:0.57 alpha:1];
      }
      else
      {
        cell.textLabel.font = [UIFont systemFontOfSize:14];
        cell.textLabel.textColor = [UIColor colorWithRed:0 green:0.478431f blue:1 alpha:1];
      }
    }
    
    NSString *title;
    NSString *text;
    [self getContentsForIndexPath: indexPath text_p: &text title_p: &title];
    
    cell.textLabel.text = title;
    cell.editingAccessoryType = UITableViewCellAccessoryDisclosureIndicator;
    cell.accessoryType = row == HISTORY_ROW_IDX ? UITableViewCellAccessoryDisclosureIndicator : UITableViewCellAccessoryNone;
    
    [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];
    [self configureAccessibilityForCell:cell atIndexPath:indexPath];
    
    return cell;
  }
  
  if(section == 1)
  {
    static NSString *CellIdentifier = @"NotesCell";
    
    NSAssert(row == 0, @"Unknown indexpath");
    
    UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
    
    if (cell == nil)
      cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:CellIdentifier];

    [self setupNotesCell:cell];

    [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];
    [self configureAccessibilityForCell:cell atIndexPath:indexPath];
    
    return cell;
  }
  
  SEL onTextChanged = [self onTextChangedForIndexPath:indexPath];
  SEL onReturn = [self onReturnPressedForIndexPath:indexPath];
  
  static NSString *CellIdentifier = @"SimpleDetailCell";
  
  SimpleDetailTableViewCell *cell = (SimpleDetailTableViewCell *)[tableView dequeueReusableCellWithIdentifier:CellIdentifier];
  
  if (cell == nil) 
  {
    cell = [[SimpleDetailTableViewCell alloc] initWithStyle:UITableViewCellStyleValue2 reuseIdentifier:CellIdentifier];
    cell.detailTextField.returnKeyType = UIReturnKeyNext;
  }
  
  [cell setTarget:self textChanged:onTextChanged enterPressed:onReturn];
  
  NSString *title;
  NSString *text;
  [self getContentsForIndexPath: indexPath text_p: &text title_p: &title];
  
  NSAssert(section == 0, @"Section 1 has been deal with above");
  
  BOOL isPasswordCell = row == PASSWORD_ROW_IDX;
  
  cell.hideDetailWhenNotEditing = !m_showingFields && isPasswordCell; // Just passwords
  cell.autoCorrect = row == TITLE_ROW_IDX;
  
  if(isPasswordCell)
  {
    UIButton *newPasswordBtn = [UIButton buttonWithType:UIButtonTypeRoundedRect];
    [newPasswordBtn addTarget:self action:@selector(onNewPassword:) forControlEvents:UIControlEventTouchUpInside];
    [newPasswordBtn setTitle:@"New" forState:UIControlStateNormal];
    [newPasswordBtn setAccessibilityLabel:@"New password"];
    newPasswordBtn.titleLabel.font = [newPasswordBtn.titleLabel.font fontWithSize:12];
    cell.editButton = newPasswordBtn;
    if(iOS6())
      cell.detailTextField.font = [UIFont fontWithName:@"DejaVuSansMono-Bold" size:15.0f];
    else
      cell.detailTextField.font = [UIFont fontWithName:@"DejaVuSansMono" size:14.0f];
    cell.detailTextField.keyboardType = [self passwordIsNumeric] ? UIKeyboardTypeNumberPad : UIKeyboardTypeDefault;
  }
  else
  {
    cell.editButton = nil;
    
    switch(row)
    {
      case EMAIL_ROW_IDX:
        cell.detailTextField.keyboardType = UIKeyboardTypeEmailAddress;
        break;
      case URL_ROW_IDX:
        cell.detailTextField.keyboardType = UIKeyboardTypeURL;
        break;
      default:
        cell.detailTextField.keyboardType = UIKeyboardTypeDefault;
        break;
    }
  }
  
  cell.textLabel.text = title;
  cell.detailTextField.text = text;
  cell.detailTextField.accessibilityLabel = title;
  cell.detailTextField.autocapitalizationType = row == 0 ? UITextAutocapitalizationTypeSentences : UITextAutocapitalizationTypeNone;

  [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];
  [self configureAccessibilityForCell:cell atIndexPath:indexPath];
  
  return cell;
}

- (UITableViewCellEditingStyle)tableView:(UITableView *)tableView editingStyleForRowAtIndexPath:(NSIndexPath *)indexPath
{
  return UITableViewCellEditingStyleNone;
}

-(BOOL)tableView:(UITableView *)tableView shouldIndentWhileEditingRowAtIndexPath:(NSIndexPath *)indexPath
{
  return NO;
}

#pragma mark - Table view delegate

- (void)didRotateFromInterfaceOrientation:(UIInterfaceOrientation)fromInterfaceOrientation
{
  [self.tableView reloadRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:0 inSection:1]] withRowAnimation:UITableViewRowAnimationNone];
}

-(void)onNewPassword:(id)sender
{
  if(![self isEditing])
    return;
  NSIndexPath *passwordIdx = [NSIndexPath indexPathForRow:PASSWORD_ROW_IDX inSection:0];
  SimpleDetailTableViewCell *cell = (SimpleDetailTableViewCell *)[self.tableView cellForRowAtIndexPath:passwordIdx];
  NSString *newPassword;
  if(m_editPwPolicyVC)
    newPassword = [m_editPwPolicyVC randomPassword];
  else
    newPassword = [m_entry randomPassword];
  
  cell.detailTextField.text = newPassword;
  [self onPasswordChangedTo:newPassword];
}

-(void)onCancelEdit:(id)unusedSender
{
  A77Log(@"Cancel entry editing");
  
  if(m_popOnCancel)
  {
    [m_file deleteEntry:m_entry];
    [self.navigationController popViewControllerAnimated:YES];
  }
  else
  {
    [self.view endEditing:YES];
    
    m_newNotes = nil;
    m_newTitle = nil;
    m_newUser = nil;
    m_newPassword = nil;
    m_newURL = nil;
    m_newEmail = nil;
    m_newGroup = nil;
    m_editPwPolicyVC = nil;
    m_historyVC = nil;

    [self setEditing:NO animated:YES];
    if(!self.editing)
    {
      self.navigationItem.title = m_entry.title;
      [self.tableView reloadData];
    }
  }
}

- (BOOL)testCanEndEditingAndNotify:(BOOL)notify animated:(BOOL)animated
{
  NSString *badField;
  NSIndexPath *badFieldIndexPath;
  
  NSString *title = m_newTitle ? m_newTitle : m_entry.title;
  NSString *password = m_newPassword ? m_newPassword : m_entry.password;
  if(!title || !title.length)
  {
    badField = @"Title";
    badFieldIndexPath = [NSIndexPath indexPathForRow:TITLE_ROW_IDX inSection:0];
  }
  else if(!password || !password.length)
  {
    badField = @"Password";
    badFieldIndexPath = [NSIndexPath indexPathForRow:PASSWORD_ROW_IDX inSection:0];
  }
  
  if(!badField)
    return YES;
  
  if(notify)
  {
    UIView *view = [MBProgressHUD rootView];
    
    MBProgressHUD *hud = [[MBProgressHUD alloc]initWithView:view];
    hud.customView = [[UIImageView alloc]initWithImage:[UIImage imageNamed:@"Bad.png"]];
    hud.mode = MBProgressHUDModeCustomView;
    hud.labelText = [NSString stringWithFormat:@"%@ cannot be empty", badField];
    hud.animationType = MBProgressHUDAnimationZoom;
    hud.opacity = .8;
    hud.removeFromSuperViewOnHide = YES;
    hud.yOffset = -50;
    [view addSubview:hud];
    [hud show:YES];
    [self performSelector:@selector(hideHud:) withObject:hud afterDelay:UIAccessibilityIsVoiceOverRunning() ? 3 : 1];
    
    [self setFocusTo:badFieldIndexPath animated:animated && (self.navigationController.visibleViewController == self)];
  }
  
  return NO;
}

-(void)setEditing:(BOOL)editing animated:(BOOL)animated
{
  if(!editing && ![self testCanEndEditingAndNotify:YES animated:YES])
    return;

  [super setEditing:editing animated:animated];
  if(self.navigationController.visibleViewController == self && m_historyVC)
    [m_historyVC setEditing:editing animated:NO];
  
  NSIndexPath *rowOnlyInEdition = [NSIndexPath indexPathForRow:POLICY_ROW_IDX inSection:0];
  NSIndexPath *rowBeforeOnlyInEdition = [NSIndexPath indexPathForRow:POLICY_ROW_IDX - 1 inSection:0];
  NSArray *rowsOnlyInEdition = @[rowOnlyInEdition];
  UITableView *tableView = self.tableView;
  
  if(editing)
  {
    NSAssert(!m_newNotes && !m_newTitle && !m_newUser && !m_newPassword && !m_newURL && !m_newEmail && !m_newGroup, @"Begining of edition should be with saved values");

    m_editPwPolicyVC = nil;
    
    if([m_entry.title isEqualToString:kNewEntryTitle])
    {
      dispatch_async(dispatch_get_main_queue(), ^(void){
        SimpleDetailTableViewCell *cell = (SimpleDetailTableViewCell *)[self.tableView cellForRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:0]];
        UITextField *detailTF = cell.detailTextField;
        [detailTF becomeFirstResponder];
        dispatch_async(dispatch_get_main_queue(), ^{
          [detailTF selectAll:self];
        });
        void (^ __block hideMenu)() = [^{
          UIMenuController *menu = [UIMenuController sharedMenuController];
          if(menu.isMenuVisible)
          {
            [menu setMenuVisible:NO animated:NO];
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-retain-cycles"
            NSAssert(CFGetRetainCount((__bridge CFTypeRef)hideMenu) == 1, @"Must free the block");
            hideMenu = nil;
#pragma clang diagnostic pop
            return;
          }
          double delayInSeconds = 0.05;
          dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, delayInSeconds * NSEC_PER_SEC);
          dispatch_after(popTime, dispatch_get_main_queue(), hideMenu);
        } copy];
        hideMenu();
      });
    }

    [tableView beginUpdates];
    
    if([tableView numberOfRowsInSection:rowOnlyInEdition.section] != rowOnlyInEdition.row + 1)
    {
      [tableView insertRowsAtIndexPaths:rowsOnlyInEdition withRowAnimation:UITableViewRowAnimationTop];
      if(iOS6())
      {
        UITableViewCell *cell = [self.tableView cellForRowAtIndexPath:rowBeforeOnlyInEdition];
        [tableView addBackgroundsToGroupedCell:cell atIndexPath:rowBeforeOnlyInEdition];
      }
    }
    if(tableView.numberOfSections != 2)
      [tableView deleteSections:[NSIndexSet indexSetWithIndex:2] withRowAnimation:UITableViewRowAnimationAutomatic];

    [tableView endUpdates];
    
    UIBarButtonItem *cancelBtn = [[UIBarButtonItem alloc]initWithBarButtonSystemItem:UIBarButtonSystemItemCancel target:self action:@selector(onCancelEdit:)];
    m_originalBackButtonItem = self.navigationItem.leftBarButtonItem;
    self.navigationItem.leftBarButtonItem = cancelBtn;
    
    if(self.navigationController.visibleViewController == self)
      [self.navigationController setToolbarHidden:YES animated:YES];
  }
  else
  {
    BOOL ok = [self.view endEditing:YES];
    (void)ok;    
    NSAssert(ok, @"Must resign first responder");
    self.navigationItem.leftBarButtonItem = m_originalBackButtonItem;
    m_originalBackButtonItem = nil;
    if(self.navigationController.visibleViewController == self)
      [self.navigationController setToolbarHidden:NO animated:YES];

    [tableView beginUpdates];
    [tableView deleteRowsAtIndexPaths:rowsOnlyInEdition withRowAnimation:UITableViewRowAnimationTop];
    [tableView insertSections:[NSIndexSet indexSetWithIndex:2] withRowAnimation:UITableViewRowAnimationAutomatic];
    [tableView endUpdates];
    if(iOS6())
    {
      UITableViewCell *cell = [self.tableView cellForRowAtIndexPath:rowBeforeOnlyInEdition];
      [tableView addBackgroundsToGroupedCell:cell atIndexPath:rowBeforeOnlyInEdition];
    }
    
    Entry *newEntry = nil;

    if(m_newTitle)
    {
      if(!newEntry) 
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      newEntry.title = m_newTitle;
      m_newTitle = nil;
    }
    if(m_newPassword)
    {
      if(!newEntry) 
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      newEntry.password = m_newPassword;
      m_newPassword = nil;
    }
    if(m_newUser)
    {
      if(!newEntry) 
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      newEntry.username = m_newUser;
      m_newUser = nil;
    }
    if(m_newURL)
    {
      if(!newEntry) 
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      newEntry.url = m_newURL;
      m_newURL = nil;
    }
    if(m_newEmail)
    {
      if(!newEntry) 
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      newEntry.email = m_newEmail;
      m_newEmail = nil;
    }
    if(m_newNotes)
    {
      if(!newEntry) 
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      newEntry.notes = m_newNotes;
      m_newNotes = nil;
    }
    if(m_newGroup)
    {
      if(!newEntry) 
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      newEntry.group = m_newGroup;
      m_newGroup = nil;
    }
    if(m_editPwPolicyVC)
    {
      if(!newEntry)
        newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
      [m_editPwPolicyVC updateEntry:newEntry];
      m_editPwPolicyVC = nil;
    }
    if(m_historyVC)
    {
      BOOL enabled = m_historyVC.historyEnabled;
      uint max = m_historyVC.maxEntries;
      BOOL clear = m_historyVC.shouldClearHistory;
      
      if(newEntry || clear || !!enabled != m_entry.pwHistoryKeep || max != m_entry.pwHistoryMaxEntries)
      {
        if(!newEntry)
          newEntry = [[Entry alloc]initWithEntry:m_entry file:m_file];
        newEntry.pwHistoryKeep = enabled;
        newEntry.pwHistoryMaxEntries = max;
        if(clear)
          [newEntry clearPwHistory];
      }
    }

    BOOL modified = newEntry != nil;
    A77Log(@"Entry editing finished. Has modifications: %d", (int)modified);
    
    if(modified)
    {
      [m_file editEntry:m_entry newEntry:newEntry];
      m_entry = newEntry;
      m_historyVC.entry = newEntry;
      [self btnCopyAndGo].enabled = m_entry.parsedURL != nil;
    }
  }
  
  UITableViewCell *notesCell = [self.tableView cellForRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:1]];
  if(notesCell)
    [self setupNotesCell:notesCell];
  for(UITableViewCell *cell in tableView.visibleCells)
  {
    NSIndexPath *indexPath = [tableView indexPathForCell:cell];
    [self configureAccessibilityForCell:cell atIndexPath:indexPath];
  }
}

- (void)hideHud:(id)iHud
{
  MBProgressHUD *hud = (MBProgressHUD *)iHud;
  [hud hide:YES];
}

- (BOOL)historyVC:(PwHistoryVC *)historyVC willSetEditing:(BOOL)editing
{
  NSAssert(m_historyVC == historyVC, @"Must be the same");
  if([self testCanEndEditingAndNotify:NO animated:NO])
    return YES;
  
  [self.navigationController popViewControllerAnimated:YES];
  [self testCanEndEditingAndNotify:YES animated:NO];
  return NO;
}

- (void)historyVC:(PwHistoryVC *)historyVC setEditing:(BOOL)editing
{
  [self setEditing:editing animated:NO];
  NSAssert(m_historyVC == historyVC, @"Must be the same");
}

- (void)displayHistoryAnimated:(BOOL)animated
{
  if(!m_historyVC)
  {
    m_historyVC = [[PwHistoryVC alloc]initWithEntry:m_entry edition:self.isEditing];
    m_historyVC.delegate = self;
  }
  [self.view endEditing:YES];
  [self.navigationController pushViewController:m_historyVC animated:animated];
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
  const NSInteger row = indexPath.row;
  const NSInteger section = indexPath.section;
  
  if(section == 2)
  {
    [tableView deselectRowAtIndexPath:indexPath animated:YES];
    return;
  }
  
  BOOL editing = self.editing;
  
  if(row == HISTORY_ROW_IDX)
  {
    [self displayHistoryAnimated:YES];
    return;
  }
  
  if (editing)
  {
    if(section != 0 || (row != GROUP_ROW_IDX && row != POLICY_ROW_IDX ))
    {
      UITableViewCell *cell = [tableView cellForRowAtIndexPath:indexPath];
      [tableView deselectRowAtIndexPath:indexPath animated:NO];
      if([cell isKindOfClass:[SimpleDetailTableViewCell class]])
      {
        SimpleDetailTableViewCell *detailCell = (SimpleDetailTableViewCell *)cell;
        [detailCell.detailTextField becomeFirstResponder];
      }
      else
      {
        NSAssert(section == 1 && row == 0, @"Must be the notes field");
        UITextView *txtView = (UITextView *)[cell.contentView viewWithTag:TEXTVIEWTAG];
        [txtView becomeFirstResponder];
      }
      return;
    }
    if(row == GROUP_ROW_IDX)
    {
      m_selectGroupViewController = [[SelectGroupViewController alloc]initWithSafeFile:m_file currentGroup:(m_newGroup ? m_newGroup : m_entry.group)];
      m_selectGroupViewController.allowCreatingNewGroup = UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPhone;
      [self.view endEditing:YES];
      [self.navigationController pushViewController:m_selectGroupViewController animated:YES];
    }
    else
    {
      NSAssert(row == POLICY_ROW_IDX, @"Don't know about other rows");
      if(!m_editPwPolicyVC)
      {
        m_editPwPolicyVC = [[EditEntryPolicyVC alloc]initWithEntry:m_entry handler:^(EditEntryPolicyVC *editVC) {
          [self.navigationController popViewControllerAnimated:YES];
        }];
      }
      [self.view endEditing:YES];
      [self.navigationController pushViewController:m_editPwPolicyVC animated:YES];
    }
    return;
  }
  
  NSString *title, *value;
  BOOL copied = NO;  
  BOOL isPassword = section == 0 && row == PASSWORD_ROW_IDX;
  if(isPassword)
  {
    Entry *baseEntry = [m_file baseEntryForEntry:m_entry];
    if(baseEntry)
    {
      NSString *password = baseEntry.password;
      if(password && password.length)
      {
        [UIPasteboard generalPasteboard].string = password;
        [PasteboardPasswordCleaner passwordCopied:password];
      }
      copied = YES;
    }
  }
  
  [self getContentsForIndexPath:indexPath text_p:&value title_p:&title];

  if(!copied)
  {
    if(value && value.length)
    {
      [UIPasteboard generalPasteboard].string = value;
      if(isPassword)
        [PasteboardPasswordCleaner passwordCopied:value];
    }
  }
  
  value = [UIPasteboard generalPasteboard].string;
  [self showHudWithMessage:[NSString stringWithFormat:@"%@ %@", title, value && value.length ? @"copied" : @"empty"]];
  [tableView deselectRowAtIndexPath:indexPath animated:YES];
} 

- (MBProgressHUD *)showHudWithMessage:(NSString *)message
{
  UITableView *tableView = self.tableView;
  MBProgressHUD *hud = [[MBProgressHUD alloc]initWithWindow:tableView.window];
  hud.customView = [[UIImageView alloc]initWithImage:[UIImage imageNamed:@"Check.png"]];
  hud.mode = MBProgressHUDModeCustomView;
  hud.labelText = message;
  hud.animationType = MBProgressHUDAnimationZoom;
  hud.opacity = .8;
  hud.removeFromSuperViewOnHide = YES;
  [tableView.window addSubview:hud];
  [hud show:YES];
  [self performSelector:@selector(hideHud:) withObject:hud afterDelay:UIAccessibilityIsVoiceOverRunning() ? 3 : 1];
  return hud;
}

@end
