//
//  GroupViewController.m
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/8/11.
//  Copyright 2011 app77. All rights reserved.
//

#import "GroupViewController.h"
#import "Entry.h"
#import "SafeFile.h"
#import "MBProgressHUD.h"
#import "EntryDetailsViewController.h"
#import "AddHelpTextUITableViewController.h"
#import "PwHistoryVC.h"
#import "SafeCtx.h"
#import "AppTheme.h"
#import "EntryCell.h"
#import "OpenInChromeController.h"
#import "PasteboardGravitationalManager.h"
#import "PasteboardPasswordCleaner.h"
#import "UIActionSheet+BlockActionSheet.h"

typedef enum SingleTapAction
{
  SingleTapActionCopyPassword,
  SingleTapActionCopyUsername,
  SingleTapActionOpenEntry,
  SingleTapActionOpenURL,
} SingleTapAction;

@interface GroupViewController () <UISearchDisplayDelegate, UISearchBarDelegate, EntryCellResponder>
{
  SafeFile *m_file;
  SafeCtx *m_safeCtx;
  NSString *m_groupName;
  NSDictionary *m_entries;
  UIView *m_disabledImage;
  NSArray *m_sections;
  NSMutableDictionary *m_state;
  NSDictionary *m_filteredEntries;
  NSArray *m_filteredSections;
  UISearchDisplayController *m_searchDisplayController;
  EntryDetailsViewController *m_currentlyDisplayedDetails;
  NSData *m_currentlyDisplayedEntryUUID;
  UITableView *m_tableView;
  
  NSString *m_savedSearchTerm;
  NSNumber *m_savedSearchButton;
  NSNumber *m_savedScrollOffset;
  
  BOOL m_rowBeingEdited;
  
  BOOL m_startInSearchMode;
  NSNumber *m_singleTapAction;
}

@end

@implementation GroupViewController

static NSString * const kSingleTapActionKey = @"singleTapAction";

static NSString * const kScrollPosition = @"GroupViewControllerScrollPosition";
static NSString * const kSearchTerm = @"GroupSearchTerm";
static NSString * const kSearchButton = @"GroupSearchMode";
static NSString * const kSmartPasteboardEnabledKey = @"smartPasteboardEnabled";

static NSString * const kHelpText = @"Tap here to add a\nnew item";
static NSInteger const kHelpBaseTag = 1200;

- (SingleTapAction)singleTapAction
{
  if(m_singleTapAction)
    return (SingleTapAction)m_singleTapAction.intValue;
    
  NSString *actionStr = [[NSUserDefaults standardUserDefaults]valueForKey:kSingleTapActionKey];
  if(actionStr.length == 0)
    m_singleTapAction = @(SingleTapActionCopyPassword);
  
  if([actionStr isEqualToString:@"copyPassword"])
    m_singleTapAction = @(SingleTapActionCopyPassword);
  else  if([actionStr isEqualToString:@"copyUsername"])
    m_singleTapAction = @(SingleTapActionCopyUsername);
  else  if([actionStr isEqualToString:@"openEntry"])
    m_singleTapAction = @(SingleTapActionOpenEntry);
  else  if([actionStr isEqualToString:@"openURL"])
    m_singleTapAction = @(SingleTapActionOpenURL);
  else
    m_singleTapAction = @(SingleTapActionCopyPassword);
  
  return (SingleTapAction)m_singleTapAction.intValue;
}

static BOOL smartPasteboardEnabled()
{
  NSNumber *value = [[NSUserDefaults standardUserDefaults]objectForKey:kSmartPasteboardEnabledKey];
  return value ? value.boolValue : YES;
}

- (void) getEffDataForTable: (UITableView *)tableView  sections:(NSArray **)effSections entries:(NSDictionary **)effEntries
{
  if(tableView == self.tableView)
  {
    *effSections = m_sections;
    *effEntries = m_entries;
  }
  else
  {
    *effSections = m_filteredSections;
    *effEntries = m_filteredEntries;
  }
}

- (Entry *) entryInTableView: (UITableView *)tableView atIndexPath: (NSIndexPath *)indexPath  
{
  NSArray *effSections;
  NSDictionary *effEntries;
  [self getEffDataForTable:tableView sections:&effSections entries:&effEntries];
  
  NSNumber *section = effSections[indexPath.section];
  Entry *entry = effEntries[section][indexPath.row];
  return entry;
}

- (void)loadSavedState
{
  NSString *searchTerm = m_state[kSearchTerm];
  BOOL searching = searchTerm && ![searchTerm isEqualToString:@""];

  // Save those to use later
  m_savedSearchButton = m_state[kSearchButton];
  m_savedSearchTerm = searchTerm;

  if(!searching)
    m_savedScrollOffset = m_state[kScrollPosition];
  else
    m_savedScrollOffset = nil;
}

- (void) prepareEntriesListFiltered:(BOOL) filter
{
  NSArray *unsorted;
  if(!m_groupName)
  {
    // Union all arrays
    NSArray *allArrays = [m_file.entries allValues];
    if(allArrays.count > 0)
    {
      unsorted = allArrays[0];
      for(int i = 1; i < allArrays.count; i++)
        unsorted = [unsorted arrayByAddingObjectsFromArray:allArrays[i]];
    }
    else
    {
      unsorted = @[];
    }
      
  }
  else
    unsorted = (m_file.entries)[m_groupName];
  
  if(filter)
  {
    NSMutableArray *filtered = [[NSMutableArray alloc]initWithCapacity:unsorted.count];
    NSString *searchText = self.searchDisplayController.searchBar.text;
    BOOL titleOnly = self.searchDisplayController.searchBar.selectedScopeButtonIndex == 0;
    
    @autoreleasepool {
      for(Entry *entry in unsorted)
        if([entry matchesSearchTerm:searchText titleOnly:titleOnly])
          [filtered addObject:entry];
    }
    unsorted = filtered;
  }
  
  NSArray *sorted = [unsorted sortedArrayUsingSelector:@selector(caseInsensitiveCompare:)];
  NSMutableDictionary *newEntries = [[NSMutableDictionary alloc]init];
  NSMutableArray *letterArray = nil;
  NSMutableArray *newSections = [[NSMutableArray alloc]init];
  
  unichar lastChar = 0;
  
  for(Entry *entry in sorted)
  {
    NSString *title = entry.title;
    unichar startChar = title.length > 0 ? [[title uppercaseString] characterAtIndex:0] : ' ';
    if(startChar == lastChar)
    {
      NSAssert2(letterArray != nil, @"Internal error @ %s:%d", __FILE__, __LINE__);
      NSAssert2(letterArray == [newEntries objectForKey:[NSNumber numberWithUnsignedShort:startChar]], @"Internal error @ %s:%d", __FILE__, __LINE__);
    }
    else 
    {
      letterArray = [[NSMutableArray alloc]init];
      NSNumber *key = @(startChar);
      newEntries[key] = letterArray;
      [newSections addObject:key];
    }

    lastChar = startChar;
    [letterArray addObject:entry];
  }

  if(!filter)
  {
    m_entries = newEntries;
    m_sections = newSections;

    // This method gets called in every insertion / deletion
    BOOL empty = m_sections.count == 0;
    self.navigationItem.rightBarButtonItem.enabled = !empty;
    [self adjustAddHelpText:kHelpText viewTagBase:kHelpBaseTag show:(empty && m_file != nil)];
  }
  else
  {
    m_filteredEntries = newEntries;
    m_filteredSections = newSections;
  }
}

#pragma mark -
#pragma mark Initialization

- (id) initWithState:(NSMutableDictionary *)state
{
  self = [super initWithStyle:UITableViewStylePlain];
  if(!self)
    return nil;

  m_state = state;
  
  m_rowBeingEdited = NO;

  // navigation bar
  self.navigationItem.rightBarButtonItem = self.editButtonItem;
  
  UIBarButtonItem *spacer = [[UIBarButtonItem alloc]initWithBarButtonSystemItem:UIBarButtonSystemItemFlexibleSpace target:nil action:nil];
  UIBarButtonItem *addBtn = [UIBarButtonItem buttonWithImageNamed:imageName(@"Plus") target:self action:@selector(onAddNewItem:)];
  addBtn.accessibilityLabel = @"Add";

  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
  {
    self.toolbarItems = @[spacer, addBtn];
  }
  else
  {
    UIBarButtonItem *closeBtn = [UIBarButtonItem buttonWithImageNamed:imageName(@"Close") target:self action:@selector(onCloseSafe:)];
    closeBtn.accessibilityLabel = @"Close";
    self.toolbarItems = @[closeBtn, spacer, addBtn];
  }  
  
  return self;
}

- (id) initWithFile:(SafeFile *)filePtr ctx:(SafeCtx *)ctx  group:(NSString *)group searchMode:(BOOL)searchMode state:(NSMutableDictionary *)state
{
  self = [self initWithState:state];
  if(!self)
    return nil;
  

  [self setSafeFile:filePtr ctx:ctx group:group searchMode:searchMode];
  
  return self;
}

- (void)dealloc
{
  m_searchDisplayController.delegate = nil;
  m_searchDisplayController.searchResultsDataSource = nil;
  m_searchDisplayController.searchResultsDelegate = nil;
  if(self.isViewLoaded)
  {
    self.tableView.delegate = nil;
    self.tableView.dataSource = nil;
  }
}

- (void) setSafeFile:(SafeFile *)filePtr ctx:(SafeCtx *)ctx group:(NSString *)group searchMode:(BOOL)searchMode
{
  if(!group)
    self.navigationItem.title = @"All items";
  else if([group isEqualToString:@""])
    self.navigationItem.title = @"Ungrouped";
  else
  {
    NSArray *groups = [group componentsSeparatedByString:@"."];
    self.navigationItem.title = groups.lastObject;
  }
  
  m_file = filePtr;
  m_safeCtx = ctx;
  
  m_groupName = group ? [group copy] : nil;
  m_startInSearchMode = searchMode;
  m_savedSearchTerm = nil;
  m_savedSearchButton = nil;
  
  self.editButtonItem.enabled = filePtr != nil;
  [[self.toolbarItems lastObject] setEnabled:filePtr != nil];
  UISearchBar *searchBar = m_searchDisplayController.searchBar;
  if(filePtr == nil)
  {
    searchBar.text = nil;
    [searchBar setUserInteractionEnabled:NO];
  }
  else
  {
    [searchBar setUserInteractionEnabled:YES]; 
  }
  
  [m_state removeObjectForKey:kScrollPosition];
  [m_state removeObjectForKey:kSearchButton];
  [m_state removeObjectForKey:kSearchTerm];

  /* To simulate changes from iCloud
#ifdef DEBUG_MODE
  if(filePtr)
  {
    double delayInSeconds = 5.0;
    dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, delayInSeconds * NSEC_PER_SEC);
    dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
      [m_file safeChangedExternally];
    });
  }
#endif */
}

- (void)reload
{
  if(!self.isViewLoaded)
    return;
  if(self.editing)
    [self setEditing:NO animated:YES];
  [self prepareEntriesListFiltered:NO];
  [self.tableView reloadData];
  UISearchDisplayController *searchController = self.searchDisplayController;
  UISearchBar *searchBar = searchController.searchBar;

  if(m_startInSearchMode || (m_savedSearchTerm && m_savedSearchTerm.length))
  {
    if(m_searchDisplayController.isActive)
    {
      [m_searchDisplayController setActive:NO animated:NO];
      [m_searchDisplayController setActive:YES animated:NO];
    }
    else
      [m_searchDisplayController setActive:YES animated:YES];
    
    searchBar.selectedScopeButtonIndex = m_savedSearchButton ? [m_savedSearchButton integerValue] : 0;
    [searchBar becomeFirstResponder];
    [self prepareEntriesListFiltered:YES];
    [self.searchDisplayController.searchResultsTableView reloadData];
    searchBar.text = m_savedSearchTerm;
  }
  else
  {
    [searchController setActive:NO animated:YES];
    searchBar.text = @"";
  }
  
  if(m_savedScrollOffset)
  {
    CGFloat offset = [m_savedScrollOffset floatValue];
    [self.tableView setContentOffset:CGPointMake(0, offset) animated:NO];
    m_savedScrollOffset = nil;
  }
  else
  {
    if(m_sections.count > 0)
    {
      NSAssert([self tableView:m_tableView numberOfRowsInSection:0], @"No empty sections allowed");
      [m_tableView scrollToRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:0] atScrollPosition:UITableViewScrollPositionNone animated:NO];
    }
  }
  
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
  {
    [UIView animateWithDuration:.3 animations:^{
      [self adjustVisibilities];
    }];
  }
}

- (void)clear
{
  if(self.editing)
    [self setEditing:NO animated:YES];

  self.editButtonItem.enabled = NO;
  [[self.toolbarItems lastObject] setEnabled:NO];  
  
  m_groupName = nil;
  m_file = nil;
  
  m_entries = nil;
  m_sections = nil;
  m_filteredEntries = nil;
  m_filteredSections = nil;

  self.navigationItem.title = nil;
  [self.tableView reloadData];
  [m_searchDisplayController setActive:NO animated:YES];
  UISearchBar *searchBar = m_searchDisplayController.searchBar;
  searchBar.text = nil;
  [searchBar setUserInteractionEnabled:NO];
  
  [self adjustAddHelpText:kHelpText viewTagBase:kHelpBaseTag show:NO];
  
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
  {
    [UIView animateWithDuration:.3 animations:^{
      [self adjustVisibilities];
    }];
  }
}

- (void)mustCloseEdition
{
  if(self.editing)
    [self setEditing:NO animated:YES]; 
  if(m_currentlyDisplayedDetails && m_currentlyDisplayedDetails.isEditing)
    [m_currentlyDisplayedDetails onCancelEdit:nil];
}

- (void)safeFileWasExternallyModified
{
  [self mustCloseEdition];
  [self loadSavedState];
  [self reload];
  if(!m_currentlyDisplayedDetails)
    return;

  UINavigationController *nav = self.navigationController;
  
  BOOL inHistory = [nav.visibleViewController.class isSubclassOfClass:[PwHistoryVC class]];
  m_currentlyDisplayedDetails = nil;
  Entry *newEntry = nil;
  
  for(NSArray *entries in m_entries.allValues)
  {
    for(Entry *entry in entries)
    {
      if([entry.uuid isEqualToData:m_currentlyDisplayedEntryUUID])
      {
        newEntry = entry;
        break;
      }
    }
  }
  
  if(!newEntry)
  {
    [self.navigationController popToViewController:self animated:YES];
    m_currentlyDisplayedEntryUUID = nil;
    m_rowBeingEdited = NO;
    return;
  }
  
  m_currentlyDisplayedDetails = [[EntryDetailsViewController alloc]initWithFile:m_file ctx:m_safeCtx entry:newEntry];
  [self.navigationController popToViewController:self animated:NO];
  NSAssert(m_currentlyDisplayedEntryUUID, @"Must not call viewWillAppear, which nulifies m_currentlyDisplayedEntryUUID");
  [self.navigationController pushViewController:m_currentlyDisplayedDetails animated:NO];
  if(inHistory)
    [m_currentlyDisplayedDetails displayHistoryAnimated:NO];
}

- (void)safeFileWasInternallyModified
{
  [m_currentlyDisplayedDetails safeWasInternallyModified];
}

- (void)groupRenamedFrom:(NSString *)oldGroupName to:(NSString *)newGroupName
{
  if([oldGroupName isEqualToString:m_groupName])
  {
    m_groupName = newGroupName;
    NSArray *groups = [newGroupName componentsSeparatedByString:@"."];
    self.navigationItem.title = groups.lastObject;
  }
  [m_currentlyDisplayedDetails groupRenamedFrom:oldGroupName to:newGroupName];
}

#pragma mark -
#pragma mark View lifecycle

- (void)adjustVisibilities
{
  if(m_file)
  {
    m_disabledImage.alpha = 0;
    m_tableView.alpha = 1;
#ifdef PWSAFE_COZI
    [self.navigationController setToolbarHidden:NO animated:YES];
#endif
  }
  else
  {
    m_disabledImage.alpha = 1;
    m_tableView.alpha = 0;
#ifdef PWSAFE_COZI
    [self.navigationController setToolbarHidden:YES animated:YES];
#endif
  }
}

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
  // Return YES for supported orientations
  return (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad || interfaceOrientation == UIInterfaceOrientationPortrait);
}

- (void)willAnimateRotationToInterfaceOrientation:(UIInterfaceOrientation)toInterfaceOrientation duration:(NSTimeInterval)duration
{
  [super willAnimateRotationToInterfaceOrientation:toInterfaceOrientation duration:duration];
  if(m_disabledImage && m_disabledImage.alpha > 0)
  {
    UIView *old = m_disabledImage;
    m_disabledImage = themeGroupViewDisabledImage(self.view.frame, UIInterfaceOrientationIsLandscape(toInterfaceOrientation));
    m_disabledImage.alpha = old.alpha;
    [self.view insertSubview:m_disabledImage aboveSubview:old];
    [old removeFromSuperview];
  }
}

- (void)willRotateToInterfaceOrientation:(UIInterfaceOrientation)toInterfaceOrientation duration:(NSTimeInterval)duration
{
  [super willRotateToInterfaceOrientation:toInterfaceOrientation duration:duration];
  if(m_disabledImage && m_disabledImage.alpha == 0)
  {
    UIView *old = m_disabledImage;
    m_disabledImage = themeGroupViewDisabledImage(self.view.frame, UIInterfaceOrientationIsLandscape(toInterfaceOrientation));
    m_disabledImage.alpha = old.alpha;
    [self.view insertSubview:m_disabledImage aboveSubview:old];
    [old removeFromSuperview];
  }  
}

- (void)didRotateFromInterfaceOrientation:(UIInterfaceOrientation)fromInterfaceOrientation
{
  [super didRotateFromInterfaceOrientation:fromInterfaceOrientation];
  [self adjustHelpFramesViewTagBase:kHelpBaseTag];
}

- (void)viewDidLoad
{
  [super viewDidLoad];
  [self.tableView themeViewDidLoadInIpadPos:TVP_RIGHT];
  
  [self prepareEntriesListFiltered:NO];
  
  BOOL searching = m_savedSearchTerm && ![m_savedSearchTerm isEqualToString:@""];
  UISearchBar *searchBar = self.searchDisplayController.searchBar;
  
  if(searching || m_startInSearchMode)
  {
    if(searching)
    {
      searchBar.text = m_savedSearchTerm;
      if(m_savedSearchButton)
        searchBar.selectedScopeButtonIndex = [m_savedSearchButton integerValue];
      [self prepareEntriesListFiltered:YES];
    }
    else
    {
      searchBar.selectedScopeButtonIndex = 0;
    }
  }
  
  m_savedSearchTerm = nil;
  m_savedSearchButton = nil;
  
  UIBarButtonItem *backButton = [[UIBarButtonItem alloc] initWithTitle:@"Entries" style:UIBarButtonItemStyleBordered target:nil action:nil];
  self.navigationItem.backBarButtonItem = backButton;
}

- (void)viewDidUnload 
{
  [super viewDidUnload];
  
  m_entries = nil;
  m_sections = nil;
  
  m_filteredSections = nil;
  m_filteredEntries = nil;
  
  m_searchDisplayController = nil;

  m_disabledImage = nil;
  
  m_tableView = nil;
}

- (void)viewWillAppear:(BOOL)animated 
{
  [super viewWillAppear:animated];
  
  m_currentlyDisplayedDetails = nil;
  m_currentlyDisplayedEntryUUID = nil;
  
  [self.navigationController setToolbarHidden:NO animated:animated];

  if(m_savedScrollOffset)
  {
    CGFloat offset = [m_savedScrollOffset floatValue];
    [self.tableView setContentOffset:CGPointMake(0, offset) animated:NO];
    m_savedScrollOffset = nil;
  } 
  
  if(m_rowBeingEdited)
  {
    // Reload internal data structures
    [self prepareEntriesListFiltered:NO];
    BOOL searching = self.searchDisplayController.searchBar.text != nil && self.searchDisplayController.searchBar.text.length != 0;
    if(searching)
      [self prepareEntriesListFiltered:YES];

    // Reload the table
    [self.tableView reloadData];
    
    if(searching)
      [self.searchDisplayController.searchResultsTableView reloadData];
    
    m_rowBeingEdited = NO;
  }

#ifdef PWSAFE_COZI
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad && m_file == nil)
    [self.navigationController setToolbarHidden:YES animated:animated];
#endif

}

- (void)viewDidAppear:(BOOL)animated
{
  [super viewDidAppear:animated];
  NSString *searchBarText = m_searchDisplayController.searchBar.text;
  if(m_startInSearchMode || (searchBarText && searchBarText.length))
  {
    [m_searchDisplayController setActive:YES animated:YES];
    [m_searchDisplayController.searchBar becomeFirstResponder];
    m_startInSearchMode = NO;
  }
}

- (void)viewWillDisappear:(BOOL)animated
{
  [m_state removeObjectForKey:kScrollPosition];
  [m_state removeObjectForKey:kSearchButton];
  [m_state removeObjectForKey:kSearchTerm];
  m_singleTapAction = nil;
  
  [super viewWillDisappear:animated];
  [self adjustAddHelpText:kHelpText viewTagBase:kHelpBaseTag show:NO];
}

- (UITableView *)tableView
{
  if(!m_tableView)
  {
    UIView *view = self.view;
    (void)view;
  }
  NSAssert(m_tableView, @"Must be loaded by now");
  return m_tableView;
}

- (void)loadView
{
  CGRect frame = CGRectMake(0, 0, 500, 500);
  self.view = [[UIView alloc]initWithFrame:frame];
  self.view.backgroundColor = kThemeDarkBgColor;
  
  UITableView *tableView = [[UITableView alloc]initWithFrame:frame style:UITableViewStylePlain];
  tableView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
  [self.view addSubview:tableView];
  m_tableView = tableView;
  tableView.delegate = self;
  tableView.dataSource = self;
  
  
  
  m_tableView.sectionIndexMinimumDisplayRowCount = 10;
  
  // Searchbar
  float w = self.navigationController.view.bounds.size.width;
  CGRect searchBarFrame = CGRectMake(0, 0, w, 44);
  UISearchBar *searchBar = [[UISearchBar alloc] initWithFrame:searchBarFrame];
  searchBar.scopeButtonTitles = @[@"Title only", @"All fields"];
  searchBar.delegate = self;
  m_tableView.tableHeaderView = searchBar;
  m_searchDisplayController = [[UISearchDisplayController alloc]initWithSearchBar:searchBar contentsController:self];
  m_searchDisplayController.delegate = self;
  m_searchDisplayController.searchResultsDataSource = self;
  m_searchDisplayController.searchResultsDelegate = self;
  
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
  {
    m_disabledImage = themeGroupViewDisabledImage(frame, UIInterfaceOrientationIsLandscape(self.navigationController.interfaceOrientation));
    [self.view addSubview:m_disabledImage];
    [self adjustVisibilities];
  }
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView 
{
  NSArray *effSections;
  NSDictionary *effEntries;
  [self getEffDataForTable:tableView sections:&effSections entries:&effEntries];  
  
  // Return the number of sections.
  return effSections.count;
}


- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)sectionIdx 
{
  NSArray *effSections;
  NSDictionary *effEntries;
  [self getEffDataForTable:tableView sections:&effSections entries:&effEntries];    
  // Return the number of rows in the section.
  NSNumber *section = effSections[sectionIdx];
  NSArray *entriesInSection = effEntries[section];
  return entriesInSection.count;
}

// Customize the appearance of table view cells.
- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath 
{
  static NSString *CellIdentifier = @"Cell";
  
  EntryCell *cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
  if (cell == nil) 
  {
    cell = [[EntryCell alloc] initWithStyle:UITableViewCellStyleSubtitle reuseIdentifier:CellIdentifier];
#ifdef THEME_USE_CUSTOM_TABLE_VIEW_CELL_SELECTION_BG
    cell.selectedBackgroundView = [tableView cellSelectionBackgroundAtIndexPath:indexPath];
    cell.textLabel.highlightedTextColor = themeTableViewSelectedTextColor;
    cell.detailTextLabel.highlightedTextColor = themeTableViewSelectedTextColor;    
#endif
    cell.accessibilityTraits = UIAccessibilityTraitButton;
    NSString *tapHint;
    SingleTapAction action = [self singleTapAction];
    switch(action)
    {
      case SingleTapActionCopyPassword:
        tapHint = @"Copies password.";
        break;
      case SingleTapActionCopyUsername:
        tapHint = @"Copies username.";
        break;
      case SingleTapActionOpenEntry:
        tapHint = @"Opens entry.";
        break;
      case SingleTapActionOpenURL:
        tapHint = @"Opens URL.";
        break;
    }
    cell.accessibilityHint = tapHint;
    
    if(!iOS6() && [self singleTapAction] != SingleTapActionOpenEntry)
    {
      UIButton *btn = [UIButton buttonWithType:UIButtonTypeCustom];
      UIImage *btnImg = [[UIImage imageNamed:@"infoBtn"]imageWithRenderingMode:UIImageRenderingModeAlwaysTemplate];
      btn.frame = CGRectMake(0, 0, 44, 44);
      [btn setImage:btnImg forState:UIControlStateNormal];
      CGSize imgSize = btnImg.size;
      CGFloat scale = [UIScreen mainScreen].scale;
      btn.imageEdgeInsets = UIEdgeInsetsMake(0, rintf(scale * ((44 - imgSize.width) / 4)) / scale, 0, 0);
      [btn addTarget:self action:@selector(onInfoBtn:) forControlEvents:UIControlEventTouchUpInside];
      cell.accessoryView = btn;
      btn.accessibilityLabel = @"Details";
      btn.accessibilityHint = @"Opens entry details.";
    }
  }
  
  // Configure the cell...
  
  NSArray *effSections;
  NSDictionary *effEntries;
  [self getEffDataForTable:tableView sections:&effSections entries:&effEntries];  
  NSNumber *section = effSections[indexPath.section];
  Entry *entry = effEntries[section][indexPath.row];

  cell.textLabel.text = entry.title;
  
  NSString *url = entry.url;
  NSString *username = entry.username;
  
  NSNumber *showDetailsObj = [[NSUserDefaults standardUserDefaults]objectForKey:@"showDetailsInListing"];
  BOOL showDetails = !showDetailsObj || showDetailsObj.boolValue;
  
  if(showDetails)
  {
    BOOL hasURL = url != nil && ![url isEqualToString:@""];
    BOOL urlHasMethod = url != nil && [url rangeOfString:@"://"].location != NSNotFound;
    BOOL hasUsername = username != nil && ![username isEqualToString:@""];
    BOOL usernameIsEmail = username != nil && [username rangeOfString:@"@"].location != NSNotFound;
    
    if(hasUsername)
    {
      if(hasURL && !usernameIsEmail && !urlHasMethod)
        cell.detailTextLabel.text = [NSString stringWithFormat:@"%@@%@", username, url];
      else
        cell.detailTextLabel.text = username;
    }
    else 
    {
      cell.detailTextLabel.text = url;
    }
  }
  else
    cell.detailTextLabel.text = nil;
  
  if([self singleTapAction] == SingleTapActionOpenEntry)
    cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
  else
  {
    if(iOS6())
    {
      cell.accessoryType = UITableViewCellAccessoryDetailDisclosureButton;
    }

  }
  
  [tableView addThemeToPlainCell:cell];
  
  return cell;
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)sectionIdx
{
  NSArray *effSections;
  NSDictionary *effEntries;
  [self getEffDataForTable:tableView sections:&effSections entries:&effEntries];
  
  NSNumber *section = effSections[sectionIdx];
  unichar sectionLetter = [section unsignedShortValue];

  return [NSString stringWithCharacters:&sectionLetter length:1];
}

- (NSArray *)sectionIndexTitlesForTableView:(UITableView *)tableView
{
  if(tableView != self.tableView)
    return nil;
  
  NSMutableArray *indexes = [NSMutableArray arrayWithCapacity:m_sections.count + 1];
  [indexes addObject:UITableViewIndexSearch];
  for(NSNumber *ch in m_sections)
  {
    unichar c = [ch unsignedCharValue];
    [indexes addObject:[NSString stringWithCharacters:&c length:1]];
  }
  return indexes;
}

- (NSInteger)tableView:(UITableView *)tableView sectionForSectionIndexTitle:(NSString *)title atIndex:(NSInteger)index
{
  if (tableView != self.tableView)
    return NSNotFound;
  
  if (index == 0) 
  {
    [tableView setContentOffset:CGPointMake(0, -tableView.contentInset.top) animated:NO];
    return NSNotFound;
  }
  return index - 1;
}

// Override to support editing the table view.
- (void)tableView:(UITableView *)tableView commitEditingStyle:(UITableViewCellEditingStyle)editingStyle forRowAtIndexPath:(NSIndexPath *)indexPath 
{
  if (editingStyle == UITableViewCellEditingStyleDelete) 
  {
    // Delete the row from the data source.
    bool lastOneInSection = [self tableView:tableView numberOfRowsInSection:indexPath.section] == 1;
    Entry *entry = [self entryInTableView:tableView atIndexPath:indexPath];
    [m_file deleteEntry:entry];
    [self prepareEntriesListFiltered:NO];
    if(tableView != self.tableView)
      [self prepareEntriesListFiltered:YES];
    
    if(lastOneInSection)
      [tableView deleteSections:[NSIndexSet indexSetWithIndex:indexPath.section] withRowAnimation:UITableViewRowAnimationLeft];
    else
      [tableView deleteRowsAtIndexPaths:@[indexPath] withRowAnimation:UITableViewRowAnimationLeft];
  }
  else if (editingStyle == UITableViewCellEditingStyleInsert) 
  {
    // Create a new instance of the appropriate class, insert it into the array, and add a new row to the table view.
  }   
}

#pragma mark - Table view delegate

- (void)scrollViewDidScroll:(UIScrollView *)scrollView
{
  if(scrollView == (UIScrollView *)self.tableView)
  {
    CGFloat offset = scrollView.contentOffset.y;
    m_state[kScrollPosition] = @(offset);
  }
}

- (void)hideHud:(id)iHud
{
  MBProgressHUD *hud = (MBProgressHUD *)iHud;
  [hud hide:YES];
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
  switch([self singleTapAction])
  {
    case SingleTapActionCopyPassword:
    default:
      [self copyPasswordAtIndexPath:indexPath tableView:tableView];
      break;
      
    case SingleTapActionCopyUsername:
      [self tableView:tableView performAction:@selector(copyUsername:) forRowAtIndexPath:indexPath withSender:nil];
      break;
      
    case SingleTapActionOpenURL:
      [self tableView:tableView performAction:@selector(openUrl:) forRowAtIndexPath:indexPath withSender:nil];
      break;
      
    case SingleTapActionOpenEntry:
      [self tableView:tableView openEntryAtIndexPath:indexPath];
      break;
  }
  
  [tableView deselectRowAtIndexPath:indexPath animated:YES];
} 

-(void)tableView:(UITableView *)tableView accessoryButtonTappedForRowWithIndexPath:(NSIndexPath *)indexPath
{
  [self tableView:tableView openEntryAtIndexPath:indexPath];
}

- (void)onInfoBtn:(UIButton *)sender
{
  UITableView *tableView = (UITableView *)sender.nextResponder;
  while(![tableView isKindOfClass:[UITableView class]])
    tableView = (UITableView *)tableView.nextResponder;
  NSAssert(tableView, @"Must find the tableView");
  if(!tableView)
    return;
  
  UITableViewCell *cell = (UITableViewCell *)sender.superview;
  while(![cell isKindOfClass:[UITableViewCell class]])
    cell = (UITableViewCell *)cell.superview;
  NSAssert(cell, @"Must find the cell");
  if(!cell)
    return;
  
  [self tableView:tableView accessoryButtonTappedForRowWithIndexPath:[tableView indexPathForCell:cell]];
}

#ifdef THEME_USE_CUSTOM_TABLE_VIEW_SECTION_HEADERS

- (UIView *)tableView:(UITableView *)tableView viewForHeaderInSection:(NSInteger)section
{
  return [tableView themeViewForHeaderInSection:section];
}

- (CGFloat)tableView:(UITableView *)tableView heightForHeaderInSection:(NSInteger)section
{
  return [tableView themeHeightForHeaderInSection:section];
}

#endif

- (void)tableView:(UITableView *)tableView openEntryAtIndexPath:(NSIndexPath *)indexPath
{
  NSArray *effSections;
  NSDictionary *effEntries;
  [self getEffDataForTable:tableView sections:&effSections entries:&effEntries];
  
  NSNumber *section = effSections[indexPath.section];
  
  Entry *entry = effEntries[section][indexPath.row];
  m_currentlyDisplayedEntryUUID = entry.uuid;
  m_currentlyDisplayedDetails = [[EntryDetailsViewController alloc]initWithFile:m_file ctx:m_safeCtx entry:entry];
  
  [self.navigationController pushViewController:m_currentlyDisplayedDetails animated:YES];
  m_rowBeingEdited = YES;
}

#pragma mark - Cell popover menu

- (BOOL)tableView:(UITableView *)tableView shouldShowMenuForRowAtIndexPath:(NSIndexPath *)indexPath
{
  UIMenuItem *userMenuItem = [[UIMenuItem alloc]initWithTitle:@"User" action:@selector(copyUsername:)];
  UIMenuItem *pwdMenuItem = [[UIMenuItem alloc]initWithTitle:@"Password" action:@selector(copyPassword:)];
  UIMenuItem *openMenuItem = [[UIMenuItem alloc]initWithTitle:@"Open" action:@selector(openUrl:)];
  UIMenuItem *emailMenuItem = [[UIMenuItem alloc]initWithTitle:@"Email" action:@selector(copyEmail:)];
  UIMenuItem *urlMenuItem = [[UIMenuItem alloc]initWithTitle:@"URL" action:@selector(copyUrl:)];
  UIMenuItem *titleMenuItem = [[UIMenuItem alloc]initWithTitle:@"Title" action:@selector(copyTitle:)];
  UIMenuItem *notesMenuItem = [[UIMenuItem alloc]initWithTitle:@"Notes" action:@selector(copyNotes:)];
  NSArray *menuItems = @[userMenuItem, pwdMenuItem, openMenuItem, emailMenuItem, urlMenuItem, titleMenuItem, notesMenuItem];
  
  if(UIAccessibilityIsVoiceOverRunning() && [UIActionSheet BlockActionSheetSupported])
  {
    UIActionSheet *actionSheet = [[UIActionSheet alloc]init];
    
    for(UIMenuItem *menuItem in menuItems)
    {
      if([self tableView:tableView canPerformAction:menuItem.action forRowAtIndexPath:indexPath withSender:nil])
        [actionSheet addButtonWithTitle:menuItem.title];
    }
    actionSheet.title = @"Entry Action Menu";
    [actionSheet addButtonWithTitle:@"Cancel"];
    [actionSheet setCancelButtonIndex:actionSheet.numberOfButtons - 1];
    [actionSheet setDidDismissBlock:^(UIActionSheet *actionSheet, NSInteger btnIndex) {
      if(btnIndex == actionSheet.cancelButtonIndex)
        return;
      
      NSString *btnTitle = [actionSheet buttonTitleAtIndex:btnIndex];
      for(UIMenuItem *menuItem in menuItems)
      {
        if(![menuItem.title isEqualToString:btnTitle])
          continue;
        [self tableView:tableView performAction:menuItem.action forRowAtIndexPath:indexPath withSender:nil];
        break;
      }
    }];

    [actionSheet showFromToolbar:self.navigationController.toolbar];
    
    return NO;
  }
  
  [[UIMenuController sharedMenuController]setMenuItems:menuItems];
  [[UIMenuController sharedMenuController]update];
  return YES;
}

- (BOOL)tableView:(UITableView *)tableView canPerformAction:(SEL)action forRowAtIndexPath:(NSIndexPath *)indexPath withSender:(id)sender
{
  BOOL isSystem = action == @selector(copy:) || action == @selector(cut:) || action == @selector(paste:) || action == @selector(delete:);
  
  if(isSystem)
    return NO;

  id value;
  if(action == @selector(copyPassword:))
  {
    Entry *entry = [self entryInTableView: tableView atIndexPath: indexPath];
    Entry *baseEntry = [m_file baseEntryForEntry:entry];
    value = (baseEntry ? baseEntry : entry).password;
  }
  else if(action == @selector(copyUsername:))
  {
    value = [self getProperty:@"username" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(openUrl:))
  {
    value = [self getProperty:@"parsedURL" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyEmail:))
  {
    value = [self getProperty:@"email" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyUrl:))
  {
    value = [self getProperty:@"url" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyTitle:))
  {
    value = [self getProperty:@"title" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyNotes:))
  {
    value = [self getProperty:@"notes" atIndexPath:indexPath tableView:tableView];
  }
  
  if(!value)
    return NO;
  
  return [value isKindOfClass:[NSString class]] ? [value length] > 0 : YES;
}

NSString * const kOpenInChromeKey;

- (void)tableView:(UITableView *)tableView performAction:(SEL)action forRowAtIndexPath:(NSIndexPath *)indexPath  withSender:(id)sender
{
  if(action == @selector(copyPassword:))
  {
    [self copyPasswordAtIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyUsername:))
  {
    NSString *username = [self getProperty:@"username" atIndexPath:indexPath tableView:tableView];
    NSString *password;
    if(smartPasteboardEnabled())
    {
      Entry *entry = [self entryInTableView: tableView atIndexPath: indexPath];
      Entry *baseEntry = [m_file baseEntryForEntry:entry];
      password = (baseEntry ? baseEntry : entry).password;
    }
    
    if(username.length == 0)
      return;
    
    if(password.length > 0)
    {
      PasteboardGravitationalManager *pgm = [PasteboardGravitationalManager sharedManager];
      [pgm copyNames:@[@"Username", @"Password"] values:@[username, password] passwordIdx:1];
      MBProgressHUD *hud = [self showHudForProperty:@"Username" onTableView:tableView hiddingAfterDelay:UIAccessibilityIsVoiceOverRunning() ? 4 : 1.5];
      hud.detailsLabelText = @"Tilt device face down to copy password\nwhen using another app.";
    }
    else
    {
      [self copyProperty:@"username" named:(NSString *)@"Username" atIndexPath:indexPath tableView:tableView];
    }
  }
  else if(action == @selector(openUrl:))
  {
    [self copyPasswordAtIndexPath:indexPath tableView:tableView];
    NSURL *url = [self getProperty:@"parsedURL" atIndexPath:indexPath tableView:tableView];
    if(url)
    {
      dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, (int64_t)(0.3 * NSEC_PER_SEC));
      dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
        NSString *scheme = url.scheme;
        BOOL web = [scheme isEqualToString:@"http"] || [scheme isEqualToString:@"https"];

        if(web && [[NSUserDefaults standardUserDefaults]boolForKey:kOpenInChromeKey] && [OpenInChromeController sharedInstance].isChromeInstalled)
          [[OpenInChromeController sharedInstance]openInChrome:url];
        else
          [[UIApplication sharedApplication] openURL:url];
      });
    }
  }
  else if(action == @selector(copyEmail:))
  {
    [self copyProperty:@"email" named:(NSString *)@"Email" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyUrl:))
  {
    [self copyProperty:@"url" named:(NSString *)@"URL" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyNotes:))
  {
    [self copyProperty:@"notes" named:(NSString *)@"Notes" atIndexPath:indexPath tableView:tableView];
  }
  else if(action == @selector(copyTitle:))
  {
    [self copyProperty:@"title" named:(NSString *)@"Title" atIndexPath:indexPath tableView:tableView];
  }
}

- (void)cellMenuPressed:(SEL)action onCell:(EntryCell *)cell onTableView:(UITableView *)tableView
{
  NSIndexPath *indexPath = [tableView indexPathForCell:cell];
  [self tableView:tableView performAction:action forRowAtIndexPath:indexPath withSender:cell];
}

- (void)copyProperty:(NSString *)property named:(NSString *)name atIndexPath:(NSIndexPath *)indexPath tableView:(UITableView *)tableView
{
  NSString *value = [self getProperty:property atIndexPath:indexPath tableView:tableView];
  if(!value.length)
    return;
  [UIPasteboard generalPasteboard].string = value;

  [self showHudForProperty:name onTableView:tableView hiddingAfterDelay:UIAccessibilityIsVoiceOverRunning() ? 3 : 1];
}

- (id)getProperty:(NSString *)property atIndexPath:(NSIndexPath *)indexPath tableView:(UITableView *)tableView
{
  Entry *entry = [self entryInTableView: tableView atIndexPath: indexPath];
  SEL sel = NSSelectorFromString(property);
  NSMethodSignature *sig = [entry methodSignatureForSelector:sel];
  NSInvocation *invocation = [NSInvocation invocationWithMethodSignature:sig];
  invocation.selector = sel;
  NSAssert(sig.methodReturnLength == sizeof(id), @"adjust accordingly");
  [invocation invokeWithTarget:entry];
  __unsafe_unretained id value = nil;
  [invocation getReturnValue:&value];
  return value;
}

- (void)copyPasswordAtIndexPath:(NSIndexPath *)indexPath tableView:(UITableView *)tableView
{
  Entry *entry = [self entryInTableView: tableView atIndexPath: indexPath];
  Entry *baseEntry = [m_file baseEntryForEntry:entry];
  NSString *password = (baseEntry ? baseEntry : entry).password;
  NSString *username = smartPasteboardEnabled() ? entry.username : nil;
  
  if(password && password.length)
  {
    if(!username.length)
    {
      [UIPasteboard generalPasteboard].string = password;
      [PasteboardPasswordCleaner passwordCopied:password];
      [self showHudForProperty:@"Password" onTableView:tableView hiddingAfterDelay:UIAccessibilityIsVoiceOverRunning() ? 3 : 1];
    }
    else
    {
      PasteboardGravitationalManager *pgm = [PasteboardGravitationalManager sharedManager];
      [pgm copyNames:@[@"Password", @"Username"] values:@[password, username] passwordIdx:0];
      MBProgressHUD *hud = [self showHudForProperty:@"Password" onTableView:tableView hiddingAfterDelay:UIAccessibilityIsVoiceOverRunning() ? 4 : 1.5];
      hud.detailsLabelText = @"Tilt device face down to copy username\nwhen using another app.";
    }
  }
}

- (MBProgressHUD *)showHudForProperty:(NSString *)property onTableView:(UITableView *)tableView hiddingAfterDelay:(NSTimeInterval)delay
{
  MBProgressHUD *hud = [[MBProgressHUD alloc]initWithWindow:tableView.window];
  hud.customView = [[UIImageView alloc]initWithImage:[UIImage imageNamed:@"Check.png"]];
  hud.mode = MBProgressHUDModeCustomView;
  hud.labelText = [NSString stringWithFormat:@"%@ copied", property];
  hud.animationType = MBProgressHUDAnimationZoom;
  hud.opacity = .8;
  hud.removeFromSuperViewOnHide = YES;
  if(tableView != self.tableView)
  {
    // In search, offset
    hud.yOffset = -50;
  }
  [tableView.window addSubview:hud];
  [hud show:YES];
  [self performSelector:@selector(hideHud:) withObject:hud afterDelay:delay];
  
  return hud;
}


#pragma mark - Actions

- (void) onAddNewItem: (id)sender
{
  if(!m_file)
    return;
  
  NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
  NSNumber *generateNewPwObj = [defaults objectForKey:@"autoGeneratePasswords"];
  BOOL generate = !generateNewPwObj || [generateNewPwObj boolValue];
  
  Entry *newEntry = [[Entry alloc]initWithFile:m_file];
  if(m_groupName && m_groupName.length)
    newEntry.group = m_groupName;
  newEntry.title = kNewEntryTitle;

  newEntry.password = generate ? [newEntry randomPassword] : nil;
  newEntry.username = m_file.defaultUsername;
  newEntry.pwHistoryKeep = m_file.defaultKeepPwHistory;
  newEntry.pwHistoryMaxEntries = m_file.defaultPwHistoryEntries;
  
  [m_file addEntry:newEntry];
  
  [self adjustAddHelpText:kHelpText viewTagBase:kHelpBaseTag show:NO];

  EntryDetailsViewController *viewController = [[EntryDetailsViewController alloc]initWithFile:m_file ctx:m_safeCtx entry:newEntry];
  viewController.popOnCancel = YES;
  viewController.editing = YES;
  [self.navigationController pushViewController:viewController animated:YES];
  m_rowBeingEdited = YES;
  
  DebugLog(@"Added new item");
}

- (void)onCloseSafe:(id)sender
{
  [m_state removeAllObjects];
  [m_safeCtx expireNow];
  m_safeCtx.opened = NO;
  [UIPasteboard generalPasteboard].string = @"";

  UIViewController *currentVC = self.navigationController.viewControllers.lastObject;
  [currentVC dismissModalViewControllerAnimated:YES];
  [self.navigationController popToRootViewControllerAnimated:YES];
  
  UIView *rootView = [MBProgressHUD rootView];
  MBProgressHUD *hud = [[MBProgressHUD alloc]initWithView:rootView];
  hud.customView = [[UIImageView alloc]initWithImage:[UIImage imageNamed:@"Check.png"]];
  hud.mode = MBProgressHUDModeCustomView;
  hud.labelText = @"Clipboard cleared";
  hud.animationType = MBProgressHUDAnimationZoom;
  hud.yOffset = -50;
  hud.opacity = .8;
  hud.removeFromSuperViewOnHide = YES;
  [rootView addSubview:hud];
  [hud show:YES];
  dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, NSEC_PER_SEC);
  dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
    [hud hide:YES];
  });
}

#pragma mark -
#pragma mark Search

- (BOOL)searchDisplayController:(UISearchDisplayController *)controller shouldReloadTableForSearchScope:(NSInteger)searchOption
{
  m_state[kSearchButton] = @(searchOption);
  [self prepareEntriesListFiltered:YES];
  return YES;
}

- (BOOL)searchDisplayController:(UISearchDisplayController *)controller shouldReloadTableForSearchString:(NSString *)searchString
{
  if(!searchString)
    searchString = @"";
  
  m_state[kSearchTerm] = searchString;
  if(searchString.length)
  {
    [self prepareEntriesListFiltered:YES];
  }
  else
  {
    m_filteredEntries = nil;
    m_filteredSections = nil;
  }
  
  return YES;
}

#pragma mark -
#pragma mark Memory management

- (void)didReceiveMemoryWarning {
    // Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
    
    // Relinquish ownership any cached data, images, etc. that aren't in use.
}



@end

