//
//  NSObject+BlockKVO.m
//  pwSafe
//
//  Created by Jorge P. Vasquez on 04/05/13.
//
//

#import "NSObject+BlockKVO.h"
#import <objc/runtime.h>

@interface KVOObserver : NSObject
{
  NSMutableDictionary *_dict;
  u_int64_t _contextCounter;
}
@end

static const void * const kKeyPathKey = &kKeyPathKey;

@implementation KVOObserver

- (id)init
{
  if(!(self = [super init]))
    return nil;
  
  _dict = [[NSMutableDictionary alloc]init];
  _contextCounter = 0;

  return self;
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
  id key = (__bridge id)context;
  KVOBlock block = _dict[key];
  NSAssert(block, @"Internal inconsistency error");
  block(object, keyPath, change);
}

- (id)addBlockObserverForKeyPath:(NSString *)keyPath options:(NSKeyValueObservingOptions)options target:(id)object handler:(KVOBlock)block
{
  NSString *context = [NSString stringWithFormat:@"%llu", ++_contextCounter];
  objc_setAssociatedObject(context, kKeyPathKey, keyPath, OBJC_ASSOCIATION_COPY_NONATOMIC);
  [object addObserver:self forKeyPath:keyPath options:options context:(__bridge void *)context];
  _dict[context] = [block copy];
  return context;
}

- (void)removeBlockObserver:(id)context target:(id)object
{
  NSString *keyPath = objc_getAssociatedObject(context, kKeyPathKey);
  NSAssert(keyPath, @"Must have a key");
  [object removeObserver:self forKeyPath:keyPath context:(__bridge void *)context];
  [_dict removeObjectForKey:context];
}

@end

static KVOObserver *GlobalObserver;

@interface KVOAutoDeleter : NSObject
{
  __weak id _target;
  id _context;
}

@end

@implementation KVOAutoDeleter

- (id)initWithTarget:(id)target context:(id)context
{
  if(!(self = [super init]))
    return nil;
  
  _target = target;
  _context = context;
  
  return self;
}

- (void)dealloc
{
  [_target removeBlockObserver:_context];
}

@end

static const void * const kAutodeleterKey = &kAutodeleterKey;

@implementation NSObject (BlockKVO)

- (id)addBlockObserverForKeyPath:(NSString *)keyPath options:(NSKeyValueObservingOptions)options autoremove:(BOOL)autoremove handler:(KVOBlock)block
{
  static dispatch_once_t onceToken;
  dispatch_once(&onceToken, ^{
    GlobalObserver = [[KVOObserver alloc]init];
  });
  
  id context = [GlobalObserver addBlockObserverForKeyPath:keyPath options:options target:self handler:block];
  if(autoremove)
  {
    KVOAutoDeleter *autoDeleter = [[KVOAutoDeleter alloc]initWithTarget:self context:context];
    KVOAutoDeleter *otherAutoDeleter = objc_getAssociatedObject(self, kAutodeleterKey);
    if(otherAutoDeleter)
      objc_setAssociatedObject(autoDeleter, kAutodeleterKey, otherAutoDeleter, OBJC_ASSOCIATION_RETAIN_NONATOMIC);
    objc_setAssociatedObject(self, kAutodeleterKey, autoDeleter, OBJC_ASSOCIATION_RETAIN_NONATOMIC);
  }
  
  return context;
}

- (void)removeBlockObserver:(id)context
{
  return [GlobalObserver removeBlockObserver:context target:self];
}

@end
