//
//  OpenSafeViewController.m
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/9/11.
//  Copyright 2011 app77. All rights reserved.
//

#import "OpenSafeViewController.h"
#import "KeepAliveTimePickerViewController.h"
#import "SafeCtx.h"
#import "SafeCtxStorage.h"
#import "AppTheme.h"
#import "SafePwdInputAccessoryView.h"

#ifndef PWSAFE_NO_MOBIUS
#import "MobiusSDK.h"
#endif

#define OPEN_SAFE_VIEW_CONTROLLER_PASSWORD_TAG 51

@implementation OpenSafeViewController
{
  SEL m_action;
  __weak id m_actionTarget;
  UIActivityIndicatorView *m_activityView;
  CGRect m_activityViewFrame;
  BOOL m_failedOnce;
#ifndef PWSAFE_NO_MOBIUS
  BOOL m_mobiusRegistered;
#endif
}

@synthesize safeCtx = m_safeCtx, safePath = m_safePath;

#pragma mark -
#pragma mark Initialization

- (void)clearPassword
{
#ifndef PWSAFE_NO_MOBIUS
  BOOL mobiusRegistered = [[MobiusSDK sharedMobius]isAccountRegistered:m_safeCtx.safeUUID];
  BOOL mobiusChanged = !mobiusRegistered != !m_mobiusRegistered;
#endif
  
  if(!m_failedOnce)
  {
    m_failedOnce = YES;
#ifdef PWSAFE_NO_MOBIUS
    [self.tableView reloadSections:[NSIndexSet indexSetWithIndex:0] withRowAnimation:UITableViewRowAnimationFade];
#else
    if(mobiusChanged)
      [self.tableView reloadData];
    else
      [self.tableView reloadSections:[NSIndexSet indexSetWithIndex:0] withRowAnimation:UITableViewRowAnimationFade];
#endif
  }

  dispatch_async(dispatch_get_main_queue(), ^{
    UITableViewCell *cell = [self.tableView cellForRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:0]];
    UITextField *passwordField = (UITextField *)[cell.contentView viewWithTag:OPEN_SAFE_VIEW_CONTROLLER_PASSWORD_TAG];
    passwordField.text = nil;
    
    // Enable text input
    self.view.userInteractionEnabled = YES;
    self.navigationItem.rightBarButtonItem.enabled = YES;
    
    // Hide activity view
    m_activityView.hidden = YES;
    [m_activityView stopAnimating];
    
    [passwordField becomeFirstResponder];
    
    
    // Disable Done button
    self.navigationItem.rightBarButtonItem.enabled = NO;
  });
}

- (id) initWithSafePath:(NSString *)safeFilePath target:(id)onDoneTarget onDone:(SEL)onDoneSelector
{
  m_safePath = [safeFilePath copy];
  m_safeCtx = [[SafeCtxStorage sharedStorage] safeCtxForPath:m_safePath];
  self = [super initWithStyle:UITableViewStyleGrouped];  
  if(!self)
    return nil;
  
#ifdef PWSAFE_VAULT
  self.navigationItem.title = @"Open Vault";
#else
  self.navigationItem.title = @"Open safe";
#endif
  m_action = onDoneSelector;
  m_actionTarget = onDoneTarget;
  m_failedOnce = NO;
  
  return self;
}

- (void)dealloc
{
  DebugLog(@"Dealloc");
  if(self.isViewLoaded)
  {
    self.tableView.delegate = nil;
    self.tableView.dataSource = nil;
  }
}

#pragma mark -
#pragma mark Actions

- (void) onCancel:(id)sender
{
  m_safeCtx = nil;
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
  [m_actionTarget performSelector:m_action withObject:self];
#pragma clang diagnostic pop
}

#ifndef PWSAFE_NO_MOBIUS

- (void)logonWithMobius:(id)sender
{
  [[MobiusSDK sharedMobius]authenticateAccount:m_safeCtx.safeUUID];
}

#endif

#pragma mark -
#pragma mark View lifecycle

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
  // Return YES for supported orientations
  return (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad || interfaceOrientation == UIInterfaceOrientationPortrait);
}

- (void)viewDidLoad 
{
  [super viewDidLoad];
  [self.tableView themeViewDidLoadInIpadPos:TVP_FORMSHEET];

  UIBarButtonItem *btn  = [[UIBarButtonItem alloc]initWithBarButtonSystemItem:UIBarButtonSystemItemCancel target:self action:@selector(onCancel:)];
  self.navigationItem.leftBarButtonItem = btn;
  
  btn = [[UIBarButtonItem alloc]initWithBarButtonSystemItem:UIBarButtonSystemItemDone target:self action:@selector(onDone:)];
  self.navigationItem.rightBarButtonItem = btn;
  btn.enabled = NO;
}


- (void)viewWillAppear:(BOOL)animated 
{
  [super viewWillAppear:animated];
  if([UIDevice currentDevice].userInterfaceIdiom != UIUserInterfaceIdiomPad)
    [self focusPasswordField];
  [self.tableView reloadRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:1 inSection:0]] withRowAnimation:UITableViewRowAnimationNone];
}

- (void)viewDidAppear:(BOOL)animated
{
  [super viewDidAppear:animated];
  [self focusPasswordField];
}

- (void)focusPasswordField
{
  UITableViewCell *cell = [self.tableView cellForRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:0]];
  UITextField *passwordField = (UITextField *)[cell.contentView viewWithTag:OPEN_SAFE_VIEW_CONTROLLER_PASSWORD_TAG];
  [passwordField becomeFirstResponder];
}

#pragma mark -
#pragma mark Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView 
{
#ifdef PWSAFE_NO_MOBIUS
  return 1;
#else
  m_mobiusRegistered = [[MobiusSDK sharedMobius]isAccountRegistered:m_safeCtx.safeUUID];
  return (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad && m_mobiusRegistered) ? 2 : 1;
#endif
}


- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section 
{
  // Return the number of rows in the section.
  
  return section == 0 ? 2 : 1;
}

- (NSString *) tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
{
  if(section != 0)
    return nil;
  return m_safeCtx.safeName;
}

// Customize the appearance of table view cells.
- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath 
{
  static NSString * const CellIdentifierPassword = @"Password";
  static NSString * const CellIdentifierTimeout = @"Timeout";
#ifndef PWSAFE_NO_MOBIUS
  static NSString * const CellIdentifierMobius = @"Mobius";
#endif
  
  UITableViewCell *cell;
  
  if(indexPath.section == 0)
  {
    UITextField *passwordField;
    if(indexPath.row == 0)
    {
      cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:CellIdentifierPassword];

      NSUInteger w = UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad ? 310 : 130;
      passwordField = [[UITextField alloc]initWithFrame:CGRectMake(160, 12, w, 21)];
      passwordField.borderStyle = UITextBorderStyleNone;
      passwordField.adjustsFontSizeToFitWidth = YES;
      passwordField.secureTextEntry = YES;
      passwordField.textColor = cell.detailTextLabel.textColor;
      passwordField.autocapitalizationType = UITextAutocapitalizationTypeNone;
      passwordField.autocorrectionType = UITextAutocorrectionTypeNo;
      passwordField.enablesReturnKeyAutomatically = YES;
      passwordField.returnKeyType = UIReturnKeyDone;
      passwordField.clearsOnBeginEditing = NO;
      passwordField.clearButtonMode = UITextFieldViewModeWhileEditing;
      passwordField.delegate = self;
      passwordField.tag = OPEN_SAFE_VIEW_CONTROLLER_PASSWORD_TAG;
      [cell.contentView addSubview:passwordField];
      [passwordField addTarget:self action:@selector(onPasswordChanged:) forControlEvents:UIControlEventEditingChanged];
      passwordField.backgroundColor = nil;
      passwordField.opaque = NO;
      passwordField.accessibilityLabel = @"Safe password";
      if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPhone)
      {
        SafePwdInputAccessoryView *accessoryView = [[SafePwdInputAccessoryView alloc]initWithFrame:CGRectMake(0, 0, self.view.frame.size.width, 44.0f)];;
        accessoryView.allowedKeyboardTypes = @[@(UIKeyboardTypeDefault), @(UIKeyboardTypeNumberPad)];
        passwordField.inputAccessoryView = accessoryView;
        passwordField.keyboardType = (UIKeyboardType)[accessoryView.allowedKeyboardTypes[0] integerValue];
        accessoryView.textField = passwordField;
#ifndef PWSAFE_NO_MOBIUS
        if([[MobiusSDK sharedMobius]isAccountRegistered:m_safeCtx.safeUUID])
          accessoryView.additionalButtons = @[ [[UIBarButtonItem alloc]initWithTitle:@"Unlock with Mobius" style:UIBarButtonItemStyleBordered target:self action:@selector(logonWithMobius:)]];
#endif
      }
      cell.textLabel.text = @"Password";
      cell.selectionStyle = UITableViewCellSelectionStyleNone;
    }
    else
    {
      NSAssert2(indexPath.row == 1, @"Internal error @ %s:%d", __FILE__, __LINE__);
      cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifierTimeout];
      if (cell == nil)
      {
        cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:CellIdentifierTimeout];
      }
      cell.textLabel.text = @"Keep open for";
      cell.detailTextLabel.text = m_safeCtx.keepAlivePeriodText;
      cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
      cell.accessibilityTraits = UIAccessibilityTraitButton;
    }
  }
  else
  {
#ifndef PWSAFE_NO_MOBIUS
    NSAssert2(indexPath.section == 1, @"Internal error @ %s:%d", __FILE__, __LINE__);
    // section 1
    cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifierMobius];
    if(!cell)
    {
      cell = [[UITableViewCell alloc]initWithStyle:UITableViewCellStyleDefault reuseIdentifier:CellIdentifierMobius];
      cell.textLabel.textAlignment = UITextAlignmentCenter;
      cell.textLabel.text = @"Unlock with Mobius";
      cell.accessibilityTraits = UIAccessibilityTraitButton;
    }
#endif
  }
  [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];
  
  return cell;
}

#pragma mark -
#pragma mark Table view delegate

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath 
{
#ifndef PWSAFE_NO_MOBIUS
  if(indexPath.section == 1)
  {
    NSAssert(indexPath.row == 0, @"What row is that?");
    [self logonWithMobius:tableView];
  }
#else
  NSAssert(indexPath.section == 0, @"What section is that?");
#endif
  
  if(indexPath.row != 1)
    return;

  KeepAliveTimePickerViewController *detailViewController = [[KeepAliveTimePickerViewController alloc] initWithSafeCtx:m_safeCtx];
  [self.navigationController pushViewController:detailViewController animated:YES];
}

- (NSString *)tableView:(UITableView *)tableView titleForFooterInSection:(NSInteger)section
{
  if(section != 0)
    return nil;
  
#ifdef PWSAFE_VAULT
  NSString *safeStr = @"vault";
#else
  NSString *safeStr = @"safe";
#endif
  
  if(m_failedOnce)
    return [NSString stringWithFormat:@"Tip: you may be able to restore an older version of your %@ on the app settings.", safeStr];
  else
    return nil;
}

#pragma mark -
#pragma mark Password field delegate

- (void)onDone:(id)sender
{
  self.navigationItem.rightBarButtonItem.enabled = NO;

  NSAssert(m_safeCtx.password && m_safeCtx.password.length, @"Cannot be empty");
  
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
  [m_actionTarget performSelector:m_action withObject:self];
#pragma clang diagnostic pop

  // Disable text input and enable activity view
  self.view.userInteractionEnabled = NO;
  
  UIView *parentView = self.tableView;
  if(!m_activityView)
  {
    m_activityView = [[UIActivityIndicatorView alloc]initWithActivityIndicatorStyle:UIActivityIndicatorViewStyleWhiteLarge];
    m_activityView.color = kGrouppedTableViewLabelColor;
    [parentView addSubview:m_activityView];
    CGSize bounds = parentView.frame.size;
    [m_activityView layoutIfNeeded];
    CGSize actvSize = m_activityView.frame.size;
    m_activityViewFrame = CGRectMake(rintf((bounds.width - actvSize.width)/2), rintf((bounds.height - actvSize.height)/2), actvSize.width, actvSize.height);
  }
  m_activityView.hidden = NO;
  m_activityView.frame = m_activityViewFrame;
  [m_activityView startAnimating];
}

- (BOOL) textFieldShouldReturn:(UITextField *)textField
{
  NSString *password = textField.text;
  if(!password || ![password length])
    return NO;
  NSAssert([password isEqualToString:m_safeCtx.password], @"Should be the same");
  [self onDone:textField];
  return NO;
}

- (void)textFieldDidBeginEditing:(UITextField *)textField
{
  if(!iOS6())
    textField.textColor = kThemeTintColor;
}

- (void)textFieldDidEndEditing:(UITextField *)textField
{
  if(!iOS6())
    textField.textColor = [UIColor blackColor];
}

-(void) onPasswordChanged:(UITextField *)textField
{
  NSString *password = textField.text;
  self.navigationItem.rightBarButtonItem.enabled = password && [password length];
  m_safeCtx.password = password;
}

@end

