//
//  PasswordPolicy.m
//  pwsafemac
//
//  Created by Jorge Peixoto Vasquez on 26/04/12.
//  Copyright (c) 2012 app77. All rights reserved.
//

#import "PasswordPolicy.h"
#import "core/PWPolicy.h"
#import "core/PWSprefs.h"
#import "core/PWScore.h"
#import "SafeFile.h"
#import "Entry.h"
#import "NSStringUtils.h"
#import <memory>

@implementation PasswordPolicy
{
  PWPolicy m_policy;
}
@synthesize policy = m_policy;

#if !__has_feature(objc_arc)
#error only for ARC
#endif

- (id)initForSafe:(SafeFile *)file
{
  if(!(self = [super init]))
    return nil;
  
  std::auto_ptr<PWSprefs> prefs(new PWSprefs);
  prefs->Load(file.pwsCore.GetDBPreferences());
  
  m_policy.flags = 0;
  m_policy.digitminlength = 0;
  m_policy.lowerminlength = 0;
  m_policy.symbolminlength = 0;
  m_policy.upperminlength = 0;
  
  if(prefs->GetPref(PWSprefs::PWUseLowercase))
  {
    m_policy.flags |= PWPolicy::UseLowercase;
    m_policy.lowerminlength = prefs->GetPref(PWSprefs::PWLowercaseMinLength);
    if(!m_policy.lowerminlength)
      m_policy.lowerminlength = 1;
  }
  if(prefs->GetPref(PWSprefs::PWUseUppercase))
  {
    m_policy.flags |= PWPolicy::UseUppercase;
    m_policy.upperminlength = prefs->GetPref(PWSprefs::PWUppercaseMinLength);
    if(!m_policy.upperminlength)
      m_policy.upperminlength = 1;
  }
  if(prefs->GetPref(PWSprefs::PWUseDigits))
  {
    m_policy.flags |= PWPolicy::UseDigits;
    m_policy.digitminlength = prefs->GetPref(PWSprefs::PWDigitMinLength);
    if(!m_policy.digitminlength)
      m_policy.digitminlength = 1;
  }
  if(prefs->GetPref(PWSprefs::PWUseSymbols))
  {
    m_policy.flags |= PWPolicy::UseSymbols;
    m_policy.symbolminlength = prefs->GetPref(PWSprefs::PWSymbolMinLength);
    if(!m_policy.symbolminlength)
      m_policy.symbolminlength = 1;
  }
  if(prefs->GetPref(PWSprefs::PWUseHexDigits))
    m_policy.flags |= PWPolicy::UseHexDigits;
  if(prefs->GetPref(PWSprefs::PWUseEasyVision))
    m_policy.flags |= PWPolicy::UseEasyVision;
  if(prefs->GetPref(PWSprefs::PWMakePronounceable))
    m_policy.flags |= PWPolicy::MakePronounceable;
  
  m_policy.length = prefs->GetPref(PWSprefs::PWDefaultLength);
  
  return self;
}

- (id)initForNamedPolicy:(NSString *)policyName inSafe:(SafeFile *)file
{
  if(!(self = [super init]))
    return nil;
  
  [file pwPolicy:m_policy named:policyName];
  
  return self;
}

- (id)initWithEmptyPolicy
{
  if(!(self = [super init]))
    return nil;
  m_policy.Empty();
  return self;
}

- (PWPolicy)fixedPolicy
{
  PWPolicy policy = m_policy;
  if(policy.flags & PWPolicy::UseHexDigits)
    policy.flags = PWPolicy::UseHexDigits;
  return policy;
}

- (void)updateSafe:(SafeFile *)safe
{
  PWPolicy policy = [self fixedPolicy];
  std::auto_ptr<PWSprefs> prefs(new PWSprefs);
  prefs->Load(safe.pwsCore.GetDBPreferences());
  
  prefs->SetPref(PWSprefs::PWUseLowercase, policy.flags & PWPolicy::UseLowercase);
  prefs->SetPref(PWSprefs::PWUseUppercase, policy.flags & PWPolicy::UseUppercase);
  prefs->SetPref(PWSprefs::PWUseDigits, policy.flags & PWPolicy::UseDigits);
  prefs->SetPref(PWSprefs::PWUseSymbols, policy.flags & PWPolicy::UseSymbols);
  prefs->SetPref(PWSprefs::PWUseHexDigits, policy.flags & PWPolicy::UseHexDigits);
  prefs->SetPref(PWSprefs::PWUseEasyVision, policy.flags & PWPolicy::UseEasyVision);
  prefs->SetPref(PWSprefs::PWMakePronounceable, policy.flags & PWPolicy::MakePronounceable);
  
  prefs->SetPref(PWSprefs::PWDefaultLength, policy.length);
  prefs->SetPref(PWSprefs::PWDigitMinLength, policy.digitminlength);
  prefs->SetPref(PWSprefs::PWLowercaseMinLength, policy.lowerminlength);
  prefs->SetPref(PWSprefs::PWSymbolMinLength, policy.symbolminlength);
  prefs->SetPref(PWSprefs::PWUppercaseMinLength, policy.upperminlength);
  
  prefs->SetPref(PWSprefs::DefaultSymbols, policy.symbols.empty() ? StringX(L"") : policy.symbols);
  
  safe.pwsCore.GetDBPreferences() = prefs->Store(); 
}

- (void)updateNamedPolicy:(NSString *)policyName inSafe:(SafeFile *)file
{
  [file updatePwPolicy:m_policy named:policyName];
}

- (id)initForEntry:(Entry *)entry
{
  if(!(self = [super init]))
    return nil;
  [entry getPasswordPolicy:m_policy];
  if(m_policy.flags == 0)
    return [self initForSafe:entry.safeFile];
  
  return self;
}

- (void)updateEntry:(Entry *)entry;
{
  PasswordPolicy *safePwPolicy = [[PasswordPolicy alloc]initForSafe:entry.safeFile];
  const PWPolicy &safePolicy = safePwPolicy.policy;
  
  PWPolicy newPolicy = [self fixedPolicy];
  
  if(safePolicy == newPolicy)
  {
    PWPolicy emptyPolicy;
    emptyPolicy.Empty();
    [entry setPasswordPolicy:emptyPolicy];
  }
  else
  {
    [entry setPasswordPolicy:newPolicy];
  }
}

- (BOOL)isEqualToPolicy:(PasswordPolicy *)other
{
  return m_policy == other->m_policy;
}

- (id)copyWithZone:(NSZone *)zone
{
  PasswordPolicy *other = [[PasswordPolicy allocWithZone:zone]init];
  other->m_policy = m_policy;
  return other;
}

@end
