//
//  PasteboardGravitationalManager.m
//  pwSafe
//
//  Created by Jorge P. Vasquez on 27/11/13.
//
//

#import "PasteboardGravitationalManager.h"
#import "PasteboardPasswordCleaner.h"

#import <CoreMotion/CoreMotion.h>

@interface PasteboardGravitationalManager()
{
  NSArray *_names;
  NSArray *_values;
  NSUInteger _idx;
  NSUInteger _passwordIdx;
  UIBackgroundTaskIdentifier _bgTask;
  CMMotionManager *_motionManager;
  NSTimer *_timer;
  BOOL _flipped;
  NSUInteger _numberFlips;
}

@end

#define kMaxFlips 2 // password -> username -> password

@implementation PasteboardGravitationalManager

+ (instancetype)sharedManager
{
  static PasteboardGravitationalManager *sharedManager;
  static dispatch_once_t onceToken;
  dispatch_once(&onceToken, ^{
    sharedManager = [[PasteboardGravitationalManager alloc]init];
  });
  return sharedManager;
}

- (instancetype)init
{
  self = [super init];
  if(!self)
    return nil;
  
  [[NSNotificationCenter defaultCenter]addObserver:self selector:@selector(didEnterBackground:) name:UIApplicationDidEnterBackgroundNotification object:nil];
  [[NSNotificationCenter defaultCenter]addObserver:self selector:@selector(willEnterForground:) name:UIApplicationWillEnterForegroundNotification object:nil];
  _bgTask = UIBackgroundTaskInvalid;
  
  return self;
}

- (void)copyNames:(NSArray *)names values:(NSArray *)values passwordIdx:(NSUInteger)passwordIdx
{
  NSParameterAssert(names.count == values.count);
  NSParameterAssert(names.count > 0);
  NSAssert(_timer == nil, @"Must be stopped");
  
  _names = names;
  _values = values;
  _idx = values.count - 1;

  _numberFlips = 0;
  
  [self advance];
}

- (void)advance
{
  if(++_idx >= _values.count)
    _idx = 0;
  
  UIPasteboard *pasteboard = [UIPasteboard generalPasteboard];
  
  NSString *value = _values[_idx];
  pasteboard.string = value;
  
  UIApplication *app = [UIApplication sharedApplication];
  
  if(app.applicationState != UIApplicationStateActive)
  {
    UILocalNotification *notif = [[UILocalNotification alloc]init];
    notif.alertBody = [NSString stringWithFormat:@"%@ copied", _names[_idx]];
    notif.hasAction = NO;
    notif.soundName = UILocalNotificationDefaultSoundName;
    
    [app presentLocalNotificationNow:notif];
  }
  
  if(_idx == _passwordIdx)
    [PasteboardPasswordCleaner passwordCopied:value];
  else
    [PasteboardPasswordCleaner passwordRemovedFromPasteboard];
}

- (void)timerFired:(NSTimer *)timer
{
  UIPasteboard *pasteboard = [UIPasteboard generalPasteboard];
  NSString *currentPasteboardValue = pasteboard.string;
  if(currentPasteboardValue.length == 0 || ![currentPasteboardValue isEqualToString:_values[_idx]])
  {
    [self stop];
    return;
  }
  
  CMAccelerometerData *acc = [_motionManager accelerometerData];
  if(!acc)
    return;
  
  if(acc.acceleration.z < 0.5)
  {
    _flipped = NO;
    return;
  }

  if(_flipped)
    return;
  
  if(++_numberFlips > kMaxFlips)
  {
    [self stop];
    return;
  }
  
  _flipped = YES;
  [self advance];
}

- (void)didEnterBackground:(NSNotification *)notif
{
  NSTimeInterval updateInterval = 1./5;

  if(!_motionManager)
  {
    _motionManager = [[CMMotionManager alloc]init];
    _motionManager.accelerometerUpdateInterval = updateInterval;
  }
  
  if(!_motionManager.isAccelerometerAvailable)
  {
    _motionManager = nil;
    return;
  }
  
  [_motionManager startAccelerometerUpdates];
  _flipped = NO;
  
  _timer = [NSTimer scheduledTimerWithTimeInterval:updateInterval target:self selector:@selector(timerFired:) userInfo:nil repeats:YES];
  
  _bgTask = [[UIApplication sharedApplication]beginBackgroundTaskWithExpirationHandler:^{
    [self stop];
  }];
}

- (void)willEnterForground:(NSNotification *)notif
{
  [self stop];
}

- (void)stop
{
  [_timer invalidate];
  _timer = nil;
  [_motionManager stopAccelerometerUpdates];
  
  UIBackgroundTaskIdentifier bgTask = _bgTask;
  _bgTask = UIBackgroundTaskInvalid;
  [[UIApplication sharedApplication]endBackgroundTask:bgTask];
}

@end
