//
//  PwGenParsViewController.m
//  pwSafe
//
//  Created by Jorge Peixoto Vasquez on 09/08/11.
//  Copyright (c) 2011 app77. All rights reserved.
//

#import "PwGenParsViewController.h"
#import "PWPolicy.h"
#import "PWScore.h"
#import "NSStringUtils.h"
#import "Entry.h"
#import "PWSprefs.h"
#import "SafeFile.h"
#import "AppTheme.h"
#import "AccessibleTableViewCell.h"

template <typename T, size_t N>
char ( &_ArraySizeHelper( T (&array)[N] ))[N];

#define _countof( array ) (sizeof( _ArraySizeHelper( array ) ))

static void getCurrentSafePolicy(PWPolicy &policy, SafeFile *file);

namespace {

enum PasswordRowType
{
  PRT_BOOL,
  PRT_INT,
  PRT_STRING,
  PRT_SEPARATOR
};


struct RowDef
{
  PasswordRowType type;
  NSString *name;
  NSString *accessibilityLabel;
  const struct RowDef *dependency;
  SEL setter;
  SEL getter;
  int PWPolicy::*policyIntField;
  unsigned int policyFlag;
  BOOL invertedDependency;
  
  BOOL dependsOn(const struct RowDef &otherRow) const
  {
    if(!dependency)
      return NO;
    if(dependency == &otherRow)
      return YES;
    return dependency->dependsOn(otherRow);
  }
};

};

static const RowDef Rows[] =
{
  /*  0 */ { PRT_BOOL, @"Hexadecimal only", nil, 0, 0, 0, 0, PWPolicy::UseHexDigits, YES },
  /*  1 */ { PRT_INT, @"Password length", nil, 0, 0, 0, &PWPolicy::length, 0, NO},
  /*  2 */ { PRT_SEPARATOR },
  /*  3 */ { PRT_BOOL, @"Lowercase characters", nil, &Rows[0], 0, 0, 0, PWPolicy::UseLowercase, NO },
  /*  4 */ { PRT_INT, @" ↳ at least", @"minimum lowercase characters count", &Rows[3], 0, 0, &PWPolicy::lowerminlength, 0, NO },
  /*  5 */ { PRT_SEPARATOR },
  /*  6 */ { PRT_BOOL, @"Uppercase characters", nil, &Rows[0], 0, 0, 0, PWPolicy::UseUppercase, NO },
  /*  7 */ { PRT_INT, @" ↳ at least", @"minimum uppercase characters count", &Rows[6], 0, 0, &PWPolicy::upperminlength, 0, NO },
  /*  8 */ { PRT_SEPARATOR },
  /*  9 */ { PRT_BOOL, @"Use digits", nil, &Rows[0], 0, 0, 0, PWPolicy::UseDigits, NO },
  /* 10 */ { PRT_INT, @" ↳ at least", @"minimum digits count", &Rows[9], 0, 0, &PWPolicy::digitminlength, 0, NO },
  /* 11 */ { PRT_SEPARATOR },
  /* 12 */ { PRT_BOOL, @"Use symbols", nil, &Rows[0], 0, 0, 0, PWPolicy::UseSymbols, NO },
  /* 13 */ { PRT_INT, @" ↳ at least", @"minimum symbols count", &Rows[12], 0, 0, &PWPolicy::symbolminlength, 0, NO },
  /* 14 */ { PRT_BOOL, @" ↳ Custom symbols", @"use custom symbols", &Rows[12], @selector(setUseCustomSymbols:), @selector(useCustomSymbols), 0, 0, NO },
  /* 15 */ { PRT_STRING, @"    ↳ Symbols", @"custom symbols", &Rows[14], @selector(setCustomSymbols:), @selector(customSymbols), 0, 0, NO },
  /* 16 */ { PRT_SEPARATOR },  
  /* 17 */ { PRT_BOOL, @"Easy to read", nil, &Rows[0], 0, 0, 0, PWPolicy::UseEasyVision, NO },
  /* 18 */ { PRT_BOOL, @"Pronounceable", nil, &Rows[0], 0, 0, 0, PWPolicy::MakePronounceable, NO },
};

@interface PwGenParsViewController () <AccessibleTableViewCellDelegate>
{
  PWPolicy m_policy;
  StringX m_originalSymbols;
  __weak id m_target;
  SEL m_action;
  BOOL m_canceled;
  BOOL m_useCustomSymbols;
  UITextField *m_currentTextField;
  SafeFile *m_file;
}

@end
  
@implementation PwGenParsViewController


- (id)initWithPolicy:(const PWPolicy &)policy file:(SafeFile *)file
{
  self = [super initWithStyle:UITableViewStyleGrouped];
  if(!self)
    return nil;
    
  m_policy = policy;
  m_originalSymbols = policy.symbols;
  m_useCustomSymbols = !policy.symbols.empty();
  m_canceled = NO;
  m_file = file;
  
  self.navigationItem.title = @"Password policy";
  self.navigationItem.rightBarButtonItem = [[UIBarButtonItem alloc]initWithBarButtonSystemItem:UIBarButtonSystemItemDone target:self action:@selector(onDone:)];
  self.navigationItem.leftBarButtonItem = [[UIBarButtonItem alloc]initWithBarButtonSystemItem:UIBarButtonSystemItemCancel target:self action:@selector(onCancel:)];
    
  return self;
}

static PWSprefs *prefsFromFile(SafeFile *file)
{
#ifdef PWSPREFS_STATICS      
  const PWSprefs *prefs = PWSprefs::GetInstance();
#else
  PWSprefs *prefs = new PWSprefs;
  @synchronized(file) {
    prefs->Load(file.pwsCore.GetDBPreferences());
  }
  dispatch_async(dispatch_get_current_queue(), ^{
    delete prefs;
  });
#endif
  return prefs;
}

static void getCurrentSafePolicy(PWPolicy &policy, SafeFile *file)
{
  std::auto_ptr<PWSprefs> prefs(new PWSprefs);
  @synchronized(file) {
    prefs->Load(file.pwsCore.GetDBPreferences());
  }
  
  policy = prefs->GetDefaultPolicy();
}

- (id)initForCurrentSafe:(SafeFile *)file
{
  PWPolicy policy;
  
  getCurrentSafePolicy(policy, file);
  
  return [self initWithPolicy:policy file:file];
}

- (id)initWithEntry:(Entry *)entry inFile:(SafeFile *)file
{
  PWPolicy policy;
  [entry getPasswordPolicy:policy];
  if(policy.flags == 0)
    getCurrentSafePolicy(policy, file);
  
  return [self initWithPolicy:policy file:file];
}

- (id) initWithPolicyNamed:(NSString *)policyName inFile:(SafeFile *)file
{
  PWPolicy policy;
  [file pwPolicy:policy named:policyName];
  
  return [self initWithPolicy:policy file:file];
}

- (void)dealloc
{
  if(self.isViewLoaded)
  {
    self.tableView.delegate = nil;
    self.tableView.dataSource = nil;
  }
}

- (void)setOnDoneTarget:(id)target action:(SEL)selector
{
  m_target = target;
  m_action = selector;
}

- (void)onCancel:(id)sender
{
  m_canceled = YES;
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
  [m_target performSelector:m_action withObject:self];
#pragma clang diagnostic pop
}

- (void)onDone:(id)sender
{
  if(m_currentTextField && UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPhone)
  {
    [m_currentTextField resignFirstResponder];
    return;
  }
  
  m_canceled = NO;
  
  if(!(m_policy.flags & PWPolicy::UseHexDigits))
  {
    int total = 0; 

    if(m_policy.flags & PWPolicy::UseDigits)
      total += m_policy.digitminlength;
    if(m_policy.flags & PWPolicy::UseLowercase)
      total += m_policy.lowerminlength;
    if(m_policy.flags & PWPolicy::UseUppercase)
      total += m_policy.upperminlength;
    if(m_policy.flags & PWPolicy::UseSymbols)
      total += m_policy.symbolminlength;
  
    if(total > m_policy.length)
    {
      m_policy.length = total;
      [self.tableView reloadRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:1 inSection:0]] withRowAnimation:UITableViewRowAnimationNone];
    }
  }

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
  [m_target performSelector:m_action withObject:self];
#pragma clang diagnostic pop
}

- (PWPolicy)fixedPolicy
{
  PWPolicy policy = m_policy;
  if(policy.flags & PWPolicy::UseHexDigits)
    policy.flags = PWPolicy::UseHexDigits;
  return policy;
}

- (void)updateEntry:(Entry *)entry;
{
  PWPolicy safePolicy;
  getCurrentSafePolicy(safePolicy, m_file);
  PWPolicy newPolicy = [self fixedPolicy];
    
  if(safePolicy == newPolicy)
  {
    PWPolicy emptyPolicy;
    emptyPolicy.Empty();
    [entry setPasswordPolicy:emptyPolicy];
  }
  else
  {
    [entry setPasswordPolicy:newPolicy];
  }
}

- (void)updateSafe
{
  std::auto_ptr<PWSprefs> prefs(new PWSprefs);
  @synchronized(m_file) {
    prefs->Load(m_file.pwsCore.GetDBPreferences());
  }
  PWPolicy newPolicy = [self fixedPolicy];
  
  prefs->SetPref(PWSprefs::PWUseLowercase, newPolicy.flags & PWPolicy::UseLowercase);
  prefs->SetPref(PWSprefs::PWUseUppercase, newPolicy.flags & PWPolicy::UseUppercase);
  prefs->SetPref(PWSprefs::PWUseDigits, newPolicy.flags & PWPolicy::UseDigits);
  prefs->SetPref(PWSprefs::PWUseSymbols, newPolicy.flags & PWPolicy::UseSymbols);
  prefs->SetPref(PWSprefs::PWUseHexDigits, newPolicy.flags & PWPolicy::UseHexDigits);
  prefs->SetPref(PWSprefs::PWUseEasyVision, newPolicy.flags & PWPolicy::UseEasyVision);
  prefs->SetPref(PWSprefs::PWMakePronounceable, newPolicy.flags & PWPolicy::MakePronounceable);
  
  prefs->SetPref(PWSprefs::PWDefaultLength, newPolicy.length);
  prefs->SetPref(PWSprefs::PWDigitMinLength, newPolicy.digitminlength);
  prefs->SetPref(PWSprefs::PWLowercaseMinLength, newPolicy.lowerminlength);
  prefs->SetPref(PWSprefs::PWSymbolMinLength, newPolicy.symbolminlength);
  prefs->SetPref(PWSprefs::PWUppercaseMinLength, newPolicy.upperminlength);
  
  prefs->SetPref(PWSprefs::DefaultSymbols, newPolicy.symbols.empty() ? StringX(L"") : newPolicy.symbols );
  
#ifndef PWSPREFS_STATICS
  @synchronized(m_file) {
    m_file.pwsCore.GetDBPreferences() = prefs->Store();
  }
#endif
  
  [m_file writeSafe];
}

- (void)updatePolicyWithName:(NSString *)policyName
{
  [m_file updatePwPolicy:m_policy named:policyName];
}

- (BOOL)canceled
{
  return m_canceled;
}

- (BOOL)passwordPolicyNumericOnly
{
  return (m_policy.flags & ( PWPolicy::UseLowercase |
                             PWPolicy::UseUppercase |
                             PWPolicy::UseDigits |
                             PWPolicy::UseSymbols |
                             PWPolicy::UseHexDigits |
                             PWPolicy::UseEasyVision |
                             PWPolicy::MakePronounceable)) == PWPolicy::UseDigits;
}

- (BOOL)empty
{
  return !m_policy.flags;
}

- (NSString *)randomPassword
{
  try
  {
    NSAssert(!m_canceled, @"No sense in generating a new password with a canceled policy");
#ifdef PWSPREFS_STATICS
    return [NSString stringWithwstring:m_policy.MakeRandomPassword()];
#else
    return [NSString stringWithwstring:m_policy.MakeRandomPassword(prefsFromFile(m_file))];
#endif
  }
  catch(const std::exception &exception)
  {
    A77Log(@"C++ exception generating password: %s", exception.what());
    return [NSString string];
  }
  catch(...)
  {
    A77Log(@"C++ unknown exception type generating password");
    return [NSString string];
  }
}

#pragma mark - Rows

- (NSNumber *)useCustomSymbols
{
  return @(m_useCustomSymbols);
}

- (void)setUseCustomSymbols:(NSNumber *)newUseCustomSymbols
{
  m_useCustomSymbols = [newUseCustomSymbols boolValue];
  if(!m_useCustomSymbols)
  {
    m_policy.symbols = m_originalSymbols;
  }
}

- (NSString *)customSymbols
{
  return [NSString stringWithwstring:m_policy.symbols];
}

- (void)setCustomSymbols:(NSString *)newCustomSymbols
{
  m_policy.symbols = [newCustomSymbols wstring];
}

- (id)readRow:(const RowDef &)row
{
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
  if(row.getter)
    return [self performSelector:row.getter];
#pragma clang diagnostic pop
  
  if(row.type == PRT_BOOL)
  {
    NSAssert(row.policyFlag, @"Must have a flag");
    return @((m_policy.flags & row.policyFlag) ? YES : NO);
  }
  
  NSAssert(row.type == PRT_INT, @"Unknown type");
  NSAssert(row.policyIntField, @"Must have a field");
  int val = m_policy.*row.policyIntField;
  return @(val);
}

- (bool) isRowEnabled:(const RowDef &)row
{
  const RowDef *dependRow = row.dependency;
  if(!dependRow)
    return YES;
  NSAssert(dependRow->type == PRT_BOOL, @"Can only depend on bool");
  BOOL value = [[self readRow:*dependRow]boolValue]; 
  if(!!dependRow->invertedDependency == !!value)
    return NO;
  if(dependRow->dependency)
    return [self isRowEnabled:*dependRow];
  return YES;
}

- (const RowDef &)rowAtIndex:(NSIndexPath *)indexPath
{
  NSInteger row = indexPath.row, section = indexPath.section;
#ifdef PWSAFE_SIMPLE_PASSWORD_POLICY
  section++;
#endif
  
  for(const RowDef &r : Rows)
  {
    if(r.type == PRT_SEPARATOR)
    {
      section--;
      continue;
    }
    if(![self isRowEnabled:r])
      continue;
    if(section)
      continue;
    if(row-- == 0)
      return r;
  }
  
  @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                 reason:[NSString stringWithFormat:@"Invalid row: %@", indexPath]
                               userInfo:nil];
}

- (void)calcDependentRowsOf:(const RowDef &)rowDef indexPaths:(NSMutableArray *)idxPaths
{
  NSInteger row = 0, section = 0;
#ifdef PWSAFE_SIMPLE_PASSWORD_POLICY
  section--;
#endif

  for(const RowDef &r : Rows)
  {
    if(r.type == PRT_SEPARATOR)
    {
      section++;
      row = 0;
      continue;
    }
    
    BOOL depends = r.dependsOn(rowDef);
    BOOL enabled = [self isRowEnabled:r];

    if(depends && enabled)
    {
      [idxPaths addObject:[NSIndexPath indexPathForRow:row inSection:section]];
      DebugLog(@"Row %@ depends on row %@", r.name, rowDef.name);
    }
    
    if([self isRowEnabled:r])
       row++;
  }
}


- (const RowDef &)rowForView:(UIView *)view
{
  UITableViewCell *cell = (UITableViewCell *)view.superview;
  while(cell && ![cell.class isSubclassOfClass:[UITableViewCell class]])
    cell = (UITableViewCell *)cell.superview;
  
  NSAssert(cell, @"Must be of the right class");
  NSIndexPath *path = [self.tableView indexPathForCell:cell];
  const RowDef &row = [self rowAtIndex:path];  
  return row;
}

- (void)onBoolChanged:(UISwitch *)sw
{
  const RowDef &row  = [self rowForView:sw];
  BOOL on = sw.on;
  BOOL adddingRows;
  if(row.invertedDependency)
    adddingRows = !on;
  else
    adddingRows = on;

  NSMutableArray *affectedRows = [NSMutableArray arrayWithCapacity:_countof(Rows)];
  
  // Turning off, calc before saving
  if(!adddingRows)
    [self calcDependentRowsOf:row indexPaths:affectedRows];
  
  // Save value
  if(row.policyFlag)
  {
    if(on)
      m_policy.flags |= row.policyFlag;
    else
      m_policy.flags &= ~row.policyFlag;
  }
  else
  {
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
    [self performSelector:row.setter withObject:@(on)];
#pragma clang diagnostic pop
  }
  
  // Turning on, calc after saving
  if(adddingRows)
    [self calcDependentRowsOf:row indexPaths:affectedRows];
      
  if(adddingRows)
    [self.tableView insertRowsAtIndexPaths:affectedRows withRowAnimation:UITableViewRowAnimationTop];
  else
    [self.tableView deleteRowsAtIndexPaths:affectedRows withRowAnimation:UITableViewRowAnimationTop];

  if(iOS6())
  {
    for(NSIndexPath *row in affectedRows)
    {
      if(row.row == 0)
        continue;
      NSIndexPath *rowBefore = [NSIndexPath indexPathForRow:row.row - 1 inSection:row.section];
      UITableViewCell *cell = [self.tableView cellForRowAtIndexPath:rowBefore];
      [self.tableView addBackgroundsToGroupedCell:cell atIndexPath:rowBefore];
    }
  }
}

- (BOOL)textField:(UITextField *)textField shouldChangeCharactersInRange:(NSRange)range replacementString:(NSString *)string
{
  const RowDef &row  = [self rowForView:textField];
  if(row.type != PRT_INT)
    return YES;
  if(string.length == 0)
    return YES;
  NSScanner *scanner = [NSScanner scannerWithString:string];
  int x;
  BOOL scanned = [scanner scanInt:&x];
  NSUInteger loc = scanner.scanLocation;
  return scanned && loc == string.length;
}

- (void)textFieldDidBeginEditing:(UITextField *)textField
{
  if(!iOS6())
    textField.textColor = kThemeTintColor;

  [textField selectAll:textField];
  m_currentTextField = textField;
}

- (void)textFieldDidEndEditing:(UITextField *)textField
{
  if(!iOS6())
    textField.textColor = [UIColor blackColor];

  if(textField == m_currentTextField)
    m_currentTextField = nil;
}

- (void) onTextChanged:(UITextField *)tf
{
  const RowDef &row  = [self rowForView:tf];
  NSString *string = tf.text;
  
  if(row.type == PRT_INT)
  {
    int value;
    if(string.length == 0)
    {
      value = 1;
    }
    else
    {
      NSScanner *scanner = [NSScanner scannerWithString:string];
      BOOL scanned = [scanner scanInt:&value];
      (void)scanned;
      NSAssert(scanned, @"should be filtered to only contain numbers");
    }
    
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
    if(row.setter)
      [self performSelector:row.setter withObject:@(value)];
    else
      m_policy.*(row.policyIntField) = (int)value;
#pragma clang diagnostic pop
    
  }
  else
  {
    NSAssert(row.type == PRT_STRING, @"Don't knwon about other types");
    NSAssert(row.setter, @"Can only set text via a setter");
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
    [self performSelector:row.setter withObject:string];
#pragma clang diagnostic pop
  }
}

#pragma mark - View lifecycle

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
  // Return YES for supported orientations
  return (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad || interfaceOrientation == UIInterfaceOrientationPortrait);
}

- (void)viewDidLoad
{
  [super viewDidLoad];
  if(self.navigationController.modalPresentationStyle == UIModalPresentationFormSheet)
    [self.tableView themeViewDidLoadInIpadPos:TVP_FORMSHEET];  
  else
    [self.tableView themeViewDidLoadInIpadPos:TVP_RIGHT];      
}

- (void)viewWillAppear:(BOOL)animated
{
  [self.navigationController setToolbarHidden:YES animated:animated];
  [super viewWillAppear:animated];
}

- (void)viewDidAppear:(BOOL)animated
{
  [super viewDidAppear:animated];
  m_canceled = NO;
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
#ifdef PWSAFE_SIMPLE_PASSWORD_POLICY
  NSInteger sections = 0;
#else
  NSInteger sections = 1;
#endif
  
  for(const RowDef &row : Rows)
  {
    if(row.type == PRT_SEPARATOR)
      sections++;
  }
  
  return sections;
}


- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
  NSInteger rows = 0;
  NSInteger sec = 0;
#ifdef PWSAFE_SIMPLE_PASSWORD_POLICY
  section++;
#endif
  
  for(const RowDef &row : Rows)
  {
    if(row.type == PRT_SEPARATOR)
    {
      sec++;
      continue;
    }
    if(sec < section)
      continue;
    if(sec > section)
      break;
    if([self isRowEnabled:row])
      rows++;
  }
  
  return rows;
}

#define PGP_TAG_SWITCH 100
#define PGP_TAG_TEXT_FIELD 101
#define TEXT_MARGIN 10.0f

CGFloat cellWidth(void) 
{
  CGFloat w = UI_USER_INTERFACE_IDIOM() != UIUserInterfaceIdiomPad ? 280.0 :
  UIInterfaceOrientationIsPortrait([UIApplication sharedApplication].statusBarOrientation)? 302.0 : 595.0;
  return w;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
  static NSString *BoolCellIdentifier = @"BoolCell";
  static NSString *IntCellIdentifier = @"IntCell";
  static NSString *StringCellIdentifier = @"StringCell";

  const RowDef &row = [self rowAtIndex:indexPath];
  NSString *cellId;
  AccessibleTableViewCell *cell;

  if(row.type == PRT_BOOL)
  {
    cellId = BoolCellIdentifier;
  }
  else if(row.type == PRT_INT)
    cellId = IntCellIdentifier;
  else
  {
    cellId = StringCellIdentifier;
    NSAssert(row.type == PRT_STRING, @"Invalid type");
  }

  cell = [tableView dequeueReusableCellWithIdentifier:cellId];
  if(!cell)
  {
    cell = [[AccessibleTableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:cellId];
    [cell.contentView layoutIfNeeded];
    CGSize cellSize = cell.contentView.frame.size;

    if(row.type == PRT_BOOL)
    {
      UISwitch *sw = [[UISwitch alloc]init];
      sw.tag = PGP_TAG_SWITCH;
      [sw sizeToFit];
      CGRect frame;
      frame.size = sw.frame.size;
      frame.origin.y = rintf((cellSize.height - frame.size.height) / 2);
      frame.origin.x = cellSize.width - frame.size.width - frame.origin.y;
      sw.frame = frame;
      sw.autoresizingMask = UIViewAutoresizingFlexibleLeftMargin;
      sw.opaque = NO;
      sw.backgroundColor = nil;
      cell.accessibilityTraits = UIAccessibilityTraitAdjustable;
      cell.isAccessibilityElement = YES;
      cell.delegate = self;
      
      [cell.contentView addSubview:sw];
      [sw addTarget:self action:@selector(onBoolChanged:) forControlEvents:UIControlEventValueChanged];
    }
    else
    {
      UITextField *tf = [[UITextField alloc]init];
      tf.tag = PGP_TAG_TEXT_FIELD;
      [cell.detailTextLabel layoutIfNeeded];
      UIFont *font = [UIFont boldSystemFontOfSize:17];
      tf.font = font;
      tf.textColor = cell.detailTextLabel.textColor;
      tf.textAlignment = NSTextAlignmentRight;
      tf.backgroundColor = nil;
      tf.opaque = NO;
      tf.delegate = self;
      tf.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleTopMargin | UIViewAutoresizingFlexibleBottomMargin;
      tf.autocorrectionType = UITextAutocorrectionTypeNo;
      [cell.contentView addSubview:tf];
      if(row.type == PRT_INT)
        tf.keyboardType = UIKeyboardTypeDecimalPad;
      [tf addTarget:self action:@selector(onTextChanged:) forControlEvents:UIControlEventEditingChanged];
      
      if(row.type == PRT_INT)
      {
        UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PGP_TAG_TEXT_FIELD];
        tf.isAccessibilityElement = NO;
        cell.accessibilityTraits = UIAccessibilityTraitAdjustable;
        cell.isAccessibilityElement = YES;
        cell.delegate = self;
      }
    }
    cell.contentView.autoresizesSubviews = YES;
  }
  
  cell.textLabel.text = row.name;
  cell.selectionStyle = UITableViewCellSelectionStyleNone;
  cell.accessibilityLabel = row.accessibilityLabel;
  
  if(row.type == PRT_BOOL)
  {
    UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PGP_TAG_SWITCH];
    sw.on = [[self readRow:row]boolValue];
  }
  else
  {
    UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PGP_TAG_TEXT_FIELD];
    tf.text = [[self readRow:row]description];
    CGRect frame;
    frame.size.height = [@"17fp" sizeWithFont:tf.font].height;
    UIFont *titleFont = [UIFont boldSystemFontOfSize:17];
    CGFloat titleW = [row.name sizeWithFont:titleFont].width;
    CGSize cellSize = cell.contentView.frame.size;
    frame.size.width = cellSize.width - titleW - 3 * TEXT_MARGIN;
    frame.origin.y = rintf((cellSize.height - frame.size.height) / 2);
    frame.origin.x = cellSize.width - frame.size.width - TEXT_MARGIN;
    tf.frame = frame;
    tf.accessibilityLabel = row.name;
  }

  [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];

  return cell;
}

#pragma mark - AccessibleTableViewCellDelegate

- (void)accessibleCellIncrement:(AccessibleTableViewCell *)cell
{
  NSIndexPath *path = [self.tableView indexPathForCell:cell];
  const RowDef &row = [self rowAtIndex:path];

  if(row.type == PRT_BOOL)
  {
    UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PGP_TAG_SWITCH];
    [sw setOn:![[self readRow:row]boolValue] animated:YES];
    [self onBoolChanged:sw];
  }
  else
  {
    NSAssert(row.type == PRT_INT, @"Only INT fields can be incremented / decremented");
    int &field = m_policy.*row.policyIntField;
    if(field >= 1024)
      return;
    field++;
    UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PGP_TAG_TEXT_FIELD];
    tf.text = [[self readRow:row]description];
  }
}

- (void)accessibleCellDecrement:(AccessibleTableViewCell *)cell
{
  NSIndexPath *path = [self.tableView indexPathForCell:cell];
  const RowDef &row = [self rowAtIndex:path];
  
  if(row.type == PRT_BOOL)
  {
    UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PGP_TAG_SWITCH];
    [sw setOn:![[self readRow:row]boolValue] animated:YES];
    [self onBoolChanged:sw];
  }
  else
  {
    NSAssert(row.type == PRT_INT, @"Only INT fields can be incremented / decremented");
    int &field = m_policy.*row.policyIntField;
    if(field <= 1)
      return;
    field--;
    UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PGP_TAG_TEXT_FIELD];
    tf.text = [[self readRow:row]description];
  }
}

- (NSString *)accessibleCellValue:(AccessibleTableViewCell *)cell
{
  NSIndexPath *path = [self.tableView indexPathForCell:cell];
  const RowDef &row = [self rowAtIndex:path];

  if(row.type == PRT_INT)
  {
    UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PGP_TAG_TEXT_FIELD];
    return tf.text;
  }
  else
  {
    return [[self readRow:row]boolValue] ? @"On" : @"Off";
  }
}

@end
