//
//  PwHistoryVC.m
//  pwSafe
//
//  Created by Jorge P. Vasquez on 04/10/12.
//
//

#import "PwHistoryVC.h"
#import "Entry.h"
#import "SafeFile.h"
#import "AppTheme.h"
#import "AccessibleTableViewCell.h"

@interface PwHistoryVC () <UITextFieldDelegate, UIActionSheetDelegate, AccessibleTableViewCellDelegate>
{
  NSArray *_history;
  NSDateFormatter *_formatter;
  BOOL _originalHistoryEnabled;
  uint _originalMaxEntries;
}

@end

@implementation PwHistoryVC

- (id)initWithEntry:(Entry *)entry edition:(BOOL)edition
{
  self = [super initWithStyle:UITableViewStyleGrouped];
  if (!self)
    return nil;

  _entry = entry;
  _history = entry.pwHistory;
  _formatter = [[NSDateFormatter alloc]init];
  _formatter.timeStyle = NSDateFormatterShortStyle;
  _formatter.dateStyle = NSDateFormatterMediumStyle;
  _originalHistoryEnabled = _historyEnabled = entry.pwHistoryKeep;
  _originalMaxEntries = _maxEntries = entry.pwHistoryMaxEntries;
  _shouldClearHistory = NO;
  
  self.navigationItem.title = @"Password History";

  if(edition)
    [super setEditing:edition animated:NO];
  
  return self;
}

- (id)initWithSafeFile:(SafeFile *)file
{
  self = [super initWithStyle:UITableViewStyleGrouped];
  if (!self)
    return nil;

  _originalHistoryEnabled = _historyEnabled = file.defaultKeepPwHistory;
  _originalMaxEntries = _maxEntries = file.defaultPwHistoryEntries;

  self.navigationItem.title = @"Password History";
  
  [super setEditing:YES animated:NO];
  
  return self;
}

- (void)dealloc
{
  if(self.isViewLoaded)
  {
    self.tableView.delegate = nil;
    self.tableView.dataSource = nil;
  }
}

- (void)viewDidLoad
{
  [super viewDidLoad];
  [self.tableView themeViewDidLoadInIpadPos:_entry ? TVP_RIGHT : TVP_FORMSHEET];
  self.tableView.allowsSelectionDuringEditing = YES;
  self.navigationItem.rightBarButtonItem = self.editButtonItem;
}

- (void)viewWillAppear:(BOOL)animated
{
  [super viewWillAppear:animated];
  [self.navigationController setToolbarHidden:YES animated:animated];
}

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
  // Return YES for supported orientations
  return (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad || interfaceOrientation == UIInterfaceOrientationPortrait);
}

- (void)actionSheet:(UIActionSheet *)actionSheet clickedButtonAtIndex:(NSInteger)buttonIndex
{
  if(buttonIndex == actionSheet.cancelButtonIndex)
    return;

  NSMutableArray *indexPaths = [[NSMutableArray alloc]initWithCapacity:_history.count];
  for(NSUInteger i = 0; i < _history.count; i++)
    [indexPaths addObject:[NSIndexPath indexPathForRow:i inSection:1]];

  _shouldClearHistory = YES;
  
  [self.tableView deleteRowsAtIndexPaths:indexPaths withRowAnimation:UITableViewRowAnimationLeft];
}

- (void)setEditing:(BOOL)editing animated:(BOOL)animated
{
  UITableView *tableView = self.tableView;
  
  NSUInteger oldCount = [self tableView:tableView numberOfRowsInSection:1];
  
  // Don't call anything on delegates if I'm not visible
  id<PwHistoryVCDelegate> delegate = self.navigationController.visibleViewController == self ? _delegate : nil;
  
  if(delegate)
  {
    if([delegate respondsToSelector:@selector(historyVC:willSetEditing:)])
    {
      if(![delegate historyVC:self willSetEditing:editing])
        return;
    }
    [delegate historyVC:self setEditing:editing];
  }
  
  [super setEditing:editing animated:animated];
  
  NSArray *indexPaths;
  if(_historyEnabled)
    indexPaths = @[[NSIndexPath indexPathForRow:0 inSection:0], [NSIndexPath indexPathForRow:1 inSection:0], [NSIndexPath indexPathForRow:2 inSection:0]];
  else
    indexPaths = @[[NSIndexPath indexPathForRow:0 inSection:0], [NSIndexPath indexPathForRow:1 inSection:0]];
  
  if(editing)
  {
    [tableView insertRowsAtIndexPaths:indexPaths withRowAnimation:animated ? UITableViewRowAnimationAutomatic : UITableViewRowAnimationNone];
  }
  else
  {
    NSUInteger newCount = [self tableView:tableView numberOfRowsInSection:1];;
    [tableView beginUpdates];
    [tableView deleteRowsAtIndexPaths:indexPaths withRowAnimation:animated ? UITableViewRowAnimationTop : UITableViewRowAnimationNone];
    if(newCount != oldCount)
    {
      NSMutableArray *rows = [NSMutableArray arrayWithCapacity:newCount];
      if(newCount > oldCount)
      {
        for(NSUInteger i = oldCount; i < newCount; i++)
          [rows addObject:[NSIndexPath indexPathForRow:i inSection:1]];
        [tableView insertRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:oldCount inSection:1]] withRowAnimation:UITableViewRowAnimationAutomatic];
      }
      else
      {
        for(NSUInteger i = newCount - 1; i <= oldCount - 1; i++)
          [rows addObject:[NSIndexPath indexPathForRow:i inSection:1]];
        [tableView deleteRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:oldCount inSection:1]] withRowAnimation:UITableViewRowAnimationAutomatic];
      }
    }
    [tableView endUpdates];
  }

  // Reset
  if(_shouldClearHistory && _entry)
    NSAssert(_history.count == 0, @"Must be empty");
  self.shouldClearHistory = NO;
}

- (void)reloadData
{
  Entry *entry = _entry;
  
  NSAssert(entry, @"Not implemented for safe edition");
  _history = entry.pwHistory;
  
  if(_originalMaxEntries == _maxEntries)
    _maxEntries = _originalMaxEntries = entry.pwHistoryMaxEntries;
  if(_originalHistoryEnabled == _historyEnabled)
    _historyEnabled = _originalHistoryEnabled = entry.pwHistoryKeep;
  
  if(self.navigationController.visibleViewController == self)
  {
    // Give it time to end animations
    dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, .4 * NSEC_PER_SEC);
    dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
      [self.tableView reloadData];
    });
  }
  else
  {
    [self.tableView reloadData];
  }
}

- (void)setEntry:(Entry *)entry
{
  _entry = entry;
  [self reloadData];
}

- (void)safeWasInternallyModified
{
  [self reloadData];
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
  return _entry ? 2 : 1;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
  if(section == 0)
  {
    if(!self.isEditing)
      return 0;
    return _historyEnabled ? 3 : 2;
  }
  return _shouldClearHistory ? 0 : _history.count;
}

#define PHVC_TAG_SWITCH 8845
#define PHVC_TAG_TEXT_FIELD 8846
#define TEXT_MARGIN 10.0f

static NSString * const BoolCellId = @"BoolCellId";
static NSString * const NumCellId = @"NumCellId";

- (AccessibleTableViewCell *)getBoolCellForTableView:(UITableView *)tableView action:(SEL)action
{
  AccessibleTableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:BoolCellId];
  if(!cell)
  {
    cell = [[AccessibleTableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:BoolCellId];
    [cell.contentView layoutIfNeeded];
    CGSize cellSize = cell.contentView.frame.size;
    
    UISwitch *sw = [[UISwitch alloc]init];
    sw.tag = PHVC_TAG_SWITCH;
    [sw sizeToFit];
    CGRect frame;
    frame.size = sw.frame.size;
    frame.origin.y = rintf((cellSize.height - frame.size.height) / 2);
    frame.origin.x = cellSize.width - frame.size.width - frame.origin.y;
    sw.frame = frame;
    sw.autoresizingMask = UIViewAutoresizingFlexibleLeftMargin;
    sw.opaque = NO;
    sw.backgroundColor = nil;
    
    [cell.contentView addSubview:sw];
    [sw addTarget:self action:action forControlEvents:UIControlEventValueChanged];
    cell.contentView.autoresizesSubviews = YES;
    cell.selectionStyle = UITableViewCellSelectionStyleNone;
    cell.accessibilityTraits = UIAccessibilityTraitAdjustable;
    cell.delegate = self;
    cell.isAccessibilityElement = YES;
  }
  
  return cell;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
  AccessibleTableViewCell *cell;
  
  NSUInteger section = indexPath.section;
  NSUInteger row = indexPath.row;
  if(section == 0)
  {
    NSUInteger rowAfterZero = _historyEnabled ? row : (row + 1);
    if(row == 0)
    {
      cell = [self getBoolCellForTableView:tableView action:@selector(onEnableHistoryChanged:)];
      UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PHVC_TAG_SWITCH];
      sw.on = _historyEnabled;
      cell.textLabel.text = @"Keep history";
    }
    else if(rowAfterZero == 1)
    {
      cell = [tableView dequeueReusableCellWithIdentifier:NumCellId];
      NSString *title = @"Passwords to keep";
      if(!cell)
      {
        cell = [[AccessibleTableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:NumCellId];
        [cell.contentView layoutIfNeeded];
        CGSize cellSize = cell.contentView.frame.size;
        
        UITextField *tf = [[UITextField alloc]init];
        tf.tag = PHVC_TAG_TEXT_FIELD;
        [cell.detailTextLabel layoutIfNeeded];
        UIFont *font = [UIFont boldSystemFontOfSize:17];
        tf.font = font;
        tf.textColor = cell.detailTextLabel.textColor;
        tf.textAlignment = NSTextAlignmentRight;
        tf.backgroundColor = nil;
        tf.opaque = NO;
        CGRect frame;
        frame.size.height = [@"17fp" sizeWithFont:font].height;
        UIFont *titleFont = [UIFont boldSystemFontOfSize:17];
        CGFloat titleW = [title sizeWithFont:titleFont].width;
        frame.size.width = cellSize.width - titleW - 3 * TEXT_MARGIN;
        frame.origin.y = rintf((cellSize.height - frame.size.height) / 2);
        frame.origin.x = cellSize.width - frame.size.width - TEXT_MARGIN;
        tf.frame = frame;
        
        tf.delegate = self;
        tf.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleTopMargin | UIViewAutoresizingFlexibleBottomMargin;
        tf.autocorrectionType = UITextAutocorrectionTypeNo;
        [cell.contentView addSubview:tf];
        tf.keyboardType = UIKeyboardTypeDecimalPad;
        [tf addTarget:self action:@selector(onTextChanged:) forControlEvents:UIControlEventEditingChanged];
        tf.accessibilityLabel = @"Number of passwords to keep";
        
        cell.contentView.autoresizesSubviews = YES;
        cell.selectionStyle = UITableViewCellSelectionStyleNone;
        cell.delegate = self;
        cell.accessibilityTraits = UIAccessibilityTraitAdjustable;
        cell.isAccessibilityElement = YES;
      }
      cell.textLabel.text = title;
      
      UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PHVC_TAG_TEXT_FIELD];
      tf.text = [NSString stringWithFormat:@"%u", _maxEntries];
    }
    else
    {
      NSAssert(rowAfterZero == 2, @"Don't know about other rows");
      if(_entry)
      {
        static NSString * const ClearHistCellID = @"ClearHistCellID";
        cell = [tableView dequeueReusableCellWithIdentifier:ClearHistCellID];
        if(!cell)
        {
          cell = [[AccessibleTableViewCell alloc]initWithStyle:UITableViewCellStyleDefault reuseIdentifier:ClearHistCellID];
          cell.selectionStyle = UITableViewCellSelectionStyleBlue;
          cell.textLabel.textAlignment = UITextAlignmentCenter;
        }
        cell.textLabel.text = @"Clear history";
        cell.accessibilityTraits = UIAccessibilityTraitButton;
      }
      else
      {
        cell = [self getBoolCellForTableView:tableView action:@selector(onUpdateAllEntriesChanged:)];
        UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PHVC_TAG_SWITCH];
        sw.on = _shouldUpdateAllEntries;
        cell.textLabel.text = @"Update all entries";
      }
    }
  }
  else
  {
    static NSString * const HistCellIdentifier = @"HistCell";
    cell = [tableView dequeueReusableCellWithIdentifier:HistCellIdentifier];
    if(!cell)
    {
      cell = [[AccessibleTableViewCell alloc]initWithStyle:UITableViewCellStyleSubtitle reuseIdentifier:HistCellIdentifier];
      if(iOS6())
        cell.textLabel.font = [UIFont fontWithName:@"DejaVuSansMono-Bold" size:18];
      else
        cell.textLabel.font = [UIFont fontWithName:@"DejaVuSansMono" size:15];
    }
    PwHistoryEntry *pwHistEntry = _history[row];
    cell.textLabel.text = pwHistEntry.password;
    cell.detailTextLabel.text = [_formatter stringFromDate:pwHistEntry.changeTimestamp];
    cell.selectionStyle = UITableViewCellSelectionStyleNone;
  }
  
  [tableView addBackgroundsToGroupedCell:cell atIndexPath:indexPath];
  
  return cell;
}

- (BOOL)tableView:(UITableView *)tableView canEditRowAtIndexPath:(NSIndexPath *)indexPath
{
  return NO;
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
{
  if(section == 0)
    return nil;
  return @"History";
}

#pragma mark - Table view delegate

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
  if(!_entry)
    return;
  if(indexPath.section != 0 || indexPath.row != [self tableView:tableView numberOfRowsInSection:0] - 1)
    return;

  [tableView deselectRowAtIndexPath:indexPath animated:YES];
  NSAssert(self.isEditing, @"Must be editing");
  UIActionSheet *actionSheet = [[UIActionSheet alloc]initWithTitle:nil delegate:self cancelButtonTitle:@"Cancel" destructiveButtonTitle:@"Clear history" otherButtonTitles:nil];
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
  {
    UITableViewCell *cell = [tableView cellForRowAtIndexPath:indexPath];
    [actionSheet showFromRect:cell.contentView.frame inView:cell animated:YES];
  }
  else
  {
    [actionSheet showInView:self.view];
  }
}

#pragma mark - Switch

- (void)onEnableHistoryChanged:(UISwitch *)sw
{
  _historyEnabled = sw.on;
  if(_historyEnabled)
    [self.tableView insertRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:1 inSection:0]] withRowAnimation:UITableViewRowAnimationTop];
  else
    [self.tableView deleteRowsAtIndexPaths:@[[NSIndexPath indexPathForRow:1 inSection:0]] withRowAnimation:UITableViewRowAnimationTop];
  
  if(iOS6())
  {
    NSIndexPath *swIndexPath = [NSIndexPath indexPathForRow:0 inSection:0];
    UITableViewCell *swCell = [self.tableView cellForRowAtIndexPath:swIndexPath];
    [self.tableView addBackgroundsToGroupedCell:swCell atIndexPath:swIndexPath];
  }
}

- (void)onUpdateAllEntriesChanged:(UISwitch *)sw
{
  _shouldUpdateAllEntries = sw.on;
}

#pragma mark - TextField

- (BOOL)textField:(UITextField *)textField shouldChangeCharactersInRange:(NSRange)range replacementString:(NSString *)string
{
  if(string.length == 0)
    return YES;
  NSScanner *scanner = [NSScanner scannerWithString:string];
  int x;
  BOOL scanned = [scanner scanInt:&x];
  NSUInteger loc = scanner.scanLocation;
  
  return scanned && loc == string.length;
}

- (void) onTextChanged:(UITextField *)tf
{
  NSString *string = tf.text;
  int value;
  if(string.length == 0)
    value = 1;
  else
  {
    NSScanner *scanner = [NSScanner scannerWithString:string];
    BOOL scanned = [scanner scanInt:&value];
    (void)scanned;
    NSAssert(scanned, @"should be filtered to only contain numbers");
    if(value <= 0)
      value = 1;
    else if(value >= UINT8_MAX)
      value = UINT8_MAX;
  }
  
  _maxEntries = value;
}

- (void)textFieldDidEndEditing:(UITextField *)textField
{
  textField.text = [NSString stringWithFormat:@"%u", _maxEntries];
}

#pragma mark - AccessibleTableViewCellDelegate

- (void)accessibleCellDecrement:(AccessibleTableViewCell *)cell
{
  NSString *reuseId = cell.reuseIdentifier;
  if([reuseId isEqualToString:BoolCellId])
  {
    UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PHVC_TAG_SWITCH];
    [sw setOn:!sw.on animated:YES];
    [sw sendActionsForControlEvents:UIControlEventValueChanged];
  }
  else if([reuseId isEqualToString:NumCellId])
  {
    if(_maxEntries <= 1)
      return;
    _maxEntries--;
    UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PHVC_TAG_TEXT_FIELD];
    tf.text = [NSString stringWithFormat:@"%u", _maxEntries];
  }
}

- (void)accessibleCellIncrement:(AccessibleTableViewCell *)cell
{
  NSString *reuseId = cell.reuseIdentifier;
  if([reuseId isEqualToString:BoolCellId])
  {
    UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PHVC_TAG_SWITCH];
    [sw setOn:!sw.on animated:YES];
    [sw sendActionsForControlEvents:UIControlEventValueChanged];
  }
  else if([reuseId isEqualToString:NumCellId])
  {
    if(_maxEntries == UINT8_MAX)
      return;
    _maxEntries++;
    UITextField *tf = (UITextField *)[cell.contentView viewWithTag:PHVC_TAG_TEXT_FIELD];
    tf.text = [NSString stringWithFormat:@"%u", _maxEntries];
  }
}

- (NSString *)accessibleCellValue:(AccessibleTableViewCell *)cell
{
  NSString *reuseId = cell.reuseIdentifier;
  if([reuseId isEqualToString:BoolCellId])
  {
    UISwitch *sw = (UISwitch *)[cell.contentView viewWithTag:PHVC_TAG_SWITCH];
    return sw.on ? @"On" : @"Off";
  }
  else if([reuseId isEqualToString:NumCellId])
  {
    return [NSString stringWithFormat:@"%u", _maxEntries];
  }
  return nil;
}

@end
