//
//  SafeCtxStorage.m
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/21/11.
//  Copyright 2011 app77. All rights reserved.
//

#import "SafeCtxStorage.h"
#import "SafeCtx.h"
#import "FileSystemFolders.h"

@implementation SafeCtxStorage
{
  NSMutableDictionary *_dictionary;
}

+ (SafeCtxStorage *)sharedStorage
{
  static SafeCtxStorage *storage = nil;
  if(!storage)
    storage = [[SafeCtxStorage alloc]init];
  return storage;
}

- (NSString *)oldStorageFileName
{
  NSArray *cachesDirs = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
  NSString *cacheDir = cachesDirs[0];
  
  return [cacheDir stringByAppendingPathComponent:@"SafeCtxStorage.dat"];
}

- (NSString *)storageFileName
{
  NSArray *libsDirs = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES);
  NSString *libsDir = libsDirs[0];
  
  return [libsDir stringByAppendingPathComponent:@"SafeCtxStorage.dat"];
}


- (NSString *)savedLibDirFileName
{
  NSArray *cachesDirs = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
  NSString *cacheDir = cachesDirs[0];
  
  return [cacheDir stringByAppendingPathComponent:@"SavedLibDir.dat"];
}

- (void)save
{
  DebugLog(@"Saving safes CTX");
  NSData *data = [NSKeyedArchiver archivedDataWithRootObject:_dictionary];
  NSString *libDir = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES)[0];
  NSString *storageFilePath = [self storageFileName];
  NSString *libDirSavePath = [self savedLibDirFileName];
  
  dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_LOW, 0), ^{
    NSError *error = nil;
    static NSString * const syncer = @"syncer";
    
    BOOL ok;
    @synchronized(syncer) {
      ok = [data writeToFile:storageFilePath options:NSDataWritingAtomic|NSDataWritingFileProtectionComplete error:&error];
    }
    if(!ok)
      NSLog(@"ERROR writing safectxstorage data: %@", error);
    
    error = nil;
    NSData *libDirData = [libDir dataUsingEncoding:NSUTF8StringEncoding];
    @synchronized(syncer) {
      ok = [libDirData writeToFile:libDirSavePath options:NSDataWritingAtomic|NSDataWritingFileProtectionComplete error:&error];
    }
    if(!ok)
      NSLog(@"ERROR writing safectxstorage data: %@", error);

    DebugLog(@"Saved safes CTX");
  });
  
}

- (id) init
{
  if(!(self = [super init]))
    return nil;
  
  NSString *savedLibDir = [NSString stringWithContentsOfFile:[self savedLibDirFileName] encoding:NSUTF8StringEncoding error:nil];
  NSString *libDir = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES)[0];
  BOOL update = !savedLibDir || ![savedLibDir isEqualToString:libDir];
  
  NSString *storageFilePath = [self storageFileName];
  NSString *oldStorageFilePath = [self oldStorageFileName];
  NSFileManager *fm = [NSFileManager defaultManager];
  
  if([fm fileExistsAtPath:oldStorageFilePath] && ![fm fileExistsAtPath:storageFilePath])
    [fm moveItemAtPath:oldStorageFilePath toPath:storageFilePath error:nil];
  
  // Load the state
  _dictionary = [NSKeyedUnarchiver unarchiveObjectWithFile:storageFilePath];
  if(!_dictionary)
  {
    _dictionary = [[NSMutableDictionary alloc]init];
  }
  else if(update)
  {
    NSDictionary *oldDictionary = _dictionary;
    _dictionary = [[NSMutableDictionary alloc]init];
    
    NSString *docsDir = localSafeDir();
    
    for(__strong NSString *key in [oldDictionary allKeys])
    {
      SafeCtx *value = oldDictionary[key];

      NSString *baseDir;
      
#if PWSAFE_WITH_DROPBOX
      NSString *dropBoxDir = dropboxDir();

      if(value.dropboxPath.length == 0)
      {
        baseDir = docsDir;
      }
      else
      {
        // Keep only newer if repeated
        SafeCtx *valueAlreadyInDic = _dictionary[key];
        if(valueAlreadyInDic && [valueAlreadyInDic.dropboxSyncTimestamp compare:value.dropboxSyncTimestamp] == NSOrderedDescending)
          continue;
        baseDir = dropBoxDir;
      }
#else
      baseDir = docsDir;
#endif
      
      key = [baseDir stringByAppendingPathComponent:[key lastPathComponent]];
      [value updateSafePath:key];
      _dictionary[key] = value;
    }

    // Save again to save updated values
    [self save];
  }
  
  if(!_dictionary)
    return nil;
  
  return self;
}

#ifndef PWSAFE_COZI
- (void)decryptPasswordsIfNeeded
{
  for(SafeCtx *ctx in [_dictionary allValues])
    if(!ctx.password)
      [ctx decryptPassword];
}
#endif

- (SafeCtx *) openedSafeCtx
{
  for(SafeCtx *ctx in [_dictionary allValues])
    if(ctx.opened)
      return ctx;
  return nil;
}

- (SafeCtx *)safeCtxForPath:(NSString *)path
{
  NSParameterAssert(path && path.length);
  path = [path precomposedStringWithCanonicalMapping];
  SafeCtx *ctx = _dictionary[path];
  if(!ctx)
  {
    DebugLog(@"Not found safe for %@, creating one", path);
    ctx = [[SafeCtx alloc]initWithPath:path];
    _dictionary[path] = ctx;
  }
  
  return ctx;
}

- (SafeCtx *)objectForKeyedSubscript:(NSString *)path
{
  return [self safeCtxForPath:path];
}

- (SafeCtx *)safeCtxForUUID:(NSString *)uuid
{
  for(SafeCtx *ctx in _dictionary.allValues)
    if([ctx.safeUUID isEqualToString:uuid])
      return ctx;
  
  return nil;
}

- (void)deleteCtxForPath:(NSString *)path
{
  path = [path precomposedStringWithCanonicalMapping];
  [_dictionary removeObjectForKey:path];
}

- (NSArray *)allSafeCtxs 
{
  return [_dictionary allValues];
}

- (void)safeMovedFromPath:(NSString *)oldPath toPath:(NSString *)newPath
{
  oldPath = [oldPath precomposedStringWithCanonicalMapping];
  newPath = [newPath precomposedStringWithCanonicalMapping];
  
  SafeCtx *ctx = _dictionary[oldPath];
  NSAssert(ctx, @"Can't move a non-existent safe");
  if(!ctx)
    return;
  NSAssert([ctx.safePath isEqualToString:oldPath], @"Must be on the place the dictionary says it is");
  
  [_dictionary removeObjectForKey:oldPath];
  [ctx updateSafePath:newPath];
  _dictionary[newPath] = ctx;
}


@end
