//
//  SafeFile+Print.m
//  pwSafe
//
//  Created by Jorge P. Vasquez on 25/09/12.
//
//

#import "SafeFile+Print.h"
#import "SafeFile.h"
#import "Entry.h"
#import <CoreText/CoreText.h>
#import <objc/runtime.h>

@interface SafeFilePaperEntry : NSObject
{
@public
  NSString *title;
  NSString *username;
  NSString *password;
  NSString *url;
  NSString *email;
  NSString *notes;
  NSInteger page;
}
@end

@implementation SafeFilePaperEntry

@end

@interface SafeFilePaperGroup : NSObject
{
@public
  NSString *groupName;
  NSArray *entries;
  NSInteger page;
}

@end

@implementation SafeFilePaperGroup

@end

@interface SafeFilePaperContext : NSObject
{
@public
  CGSize paperSize;
  CGRect printableRect;
  NSMutableArray *groups; // Array of SafeFilePaperGroup
  NSString *safeName;
  NSString *dateString;
  CTFontRef font;
  CTFontRef titleFont;
  CTFontRef notesFont;
  CTFontRef passwordFont;
  CTFontRef headerFont;
  CTFontRef groupFont;
  CTFontRef headerInfoFont;
  NSInteger nPages;
}
@end

@implementation SafeFilePaperContext

#if !__IPHONE_OS_VERSION_MAX_ALLOWED

__attribute__((ns_returns_autoreleased))
static id NSColorToCGColor(id self, SEL cmd, ...)
{
  const NSInteger numberOfComponents = [self numberOfComponents];
  CGFloat components[numberOfComponents];
  CGColorSpaceRef colorSpace = [[self colorSpace] CGColorSpace];
  
  [self getComponents:components];
  
  return (__bridge id)CGColorCreate(colorSpace, components);
}

+ (void)initialize
{
  NSColor *black = [NSColor blackColor];
  if([black respondsToSelector:@selector(CGColor)])
    return;

  class_addMethod([NSColor class], @selector(CGColor), NSColorToCGColor, "@@:");
  NSAssert([black respondsToSelector:@selector(CGColor)], @"Now it must work");
}

#endif

- (void)dealloc
{
  CFRelease(font);
  CFRelease(titleFont);
  CFRelease(passwordFont);
  CFRelease(notesFont);
  CFRelease(headerFont);
  CFRelease(headerInfoFont);
  CFRelease(groupFont);
}

@end

@implementation SafeFile (Print)

- (void)saveAsPDFToPath:(NSString *)pdfFilePath password:(NSString *)password name:(NSString *)friendlyName pageSize:(CGSize)pageSize callback:(SafeFilePDFCallback)callback
{
  NSString *appVersion = [[NSBundle mainBundle]objectForInfoDictionaryKey:@"CFBundleShortVersionString"];
#ifdef __IPHONE_OS_VERSION_MAX_ALLOWED
#ifdef PWSAFE_COZI
  NSString *creator = [@"Cozi Password Vault v. " stringByAppendingString:appVersion];
#else
  NSString *creator = [@"pwSafe for iOS v. " stringByAppendingString:appVersion];
#endif
#else
  NSString *creator = [@"pwSafe for Mac v. " stringByAppendingString:appVersion];
#endif
  NSMutableDictionary *options = [@{
  (id)kCGPDFContextCreator: creator,
  (id)kCGPDFContextTitle: friendlyName,
  } mutableCopy];

  if(password && password.length)
  {
    if(![password canBeConvertedToEncoding:NSASCIIStringEncoding])
    {
      NSError *error = [NSError errorWithDomain:@"SafeFilePrint" code:-100 userInfo:@{NSLocalizedDescriptionKey: @"Could not convert password to ASCII"}];
      dispatch_async(dispatch_get_main_queue(), ^{
        callback(NO, error, pdfFilePath);
      });
      return;
    }
    [options addEntriesFromDictionary:@{
     (id)kCGPDFContextOwnerPassword: password,
     (id)kCGPDFContextUserPassword: password,
     (id)kCGPDFContextEncryptionKeyLength: @128
     }];
  }
  CGRect pageRect = CGRectMake(0, 0, pageSize.width, pageSize.height);
  SafeFilePaperContext *paperContext = [self createPaperContextForPaperSize:pageRect.size name:(NSString *)friendlyName printableRect:pageRect];

  dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
    CGContextRef pdf = CGPDFContextCreateWithURL((__bridge CFURLRef)[NSURL fileURLWithPath:pdfFilePath], &pageRect, (__bridge CFDictionaryRef)options);
    if(!pdf)
    {
      NSError *error = [NSError errorWithDomain:@"SafeFilePrint" code:-101 userInfo:@{NSLocalizedDescriptionKey: @"Could not create PDF file"}];
      dispatch_async(dispatch_get_main_queue(), ^{
        callback(NO, error, pdfFilePath);
      });
    }
    
    NSDictionary *pageOptions = @{(id)kCGPDFContextMediaBox: [NSData dataWithBytes:&pageRect length:sizeof(pageRect)]};
    NSInteger nPages = [self numberOfPagesInPaperContext:paperContext];
    
    for(NSInteger page = 0; page < nPages; page++)
    {
      CGPDFContextBeginPage(pdf, (__bridge CFDictionaryRef)pageOptions);
      [self drawPage:page fromPaperContext:paperContext inGraphicsContext:pdf];
      CGPDFContextEndPage(pdf);
    }
    CGPDFContextClose(pdf);
    CGContextRelease(pdf);

    dispatch_async(dispatch_get_main_queue(), ^{
      callback(YES, nil, pdfFilePath);
    });
  });
}

#define SAFEFILEPDF_MIN_MARGIN 36.0f  // 1/2 inch margin
#define SAFEFILEPDF_HEADER_H   45.0f
#define SAFEFILEPDF_HEADER_V_MARGIN 5.0f
#define SAFEFILEPDF_GROUP_TITLE_BOX_PAD 3.0f
#define SAFEFILEPDF_GROUP_CHILD_OFFSET 8.0f;
#define SAFEFILEPDF_ENTRY_MARGIN 3.0f
#define SAFEFILEPDF_LOGO_MARGIN 6.0f

CGFloat textHeight(CTFontRef font, CGFloat maxW, NSString *text)
{
  text = [text stringByReplacingOccurrencesOfString:@"\r\n" withString:@"\n"];
  
  NSAttributedString *aStr = [[NSAttributedString alloc]initWithString:text attributes:@{ (__bridge id)kCTFontAttributeName: (__bridge id)font }];
  CTFramesetterRef frameSetter = CTFramesetterCreateWithAttributedString((__bridge CFAttributedStringRef)aStr);
  CFRange txtRange;
  CGSize size = CTFramesetterSuggestFrameSizeWithConstraints(frameSetter, CFRangeMake(0, text.length), NULL, CGSizeMake(maxW, CGFLOAT_MAX), &txtRange);
  CFRelease(frameSetter);
  return size.height;
}

- (SafeFilePaperContext *)createPaperContextForPaperSize:(CGSize)paperSize name:(NSString *)friendlyName printableRect:(CGRect)printableRect
{
  SafeFilePaperContext *ctx = [[SafeFilePaperContext alloc]init];
  ctx->paperSize = paperSize;
  ctx->safeName = friendlyName;

  // Calculate effective margin
  //
  CGRect *effPrintableRect = &(ctx->printableRect);
  
  *effPrintableRect = printableRect;
  // Left
  if(effPrintableRect->origin.x < SAFEFILEPDF_MIN_MARGIN)
  {
    CGFloat delta = SAFEFILEPDF_MIN_MARGIN - effPrintableRect->origin.x;
    effPrintableRect->origin.x = SAFEFILEPDF_MIN_MARGIN;
    effPrintableRect->size.width -= delta;
  }
  // Bottom
  if(effPrintableRect->origin.y < SAFEFILEPDF_MIN_MARGIN)
  {
    CGFloat delta = SAFEFILEPDF_MIN_MARGIN - effPrintableRect->origin.y;
    effPrintableRect->origin.y = SAFEFILEPDF_MIN_MARGIN;
    effPrintableRect->size.height -= delta;
  }
  // Right
  if(effPrintableRect->origin.x + effPrintableRect->size.width > paperSize.width - SAFEFILEPDF_MIN_MARGIN)
  {
    effPrintableRect->size.width = paperSize.width - SAFEFILEPDF_MIN_MARGIN - effPrintableRect->origin.x;
  }
  // Top
  if(effPrintableRect->origin.y + effPrintableRect->size.height > paperSize.height - SAFEFILEPDF_MIN_MARGIN)
  {
    effPrintableRect->size.height = paperSize.height - SAFEFILEPDF_MIN_MARGIN - effPrintableRect->origin.y;
  }
  
  // Snapshot data to be printed
  NSArray *groupNames = self.sortedGroups;
  NSMutableArray *groups = [NSMutableArray arrayWithCapacity:groupNames.count];
  for(NSString *groupName in self.sortedGroups)
  {
    NSArray *origEntries = self.entries[groupName];
    NSMutableArray *entries = [NSMutableArray arrayWithCapacity:origEntries.count];
    for(Entry *origEntry in origEntries)
    {
      SafeFilePaperEntry *entry = [[SafeFilePaperEntry alloc]init];
      entry->title = [origEntry.title copy];
      entry->username = [origEntry.username copy];

      Entry *baseEntry = [self baseEntryForEntry:origEntry];
      entry->password = [(baseEntry ? baseEntry : origEntry).password copy];
      entry->url = [origEntry.url copy];
      entry->email = [origEntry.email copy];
      entry->notes = [origEntry.notes copy];
      [entries addObject:entry];
    }
    [entries sortUsingComparator:^NSComparisonResult(SafeFilePaperEntry *e1, SafeFilePaperEntry *e2) {
      return [e1->title caseInsensitiveCompare:e2->title];
    }];
    
    SafeFilePaperGroup *group = [[SafeFilePaperGroup alloc]init];
    group->groupName = groupName;
    group->entries = entries;
    [groups addObject:group];
  }
  ctx->groups = groups;

  // Init fonts
  ctx->font = CTFontCreateWithName((__bridge CFStringRef)@"HelveticaNeue", 11, NULL);
  ctx->titleFont = CTFontCreateWithName((__bridge CFStringRef)@"HelveticaNeue", 13, NULL);
  ctx->groupFont = CTFontCreateWithName((__bridge CFStringRef)@"HelveticaNeue-Bold", 13, NULL);
  ctx->notesFont = CTFontCreateWithName((__bridge CFStringRef)@"HelveticaNeue-Italic", 11, NULL);
  ctx->passwordFont = CTFontCreateWithName((__bridge CFStringRef)@"DejaVuSansMono", 11, NULL);
  ctx->headerFont = CTFontCreateWithName((__bridge CFStringRef)@"HelveticaNeue", 22, NULL);
  ctx->headerInfoFont = CTFontCreateWithName((__bridge CFStringRef)@"HelveticaNeue", 9, NULL);
//  NSLog(@"ctx->font: %@", CTFontCopyLocalizedName(ctx->font, kCTFontFullNameKey, NULL));
//  NSLog(@"ctx->titleFont: %@", CTFontCopyLocalizedName(ctx->titleFont, kCTFontFullNameKey, NULL));
//  NSLog(@"ctx->groupFont: %@", CTFontCopyLocalizedName(ctx->groupFont, kCTFontFullNameKey, NULL));
//  NSLog(@"ctx->notesFont: %@", CTFontCopyLocalizedName(ctx->notesFont, kCTFontFullNameKey, NULL));
//  NSLog(@"ctx->passwordFont: %@", CTFontCopyLocalizedName(ctx->passwordFont, kCTFontFullNameKey, NULL));
//  NSLog(@"ctx->headerFont: %@", CTFontCopyLocalizedName(ctx->headerFont, kCTFontFullNameKey, NULL));
//  NSLog(@"ctx->headerInfoFont: %@", CTFontCopyLocalizedName(ctx->headerInfoFont, kCTFontFullNameKey, NULL));
  
  CGFloat lineH = CTFontGetAscent(ctx->font) + CTFontGetDescent(ctx->font) + CTFontGetLeading(ctx->font);
  CGFloat titleLineH = CTFontGetAscent(ctx->titleFont) + CTFontGetDescent(ctx->titleFont) + CTFontGetLeading(ctx->titleFont);
  CGFloat groupLineH = CTFontGetAscent(ctx->groupFont) + CTFontGetDescent(ctx->groupFont) + CTFontGetLeading(ctx->groupFont);
  
  CGFloat pageH = effPrintableRect->size.height - SAFEFILEPDF_HEADER_H - SAFEFILEPDF_HEADER_V_MARGIN;
  CGFloat pageW = effPrintableRect->size.width;
  NSInteger nPages = 1;
  CGFloat spaceLeft = pageH;
  
  // Calculate number of pages
  for(SafeFilePaperGroup *group in groups)
  {
    CGFloat groupH = SAFEFILEPDF_ENTRY_MARGIN;
    
    CGFloat xOffset;
    BOOL inGroup = group->groupName && group->groupName.length;
    if(inGroup)
    {
      groupH = groupLineH + 2 * SAFEFILEPDF_GROUP_TITLE_BOX_PAD;
      NSArray *groupPath = [group->groupName componentsSeparatedByString:@"."];
      NSUInteger nIndents = groupPath.count > 1 ? groupPath.count - 1 : 0;
      xOffset = nIndents * SAFEFILEPDF_GROUP_CHILD_OFFSET;
    }
    else
    {
      groupH = 0;
      xOffset = 0;
    }
    
    if(spaceLeft < groupH)
    {
      group->page = nPages++;
      spaceLeft = pageH;
    }
    else
    {
      group->page = nPages - 1;
    }
    spaceLeft -= groupH;
    
    BOOL firstEntry = YES;
    for(SafeFilePaperEntry *entry in group->entries)
    {
      CGFloat entryH = titleLineH + lineH + SAFEFILEPDF_ENTRY_MARGIN;
      if(entry->notes && entry->notes.length)
        entryH += textHeight(ctx->notesFont, pageW - xOffset, entry->notes);
      
      if(spaceLeft < entryH)
      {
        entry->page = nPages++;
        spaceLeft = pageH;
        if(inGroup)
        {
          // Bring the group header along for first entry
          if(firstEntry)
          {
            group->page++;
            NSAssert(group->page == entry->page, @"Must be together");
          }
          // Otherwise duplicate the group header
          spaceLeft -= groupH;
        }
      }
      else
      {
        entry->page = nPages - 1;
      }
      spaceLeft -= entryH;
      firstEntry = NO;
    }
  }
  ctx->nPages = nPages;
  
  // Date
  NSDateFormatter *dateFormatter = [[NSDateFormatter alloc]init];
  dateFormatter.dateStyle = NSDateFormatterMediumStyle;
  dateFormatter.timeStyle = NSDateFormatterShortStyle;
  NSDate *date = [NSDate date];
  ctx->dateString = [@"Printed on " stringByAppendingString:[dateFormatter stringFromDate:date]];
  
  return ctx;
}

- (NSInteger)numberOfPagesInPaperContext:(SafeFilePaperContext *)paperContext
{
  return paperContext->nPages;
}

#ifdef __IPHONE_OS_VERSION_MAX_ALLOWED
#define Color UIColor
#else
#define Color NSColor
#endif

static void drawText(CGContextRef context, CGPoint pos, CTFontRef font, NSString *text)
{
  NSAttributedString *attStr = [[NSAttributedString alloc]initWithString:text attributes:@{ (__bridge id)kCTFontAttributeName: (__bridge id)font }];
  CTLineRef line = CTLineCreateWithAttributedString((__bridge CFAttributedStringRef)attStr);
  CGContextSetTextPosition(context, pos.x, pos.y);
  CTLineDraw(line, context);
  CFRelease(line);
}

CGSize textSize(CTFontRef font, NSString *text)
{
  NSAttributedString *attStr = [[NSAttributedString alloc]initWithString:text attributes:@{ (__bridge id)kCTFontAttributeName: (__bridge id)font }];
  CTLineRef line = CTLineCreateWithAttributedString((__bridge CFAttributedStringRef)attStr);
  CGSize size;
  CGFloat a, d, l;
  size.width = CTLineGetTypographicBounds(line, &a, &d, &l);
  size.height = a + d;
  CFRelease(line);
  
  return size;
}

static void drawHeader(SafeFilePaperContext *paperContext, CGContextRef context, NSInteger page)
{
  CGRect effPrintableRect = paperContext->printableRect;
  CGSize size = effPrintableRect.size;
  CGFloat marginBottom = effPrintableRect.origin.y;
  CGFloat marginLeft = effPrintableRect.origin.x;
  CGFloat marginTopY = size.height + marginBottom;
  
  // Rectangle
  CGRect topRect = CGRectMake(marginLeft, marginTopY - SAFEFILEPDF_HEADER_H, size.width, SAFEFILEPDF_HEADER_H);
  CGContextSetStrokeColorWithColor(context, [Color blackColor].CGColor);
  CGContextSetFillColorWithColor(context, [Color clearColor].CGColor);
  CGContextStrokeRectWithWidth(context, topRect, 1.0f);
  
#ifdef PWSAFE_COZI
  UIImage *logo = [UIImage imageNamed:@"CoziVaultLogo@2x"];
  CGSize originalSize = logo.size;
  CGFloat logoH = topRect.size.height - 2 * SAFEFILEPDF_LOGO_MARGIN;
  CGFloat logoW = originalSize.width / originalSize.height * logoH;
  CGRect logoRect = CGRectMake(topRect.origin.x + 6.0f, topRect.origin.y + 6.0f, logoW, logoH);
  CGContextDrawImage (context, logoRect, logo.CGImage);
#else
  // Logo
  CGPDFDocumentRef elephant = CGPDFDocumentCreateWithURL((__bridge CFURLRef)[[NSBundle mainBundle]URLForResource:@"ElefantePrint" withExtension:@"pdf"]);
  CGPDFPageRef elephantPage = CGPDFDocumentGetPage(elephant, 1);
  CGContextSaveGState(context);
  CGSize elephantOriginalSize = CGPDFPageGetBoxRect(elephantPage, kCGPDFCropBox).size;
  CGFloat elephantH = topRect.size.height - 2 * SAFEFILEPDF_LOGO_MARGIN;
  CGFloat elephantW = elephantOriginalSize.width / elephantOriginalSize.height * elephantH;
  CGRect elephantRect = CGRectMake(topRect.origin.x + 6.0f, topRect.origin.y + 6.0f, elephantW, elephantH);
  CGAffineTransform pdfTransform = CGPDFPageGetDrawingTransform(elephantPage, kCGPDFCropBox, elephantRect, 0, YES);
  CGContextConcatCTM(context, pdfTransform);
  CGContextDrawPDFPage(context, elephantPage);
  CGContextRestoreGState(context);
  CFRelease(elephant);
#endif
  
  // Safe Name
  NSString *headerTxt = paperContext->safeName;
  CGSize headerSize = textSize(paperContext->headerFont, headerTxt);
  CGFloat headerDescent = CTFontGetDescent(paperContext->headerFont);
  CGPoint titlePos = CGPointMake(marginLeft + (size.width - headerSize.width) / 2, marginTopY - headerSize.height + headerDescent - (SAFEFILEPDF_HEADER_H - headerSize.height) / 2);
  drawText(context, titlePos, paperContext->headerFont, headerTxt);

  // Page number
  NSString *pageTxt = [NSString stringWithFormat:@"Page %ld", (long)(page + 1)];
  CGSize pageTxtSize = textSize(paperContext->headerInfoFont, pageTxt);
  CGPoint pagePos = CGPointMake(topRect.origin.x + topRect.size.width - SAFEFILEPDF_LOGO_MARGIN - pageTxtSize.width, topRect.origin.y + SAFEFILEPDF_LOGO_MARGIN);
  drawText(context, pagePos, paperContext->headerInfoFont, pageTxt);
  
  // Date
  NSString *dateTxt = paperContext->dateString;
  CGSize dateTxtSize = textSize(paperContext->headerInfoFont, dateTxt);
  CGFloat descent = CTFontGetDescent(paperContext->headerInfoFont);
  CGPoint datePos = CGPointMake(topRect.origin.x + topRect.size.width - SAFEFILEPDF_LOGO_MARGIN - dateTxtSize.width, marginTopY - SAFEFILEPDF_LOGO_MARGIN - dateTxtSize.height + descent);
  drawText(context, datePos, paperContext->headerInfoFont, dateTxt);
}

- (void)drawPage:(NSInteger)page fromPaperContext:(SafeFilePaperContext *)paperContext inGraphicsContext:(CGContextRef)context
{
  CGContextSaveGState(context);
  
  drawHeader(paperContext, context, page);

  CGRect effPrintableRect = paperContext->printableRect;
  CGFloat groupLineH = CTFontGetAscent(paperContext->groupFont) + CTFontGetDescent(paperContext->groupFont) + CTFontGetLeading(paperContext->groupFont);
  CGFloat groupDescent = CTFontGetDescent(paperContext->groupFont);
  CGFloat titleDescent = CTFontGetDescent(paperContext->titleFont);
  CGFloat descent = CTFontGetDescent(paperContext->font);
  CGFloat lineH = CTFontGetAscent(paperContext->font) + CTFontGetDescent(paperContext->font) + CTFontGetLeading(paperContext->font);
  CGFloat titleLineH = CTFontGetAscent(paperContext->titleFont) + CTFontGetDescent(paperContext->titleFont) + CTFontGetLeading(paperContext->titleFont);
  // Draw groups and entries
  CGFloat marginLeft = effPrintableRect.origin.x;
  CGFloat y = effPrintableRect.origin.y + effPrintableRect.size.height - SAFEFILEPDF_HEADER_H - SAFEFILEPDF_HEADER_V_MARGIN;
  CGFloat minY = effPrintableRect.origin.y;
  
  for(SafeFilePaperGroup *group in paperContext->groups)
  {
    if(group->page < page && (!group->entries.count || ((SafeFilePaperEntry *)group->entries.lastObject)->page < page))
      continue;
    if(group->page > page)
      continue;
    
    CGFloat xOffset = 0;
    if(group->groupName && group->groupName.length)
    {
      y -= SAFEFILEPDF_ENTRY_MARGIN;
      CGFloat groupTitleH = groupLineH + 2 * SAFEFILEPDF_GROUP_TITLE_BOX_PAD;
      if(y - groupTitleH < minY)
        break;
      
      NSArray *groupPath = [group->groupName componentsSeparatedByString:@"."];
      NSString *groupName = groupPath.lastObject;
      NSUInteger nIndents = groupPath.count > 1 ? groupPath.count - 1 : 0;
      
      xOffset = nIndents * SAFEFILEPDF_GROUP_CHILD_OFFSET;
      
      // Draw the group title rectangle
      CGRect groupRect = CGRectMake(marginLeft + xOffset, y - groupLineH - 2 * SAFEFILEPDF_GROUP_TITLE_BOX_PAD, effPrintableRect.size.width - xOffset, groupLineH + 2 * SAFEFILEPDF_GROUP_TITLE_BOX_PAD);
      CGContextSetFillColorWithColor(context, [Color darkGrayColor].CGColor);
      CGContextFillRect(context, groupRect);
      
      // Draw white text over it
      NSAttributedString *attStr = [[NSAttributedString alloc]initWithString:groupName
                                                                  attributes:@{
                                      (__bridge id)kCTFontAttributeName: (__bridge id)paperContext->groupFont,
                                      (__bridge id)kCTForegroundColorAttributeName: (__bridge id)[Color whiteColor].CGColor
                                    }];
      CTLineRef line = CTLineCreateWithAttributedString((__bridge CFAttributedStringRef)attStr);
      CGContextSetTextPosition(context, SAFEFILEPDF_GROUP_TITLE_BOX_PAD + marginLeft + xOffset, y - (SAFEFILEPDF_GROUP_TITLE_BOX_PAD + groupLineH - groupDescent));
      CTLineDraw(line, context);
      CFRelease(line);
      
      y -= groupTitleH;
    }
    
    BOOL firstEntry = YES;

    for(SafeFilePaperEntry *entry in group->entries)
    {
      if(entry->page > page)
        break;
      if(entry->page < page)
        continue;

      y -= SAFEFILEPDF_ENTRY_MARGIN;
      if(firstEntry)
        firstEntry = NO;
      else
      {
        CGContextSetLineCap(context, kCGLineCapSquare);
        CGContextSetStrokeColorWithColor(context, [Color blackColor].CGColor);
        CGContextSetLineWidth(context, 0.5f);
        CGFloat lineY = y + SAFEFILEPDF_ENTRY_MARGIN / 3;
        CGContextMoveToPoint(context, marginLeft + xOffset, lineY);
        CGContextAddLineToPoint(context, effPrintableRect.origin.x + effPrintableRect.size.width, lineY);
        CGContextStrokePath(context);
      }
      
      // First line: TITLE  emai   user
      CGFloat baseLineY = y - titleLineH + titleDescent;
      CGFloat xDataOffset = xOffset + SAFEFILEPDF_GROUP_CHILD_OFFSET;
      
      if(entry->title && entry->title.length)
      {
        drawText(context, CGPointMake(xOffset + marginLeft, baseLineY), paperContext->titleFont, entry->title);
      }
      if(entry->email && entry->email.length)
      {
        CGFloat emailW = textSize(paperContext->font, entry->email).width;
        CGFloat identedW = effPrintableRect.size.width - xDataOffset;
        drawText(context, CGPointMake(marginLeft + xDataOffset + (identedW - emailW) / 2, baseLineY), paperContext->font, entry->email);
      }
      if(entry->username && entry->username.length)
      {
        CGFloat userW = textSize(paperContext->font, entry->username).width;
        drawText(context, CGPointMake(effPrintableRect.origin.x + effPrintableRect.size.width - userW, baseLineY), paperContext->font, entry->username);
      }

      y -= titleLineH;
      
      // Second line: URL   password
      
      baseLineY = y - lineH + descent;
      
      if(entry->url && entry->url.length)
      {
        drawText(context, CGPointMake(xDataOffset + marginLeft, baseLineY), paperContext->font, entry->url);
      }
      if(entry->password && entry->password.length)
      {
        CGFloat passW = textSize(paperContext->passwordFont, entry->password).width;
        drawText(context, CGPointMake(effPrintableRect.origin.x + effPrintableRect.size.width - passW, baseLineY), paperContext->passwordFont, entry->password);
      }
      y -= lineH;
      
      // Notes
      if(entry->notes && entry->notes.length)
      {
        NSString *notes = [entry->notes stringByReplacingOccurrencesOfString:@"\r\n" withString:@"\n"];
        NSAttributedString *aStr = [[NSAttributedString alloc]initWithString:notes attributes:@{ (__bridge id)kCTFontAttributeName: (__bridge id)paperContext->notesFont }];
        CTFramesetterRef frameSetter = CTFramesetterCreateWithAttributedString((__bridge CFAttributedStringRef)aStr);
        CFRange txtRange;
        CFRange range = CFRangeMake(0, aStr.length);
        CGSize size = CTFramesetterSuggestFrameSizeWithConstraints(frameSetter, range, NULL, CGSizeMake(effPrintableRect.size.width - xDataOffset, CGFLOAT_MAX), &txtRange);
        CGRect rect = CGRectMake(marginLeft + xDataOffset, y - size.height, effPrintableRect.size.width - xDataOffset, size.height);
        CGPathRef path = CGPathCreateWithRect(rect, NULL);
        CTFrameRef frame = CTFramesetterCreateFrame(frameSetter, range, path, NULL);
        CFRelease(path);
        CFRelease(frameSetter);
        CTFrameDraw(frame, context);
        CFRelease(frame);
        
        y -= size.height;
      }
    }
  }
  
  
  CGContextRestoreGState(context);
}

@end


#ifdef __IPHONE_OS_VERSION_MAX_ALLOWED

@implementation SafeFilePageRenderer
{
  SafeFile *_file;
  SafeFilePaperContext *_context;
  NSString *_name;
}

- (id)initWithSafeFile:(SafeFile *)safeFile name:(NSString *)friendlyName
{
  if(!(self = [super init]))
    return nil;
  
  _file = safeFile;
  _name = friendlyName;
  
  return self;
}

- (void)drawContentForPageAtIndex:(NSInteger)pageIndex inRect:(CGRect)contentRect
{
  CGContextRef context = UIGraphicsGetCurrentContext();
  CGContextSaveGState(context);
  
  // Transform to translate to the Quartz coordinate system
  CGSize paperSize = self.paperRect.size;
  CGAffineTransform flipVertical = CGAffineTransformMake(1, 0, 0, -1, 0, paperSize.height);

  if(!_context)
  {
    CGRect printableRect = CGRectApplyAffineTransform(self.printableRect, flipVertical);
    _context = [_file createPaperContextForPaperSize:paperSize name:_name printableRect:printableRect];
  }
  
  CGContextConcatCTM(context, flipVertical);
  [_file drawPage:pageIndex fromPaperContext:_context inGraphicsContext:context];
  
  CGContextRestoreGState(context);
}

- (NSInteger)numberOfPages
{
  if(!_context)
    return -1;
  return [_file numberOfPagesInPaperContext:_context];
}

@end

#else

@implementation SafeFilePrintView
{
  SafeFile *_file;
  SafeFilePaperContext *_context;
  NSInteger _page;
}

- (id)initWithSafeFile:(SafeFile *)safeFile
{
  if(!(self = [super init]))
    return nil;
  
  _file = safeFile;
  
  return self;
}

- (BOOL)knowsPageRange:(NSRangePointer)range
{
  NSPrintOperation *printOp = [NSPrintOperation currentOperation];
  NSPrintInfo *printInfo = [printOp printInfo];
  
  NSSize paperSize = printInfo.paperSize;
  NSRect printableRect = printInfo.imageablePageBounds;

  if(!_context || !CGSizeEqualToSize(paperSize, _context->paperSize) || !CGRectEqualToRect(_context->printableRect, printableRect))
    _context = [_file createPaperContextForPaperSize:paperSize name:_file.safePath.lastPathComponent.stringByDeletingPathExtension printableRect:printableRect];
  
  NSInteger nPages = [_file numberOfPagesInPaperContext:_context];
  
  NSRect frame;
  frame.size = paperSize;
  frame.size.width *= nPages;
  frame.origin = NSZeroPoint;
  self.frame = frame;

  range->length = nPages;
  range->location = 1;
  
  return YES;
}

- (NSRect)rectForPage:(NSInteger)page
{
  NSAssert(page >= 1, @"Pages are 1 based");
  NSRect rect;
  rect.size = _context->paperSize;
  rect.origin.y = 0;
  rect.origin.x = rect.size.width * (page - 1);
  
  return rect;
}

- (void)drawRect:(NSRect)dirtyRect
{
  NSInteger page = dirtyRect.origin.x / _context->paperSize.width; // 0 based
  NSGraphicsContext *nsContext = [NSGraphicsContext currentContext];
  CGContextRef context = nsContext.graphicsPort;
  CGContextSaveGState(context);
  CGContextTranslateCTM(context, dirtyRect.origin.x, 0);
  [_file drawPage:page fromPaperContext:_context inGraphicsContext:context];
  CGContextRestoreGState(context);
}

@end

#endif