//
//  File.h
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/7/11.
//  Copyright 2011 app77. All rights reserved.
//

#import <Foundation/Foundation.h>

@class Entry;
@protocol SafeFileDelegate;

extern NSString * const kPasswordSafeStatusCodeKey;
extern NSString * const kPasswordSafeErrorDomain;
extern NSNumber * const kWrongPasswordStatusCode;

@interface SafeFile : NSObject 

@property (nonatomic, readonly) NSMutableDictionary *entries;
@property (nonatomic, readonly, strong) NSArray *sortedGroups;
@property (nonatomic, unsafe_unretained) id <SafeFileDelegate> delegate;
@property (atomic, readonly, strong) NSString *safePath;
@property (nonatomic, strong) NSData *safeUUID;
@property (nonatomic, strong, readonly) NSError* lastWriteError;
@property (nonatomic, readonly, strong) NSArray *namedPwPoliciesNames;
@property (nonatomic, readonly) BOOL defaultPwPolicyNumericOnly;
@property (nonatomic) BOOL defaultKeepPwHistory;
@property (nonatomic) uint defaultPwHistoryEntries;
@property (nonatomic, strong) NSString *defaultUsername;

+ (void)verifyPassword:(NSString *)password ofFilePath:(NSString *)filePath withCallback:(void (^)(BOOL ok, NSError *error))callback;
+ (void)internalVerifyPassword:(NSString *)password ofFilePath:(NSString *)filePath withCallback:(void (^)(BOOL ok, NSError *error))callback;

- (id) initWithPassword:(NSString *)password;
- (id) initWithFilePath:(NSString *)path password:(NSString *)password;
- (void) deleteEntry:(Entry *)entry;
- (void) editEntry:(Entry *)oldEntry newEntry:(Entry *)newEntry;
- (void) addEntry:(Entry *)newEntry;
- (void) addGroup:(NSString *)newGroup;
- (NSArray *) deleteGroup:(NSString *)groupName;
- (void) changePassword:(NSString *)newPassword;
- (void) renameGroupFrom:(NSString *)currentGroupName to:(NSString *)newGroupName;
- (void) enablePwHistoryOnAllEntries:(BOOL)enable withMaxEntries:(uint)maxEntries;
- (void) mergeChangesFromOtherFile:(SafeFile *)otherFile writeNow:(BOOL)writeNow;

- (void) writeSafe;

- (void) beginMultipleChanges;
- (void) endMultipeChanges:(BOOL)writeNow;

- (void) waitForSavingComplete;

- (Entry *)baseEntryForEntry:(Entry *)entry;
- (void) newRandomSafeUUID;

- (void) updateFilePath:(NSString *)newFilePath;
- (void) setBackupFileName:(NSString *)backupFilePath;

typedef void (^SafeFileImportCompletionBlock)(BOOL ok, NSString *report, NSError *error);

/// \param separator is the field separator
/// \param delimiter is the line delimiter for the notes field and the substitution for dots on the title field
- (void) importTextFile:(NSString *)textFilePath separator:(unichar)separator delimiter:(unichar)delimiter completion:(SafeFileImportCompletionBlock)block;

typedef void (^SafeFileExportCompletionBlock)(BOOL ok, NSError *error);

/// \param delimiter is the line delimiter for the notes field and the substitution for dots on the title field
- (void) exportTextFile:(NSString *)textFilePath delimiter:(unichar)delimiter completion:(SafeFileExportCompletionBlock)block;

// Protected

- (BOOL)loadSafeError:(NSError **)error;
- (void)internalWriteSafe;

#ifdef __cplusplus
class PWScore;
- (PWScore &)pwsCore;

class PWPolicy;
- (void)pwPolicy:(PWPolicy &)policy named:(NSString *)name;
- (void)addPwPolicy:(const PWPolicy &)policy named:(NSString *)name;
- (BOOL)deletePwPolicyNamed:(NSString *)policyName;
- (void)updatePwPolicy:(const PWPolicy &)policy named:(NSString *)policyName;
- (void)defaultPwPolicy:(PWPolicy &)policy;

#endif

@end

@protocol SafeFileDelegate <NSObject>

@required
- (void) safeFileChangesWrittenToDisk:(SafeFile *)file;

@optional
- (void) safeFileWillBeModifiedExternally:(SafeFile *)file;
- (void) safeFileWasModifiedExternally:(SafeFile *)file;
- (void) safeFileMustBeClosed:(SafeFile *)file;

@end




