//
//  File.mm
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/7/11.
//  Copyright 2011 app77. All rights reserved.
//

#import "SafeFile.h"

#include "core/PWScore.h"
#include "core/Command.h"
#include "core/PWSrand.h"

#import "Entry.h"
#import "NSStringUtils.h"
#import "NSArray+Find.h"

#include <map>
#include <memory>

#include <sys/stat.h>
#include <unistd.h>
#include <sys/stat.h>

NSString * const kPasswordSafeStatusCodeKey = @"PasswordSafeStatusCodeKey";
NSString * const kPasswordSafeErrorDomain = @"Password Safe";
NSNumber * const kWrongPasswordStatusCode = @(PWScore::WRONG_PASSWORD);

@implementation SafeFile
{
  NSMutableDictionary *m_entries;  //!< Dictionary containing NSMutableArray of groups. Entries with no groups are on a "nil" key.
  NSArray *m_sortedGroups;
  PWScore m_core;
  __strong NSString *m_password;
  __strong NSString *m_path;
  StringX m_wPassword;
  StringX m_wPath;
  __weak id <SafeFileDelegate> m_delegate;
  BOOL m_writePending;
  std::map<const CItemData *, Entry *> m_entriesDict;
  BOOL m_multipleChanges;
  dispatch_group_t m_writeDispatchGroup;
  NSError *m_lastWriteError;
  NSString *m_backupFileName;
}

@synthesize entries = m_entries;

- (PWScore &)pwsCore
{
  return m_core;
}

- (NSString *)safePath
{
  NSString *retval = nil;
  @synchronized(self) {
    retval = m_path;
  }
  return retval;
}

static BOOL setStatus(int status, NSError **error)
{
  NSString *lastError = nil;
  
  switch (status) 
  {
    case PWScore::SUCCESS:
    case PWScore::OK_WITH_ERRORS:
    case PWScore::OK_WITH_VALIDATION_ERRORS:
      lastError = nil;
      break;
    case PWScore::FAILURE: lastError = @"Unspecified failure"; break;
    case PWScore::CANT_OPEN_FILE: lastError = @"Can't open file"; break;
    case PWScore::USER_CANCEL: lastError = @"Canceled by user"; break;
    case PWScore::WRONG_PASSWORD: lastError = @"Wrong password"; break;
    case PWScore::BAD_DIGEST: lastError = @"Bad digest"; break;
    case PWScore::TRUNCATED_FILE: lastError = @"Truncated file"; break;
    case PWScore::READ_FAIL: lastError = @"Error reading file"; break;
    case PWScore::WRITE_FAIL: lastError = @"Error writing file"; break;
    case PWScore::UNKNOWN_VERSION: lastError = @"Unknown file version"; break;
    case PWScore::INVALID_FORMAT: lastError = @"Invalid file format"; break;
    case PWScore::USER_EXIT: lastError = @"User exited"; break;
    case PWScore::XML_FAILED_VALIDATION: lastError = @"XML failed validation";  break;
    case PWScore::XML_FAILED_IMPORT: lastError = @"XML failed import";  break;
    case PWScore::LIMIT_REACHED: lastError = @"Limit reached"; break;
    case PWScore::UNIMPLEMENTED: lastError = @"unimplemented"; break;
    case PWScore::NO_ENTRIES_EXPORTED: lastError = @"No entries exported"; break;
    default: lastError = [NSString stringWithFormat:@"Unknown error %d", status]; break;
  }
  
  if(!lastError)
    return YES;
  
  if(error)
  {
    NSDictionary *errorDict = @{NSLocalizedDescriptionKey: lastError, kPasswordSafeStatusCodeKey : @(status)};
    *error = [NSError errorWithDomain:kPasswordSafeErrorDomain code:-1 userInfo:errorDict];
  }
  
  return NO;
}

- (BOOL)loadSafeError:(NSError **)error
{
  NSAssert(dispatch_get_current_queue() != dispatch_get_main_queue(), @"Should not run on main queue");
  DebugLog(@"Will enter sync block for reading");
  @synchronized(self)
  {
    DebugLog(@"Entered sync block for reading");
    int pwst = m_core.CheckPasskey(m_wPath, m_wPassword);
    if(!setStatus(pwst, error))
      return NO;
    
    m_core.SetCurFile(m_wPath);
    
    CReport report;
    pwst = m_core.ReadCurFile(m_wPassword, true, 0, &report);
    if(!setStatus(pwst, error))
      return NO;
    
    if(pwst == PWScore::OK_WITH_VALIDATION_ERRORS)
    {
#ifdef TESTFLIGHT_SDK_VERSION
      A77Log(@"Validation errors");
#ifdef DEBUG_MODE
      NSLog(@"Report:\n%@", [NSString stringWithwstring:report.GetString()]);
#endif
#else
      NSLog(@"Non-fatal errors detected when opening safe %@\n%@", m_path.lastPathComponent, [NSString stringWithwstring:report.GetString()]);
#endif
    }

    [self makeModel];
    
    return YES;
  }
}

- (id) initWithFilePath:(NSString *)path password:(NSString *)password
{
  if(!(self = [super init]))
    return nil;
  
  m_password = password ? [[NSString stringWithString:password] copy] : nil;
  m_path = [[NSString stringWithString:path] copy];
  m_wPassword = password ? password.wstring : StringX();
  m_wPath = path.wstring;
  m_delegate = nil;
  m_writePending = NO;
  
  m_writeDispatchGroup = dispatch_group_create();
  
  return self;
}

- (id) initWithPassword:(NSString *)password
{
  if(!(self = [super init]))
    return nil;
  
  m_password = password ? [[NSString stringWithString:password] copy] : nil;
  m_wPassword = password ? password.wstring : StringX();
  m_delegate = nil;
  m_writePending = NO;
  m_writeDispatchGroup = dispatch_group_create();
  
  // Create the root group
  m_entries = [[NSMutableDictionary alloc] init];
  m_entries[@""] = [[NSMutableArray alloc]init];
  
  m_core.NewFile(m_wPassword);
  
  return self;
}

- (void)dealloc
{
  dispatch_release(m_writeDispatchGroup);
  DebugLog(@"-[SafeFile dealloc]");
}

- (void)makeModel
{
  // Populate entries
  if(m_entries)
    [m_entries removeAllObjects];
  else
    m_entries = [[NSMutableDictionary alloc] init];
  m_entriesDict.clear();
  
  // Create the root group
  m_entries[@""] = [[NSMutableArray alloc]init];
  
  for(ItemListConstIter it = m_core.GetEntryIter(); it != m_core.GetEntryEndIter(); it++)
  {
    const CItemData &item = m_core.GetEntry(it);
    Entry *entry = [[Entry alloc] initWithItemData:item file:self];
    m_entriesDict.insert(std::make_pair(&item, entry));
    
    NSMutableArray *group = nil;
    NSArray *groups = [entry.group componentsSeparatedByString:@"."];
    NSMutableString *groupName = nil;
    for(NSString *subGroupName in groups)
    {
      if(!groupName)
        groupName = [NSMutableString stringWithString:subGroupName];
      else
        [groupName appendFormat:@".%@", subGroupName];
      
      group = m_entries[groupName];
      if(group)
        continue;
      group = [[NSMutableArray alloc]init];
      [m_entries setValue:group forKey:groupName];
    }
    
    NSAssert([groupName isEqual:entry.group], @"Must have reconstructed group");
    NSAssert(group, @"At this point, group should be created");
    NSAssert(group == m_entries[entry.group], @"Last group is entry's group");
    
    [group addObject:entry];
  }
  
  // Load empty groups
  const auto emptyGroups = m_core.GetEmptyGroups();
  for(auto it = emptyGroups.begin(); it != emptyGroups.end(); it++)
  {
    NSString *groupName = [NSString stringWithwstring:*it];
    if(m_entries[groupName] == nil)
      m_entries[groupName] = [[NSMutableArray alloc]init];
  }
  
  for(NSMutableArray *entries in m_entries.allValues)
    [entries sortUsingSelector:@selector(caseInsensitiveCompare:)];
  
  m_sortedGroups = nil;
}

- (void)updateFilePath:(NSString *)newFilePath
{
  @synchronized(self) {
    m_wPath = newFilePath.wstring;
    m_path = [newFilePath copy];
  }
}

- (NSArray *)namedPwPoliciesNames
{
  std::vector<stringT> names;
  @synchronized(self) {
    m_core.GetPolicyNames(names);
  }
  if(names.empty())
    return @[];
  NSMutableArray *ret = [NSMutableArray arrayWithCapacity:names.size()];
  for(const auto &name : names)
    [ret addObject:[NSString stringWithWstring:name]];
  return ret;
}

- (void)pwPolicy:(PWPolicy &)policy named:(NSString *)name
{
  @synchronized(self) {
    const auto &policies = m_core.GetPasswordPolicies();
    auto it = policies.find(name.wstring);
    NSAssert(it != policies.end(), @"Policy must exist");
    const auto &pol = it->second;
    policy = pol;
    policy.usecount = pol.usecount;
  }
}

- (void)defaultPwPolicy:(PWPolicy &)policy
{
  PWSprefs *prefs = new PWSprefs;
  @synchronized(self) {
    prefs->Load(m_core.GetDBPreferences());
  }
  policy = prefs->GetDefaultPolicy();
  delete prefs;
}

- (BOOL)defaultPwPolicyNumericOnly
{
  PWPolicy pol;
  [self defaultPwPolicy:pol];
  return (pol.flags & (PWPolicy::UseLowercase |
                       PWPolicy::UseUppercase |
                       PWPolicy::UseDigits |
                       PWPolicy::UseSymbols |
                       PWPolicy::UseHexDigits |
                       PWPolicy::UseEasyVision |
                       PWPolicy::MakePronounceable)) == PWPolicy::UseDigits;
}

#pragma mark - Changes

- (id<SafeFileDelegate>) delegate
{
  return m_delegate;
}

- (void) setDelegate:(id<SafeFileDelegate>) newDelegate
{
  m_delegate = newDelegate;
}

static BOOL replaceFileManually(NSURL *pathURL, NSURL *tmpURL, NSString *backupFileName, NSError **error)
{
  NSFileManager *fm = [NSFileManager defaultManager];
  NSURL *baseURL = pathURL.URLByDeletingLastPathComponent;
  NSURL *backupURL = [baseURL URLByAppendingPathComponent:backupFileName isDirectory:NO];
  
  if([fm fileExistsAtPath:backupURL.path] && ![fm removeItemAtURL:backupURL error:error])
  {
    DebugLog(@"Deleting of old backup file failed: %@", *error);
    return NO;
  }
  
  BOOL existedBefore = [fm fileExistsAtPath:pathURL.path];
  if(existedBefore && ![fm moveItemAtURL:pathURL toURL:backupURL error:error])
  {
    DebugLog(@"Moving old file to backup failed %@", *error);
    return NO;
  }
  
  if(![fm moveItemAtURL:tmpURL toURL:pathURL error:error])
  {
    NSError *secondaryError = nil;
    if(existedBefore && ![fm moveItemAtURL:backupURL toURL:pathURL error:&secondaryError])
      DebugLog(@"Failed to move backup back into place: %@, ", secondaryError);
    DebugLog(@"Failed to move temp file into place: %@", *error);
    return NO;
  }
  
  DebugLog(@"replaceFileManually succeeded");
  return YES;
}

extern "C" BOOL hideFile(NSString *filePath)
{
  const char *c_filePath = filePath.UTF8String;
  
  struct stat st;
  if(0 != ::stat(c_filePath, &st))
    return NO;
  if(st.st_flags & UF_HIDDEN)
    return YES;
  
  return 0 == ::chflags(c_filePath, st.st_flags | UF_HIDDEN);
}

- (void)internalWriteSafe
{
  @autoreleasepool 
  {
    @synchronized(self)
    {
      m_lastWriteError = nil;
      NSFileManager *fm = [NSFileManager defaultManager];
      NSError *error = nil;
      
      if(![fm fileExistsAtPath:m_path])
      {
        int ret = m_core.WriteFile(m_wPath);
        if(ret != 0)
          [fm removeItemAtPath:m_path error:nil];
        else
        {
          error = nil;
          setStatus(ret, &error);
          m_lastWriteError = error;
        }
        m_writePending = NO;     
        return;
      }
      NSURL *pathURL = [NSURL fileURLWithPath:m_path];

      error = nil;
      NSURL *tempDirURL = [fm URLForDirectory:NSItemReplacementDirectory inDomain:NSUserDomainMask appropriateForURL:pathURL create:YES error:&error];
      if(!tempDirURL)
      {
        m_lastWriteError = error;
        m_writePending = NO;
        return;
      }
      NSString *fileName = m_path.lastPathComponent;
      NSString *tmpFileName = [fileName stringByAppendingPathExtension:@"tmp"];
      NSURL *tmpURL = [tempDirURL URLByAppendingPathComponent:tmpFileName isDirectory:NO];
      NSString *backupFileName;
      NSFileManagerItemReplacementOptions options;
      
#ifdef __IPHONE_OS_VERSION_MAX_ALLOWED
      backupFileName = [fileName stringByAppendingPathExtension:@"old"];
      options = NSFileManagerItemReplacementWithoutDeletingBackupItem;
#else
      if(m_backupFileName.length > 0)
      {
        backupFileName = m_backupFileName;
        options = NSFileManagerItemReplacementWithoutDeletingBackupItem;
      }
      else
      {
        backupFileName = [fileName stringByAppendingPathExtension:@"macOld"];
        options = 0;
      }
#endif
      DebugLog(@"fileName = %@, backupFileName = %@, tmpFileName = %@, nEntries = %lu", fileName, backupFileName, tmpURL.lastPathComponent, m_core.GetNumEntries());
      
      int ret = m_core.WriteFile(tmpURL.path.wstring);
      if(ret != 0)
      {
        DebugLog(@"m_core.WriteFile failed: %d", ret);
        error = nil;
        setStatus(ret, &error);
        m_lastWriteError = error;
        [fm removeItemAtURL:tmpURL error:nil];
      }
      else
      {
        NSURL *newURL = nil;
        error = nil;
        BOOL ok = [fm replaceItemAtURL:pathURL withItemAtURL:tmpURL backupItemName:backupFileName options:options resultingItemURL:&newURL error:&error];
        DebugLog(@"fm replaceItemAtURL result: %d", (int)ok);
        if(!ok)
        {
          if(error)
          {
            m_lastWriteError = error;
            [fm removeItemAtURL:tmpURL error:nil];
          }
          else
          {
#ifdef TESTFLIGHT_SDK_VERSION
            A77Log(@"replaceItemAtURL failed without error");
#endif
            if(!replaceFileManually(pathURL, tmpURL, backupFileName, &error))
              m_lastWriteError = error;
          }
        }
      }
      
      hideFile(backupFileName);
      
      [fm removeItemAtURL:tmpURL error:nil];
      BOOL ok = [fm removeItemAtURL:tempDirURL error:nil];
      (void)ok;
      NSAssert(ok, @"Better remove temp dir");
      
      m_writePending = NO;      
    }
    DebugLog(@"Write complete, m_lastWriteError = %@", m_lastWriteError);
    dispatch_async(dispatch_get_main_queue(), ^{
      BOOL writePending;
      @synchronized(self)
      {
        writePending = m_writePending;
      }
      if(writePending)
      {
        [self writeSafe];
        return;
      }
      [m_delegate safeFileChangesWrittenToDisk:self];
    });
  }
}

- (void)writeSafe
{
  if(m_multipleChanges || !m_path || !m_path.length)
  {
    A77Log(@"Will not write safe now due to %d / %p / %lu", m_multipleChanges, m_path, (u_long)m_path.length);
    return;
  }
  
  @synchronized(self)
  {
    // Adjust empty groups
    std::vector<StringX> newEmptyGroups;
    newEmptyGroups.reserve(m_entries.count);
    auto &newEmptyGroupsRef = newEmptyGroups;
    [m_entries enumerateKeysAndObjectsUsingBlock:^(NSString *groupName, NSArray *entries, BOOL *stop) {
      if(entries.count == 0)
        newEmptyGroupsRef.push_back(groupName.wstring);
    }];
    DBEmptyGroupsCommand *emptyGroups = DBEmptyGroupsCommand::Create(&m_core, newEmptyGroupsRef, DBEmptyGroupsCommand::EG_REPLACEALL);
    m_core.Execute(emptyGroups);
    m_core.ClearCommands();
    
    // Take care of multiple write commands
    if(m_writePending)
      return;
    m_writePending = YES;
  }
  
  dispatch_group_async(m_writeDispatchGroup, dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
    [self internalWriteSafe];
  });
}

- (void)setBackupFileName:(NSString *)backupFilePath
{
  m_backupFileName = backupFilePath ? [backupFilePath copy] : nil;
}

- (void)waitForSavingComplete
{
  dispatch_group_wait(m_writeDispatchGroup, DISPATCH_TIME_FOREVER);
}

- (NSError *)lastWriteError
{
  NSError *ret;
  @synchronized(self)
  {
    ret = m_lastWriteError;
  }
  return ret;
}

#pragma mark - Modification operations

- (void) changePassword:(NSString *)newPassword
{
  @synchronized(self)
  {
    m_core.ChangePasskey(newPassword.wstring);
  }
  [self writeSafe];
}

- (void)deleteEntry:(Entry *)entry
{
  m_sortedGroups = nil;
  
  NSMutableArray *group = m_entries[entry.group];
  [group removeObject:entry];  
  const CItemData *item = entry.cItemData;
  m_entriesDict.erase(item);
  
  DeleteEntryCommand *cmd = DeleteEntryCommand::Create(&m_core, *item);
  @synchronized(self)
  {
    m_core.Execute(cmd);
  }
  [self writeSafe];
}

- (void) editEntry:(Entry *)oldEntry newEntry:(Entry *)newEntry
{
  m_sortedGroups = nil;
  
  NSMutableArray *group = m_entries[oldEntry.group];
  NSAssert(group, @"Old entry must have a group");
  NSUInteger idx = group ? [group indexOfObjectIdenticalTo:oldEntry] : NSNotFound;
  NSAssert(idx != NSNotFound, @"Group must contain old entry");
  
  newEntry.mTime = [NSDate date];
  
  const CItemData *oldItem = oldEntry.cItemData;
  const CItemData *newItem = newEntry.cItemData;
  
  NSMutableArray *newGroup = [newEntry.group isEqualToString:oldEntry.group] ? group : m_entries[newEntry.group];
  if(newGroup == group && idx != NSNotFound)
  {
    group[idx] = newEntry;
    [group sortUsingSelector:@selector(caseInsensitiveCompare:)];
  }
  else
  {
    NSAssert(newGroup, @"New group must exist");
    if(idx != NSNotFound)
      [group removeObjectAtIndex:idx];
    
    if(newGroup)
    {
      [newGroup addObject:newEntry];
      [newGroup sortUsingSelector:@selector(caseInsensitiveCompare:)];
    }
  }
  m_entriesDict.erase(oldItem);
  m_entriesDict.insert(std::make_pair(newItem, newEntry));  

  EditEntryCommand *cmd = EditEntryCommand::Create(&m_core, *oldItem, *newItem);
  int ret;
  @synchronized(self)
  {
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"editEntry, m_core.Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
}

- (void) addEntry:(Entry *)newEntry
{
  m_sortedGroups = nil;
  
  newEntry.mTime = [NSDate date];
  
  const CItemData *newItem = newEntry.cItemData;
  AddEntryCommand *cmd = AddEntryCommand::Create(&m_core, *newItem);
  @synchronized(self)
  {
    m_core.Execute(cmd);
  }
  // Don't write file, it will be modified shortly
  
  NSString *groupName = newEntry.group;
  DebugLog(@"New entry group: 0x%p '%@' [%lu]", groupName, groupName, (unsigned long)groupName.length);
  NSMutableArray *group = m_entries[groupName];
  NSAssert(group, @"Must exist");
  [group addObject:newEntry];
  [group sortUsingSelector:@selector(caseInsensitiveCompare:)];
  m_entriesDict.insert(std::make_pair(newItem, newEntry));  
}

- (void) addGroup:(NSString *)newGroup
{
  m_sortedGroups = nil;
  NSAssert(m_entries[newGroup] == nil, @"Cannot add a group which already exists");
  m_entries[newGroup] = [[NSMutableArray alloc]init];
  [self writeSafe];
}


- (NSArray *) deleteGroup:(NSString *)parentGroupName
{
  m_sortedGroups = nil;
  
  NSMutableArray *groups = [NSMutableArray arrayWithObject:parentGroupName];
  NSString *prefix = [parentGroupName stringByAppendingString:@"."];
  
  for(NSString *childGroup in m_entries.allKeys)
    if([childGroup rangeOfString:prefix].location == 0)
      [groups addObject:childGroup];

  BOOL deleted = NO;
  
  @synchronized(self)
  {
    MultiCommands *multiCmd = MultiCommands::Create(&m_core);
    for(NSString *groupName in groups)
    {
      NSArray *group = m_entries[groupName];
      if(!group)
        continue;
      
      for(Entry *entry in group)
      {
        const CItemData *itemData = entry.cItemData;
        DeleteEntryCommand *cmd = DeleteEntryCommand::Create(&m_core, *itemData);
        multiCmd->Add(cmd);
        m_entriesDict.erase(itemData);
      }
      
      [m_entries removeObjectForKey:groupName];
      deleted = YES;
    }
    
    if(multiCmd->GetSize())
      m_core.Execute(multiCmd);
  }
  
  if(deleted)
    [self writeSafe];
  
  return groups;
}

- (void)renameGroupFrom:(NSString *)currentGroupName to:(NSString *)newGroupName
{
  m_sortedGroups = nil;
  
  NSMutableArray *groups = [NSMutableArray arrayWithObject:currentGroupName];
  NSString *prefix = [currentGroupName stringByAppendingString:@"."];
  NSUInteger prefixLen = prefix.length;
  NSString *newPrefix = [newGroupName stringByAppendingString:@"."];
  
  for(NSString *childGroup in m_entries.allKeys)
    if([childGroup rangeOfString:prefix].location == 0)
      [groups addObject:childGroup];

  BOOL modified = NO;
  
  @synchronized(self)
  {
    MultiCommands *multiCmd = MultiCommands::Create(&m_core);

    for(NSString *oldGroupName in groups)
    {
      NSArray *oldGroup = m_entries[oldGroupName];
      if(!oldGroup)
        continue;
      
      NSString *newName = [oldGroupName isEqualToString:currentGroupName] ? newGroupName : [newPrefix stringByAppendingString:[oldGroupName substringFromIndex:prefixLen]];
      StringX newNameS = newName.wstring;
      
      for(Entry *entry in oldGroup)
      {
        DebugLog(@"In entry [%@] changing group: [%@] -> [%@]", entry.title, entry.group, newName);

        const CItemData *oldItem = entry.cItemData;
        CItemData *newItem = new CItemData(*oldItem);
        newItem->SetGroup(newNameS);
        [entry setCItemData:newItem isConst:NO];
        
        EditEntryCommand *cmd = EditEntryCommand::Create(&m_core, *oldItem, *newItem);
        multiCmd->Add(cmd);
        
        m_entriesDict.erase(oldItem);
        m_entriesDict.insert(std::make_pair(newItem, entry));
      }
      m_entries[newName] = oldGroup;
      [m_entries removeObjectForKey:oldGroupName];
      modified = YES;
    }
    if(multiCmd->GetSize())
      m_core.Execute(multiCmd);
  }
  
  if(modified)
    [self writeSafe];  
}

- (void)beginMultipleChanges
{
  NSAssert(!m_multipleChanges, @"This class does not support recursive multiple changes");
  m_multipleChanges = YES;
}

- (void)endMultipeChanges:(BOOL)writeNow
{
  m_multipleChanges = NO;
  if(writeNow)
    [self writeSafe];
}

- (NSArray *)sortedGroups
{
  if(m_sortedGroups)
    return m_sortedGroups;
  m_sortedGroups = [[self.entries allKeys] sortedArrayUsingComparator:^NSComparisonResult(NSString *g1, NSString *g2) {
    NSArray *a1 = [g1 componentsSeparatedByString:@"."];
    NSArray *a2 = [g2 componentsSeparatedByString:@"."];

    NSUInteger i;
    for(i = 0; i < a1.count; i++)
    {
      if(a2.count <= i)
        return NSOrderedDescending;
      
      NSString *c1 = a1[i];
      NSString *c2 = a2[i];
      NSComparisonResult r = [c1 caseInsensitiveCompare:c2];
      if(r != NSOrderedSame)
        return r;
    }
    if(a2.count > i)
      return NSOrderedAscending;
    
    return NSOrderedSame;
  }];
  return m_sortedGroups;
}

- (NSString *)defaultUsername
{
#ifdef PWSPREFS_STATICS  
  PWSprefs *prefs = PWSprefs::GetInstance();
#else
  std::auto_ptr<PWSprefs> prefs (new PWSprefs);
  @synchronized(self) {
    prefs->Load(m_core.GetDBPreferences());
  }
#endif
  auto defaultUsername = prefs->GetPref(PWSprefs::DefaultUsername);
  if(defaultUsername.length())
    return [NSString stringWithwstring:defaultUsername];
  else
    return @"";
}

- (void)setDefaultUsername:(NSString *)defaultUsername
{
#ifdef PWSPREFS_STATICS
  PWSprefs *prefs = PWSprefs::GetInstance();
#else
  std::auto_ptr<PWSprefs> prefs (new PWSprefs);
  @synchronized(self) {
    prefs->Load(m_core.GetDBPreferences());
  }
#endif
  prefs->SetPref(PWSprefs::DefaultUsername, defaultUsername.wstring);
  StringX prefsStr(prefs->Store());
  auto *cmd = DBPrefsCommand::Create(&m_core, prefsStr);
  int ret;
  @synchronized(self)
  {
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"setDefaultUsername, m_core.Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
}

- (BOOL) defaultKeepPwHistory
{
#ifdef PWSPREFS_STATICS
  PWSprefs *prefs = PWSprefs::GetInstance();
#else
  std::auto_ptr<PWSprefs> prefs (new PWSprefs);
  @synchronized(self) {
    prefs->Load(m_core.GetDBPreferences());
  }
#endif
  bool keep = prefs->GetPref(PWSprefs::SavePasswordHistory);
  return !!keep;
}

- (void)setDefaultKeepPwHistory:(BOOL)defaultKeepPwHistory
{
#ifdef PWSPREFS_STATICS
  PWSprefs *prefs = PWSprefs::GetInstance();
#else
  std::auto_ptr<PWSprefs> prefs (new PWSprefs);
  @synchronized(self) {
    prefs->Load(m_core.GetDBPreferences());
  }
#endif
  prefs->SetPref(PWSprefs::SavePasswordHistory, defaultKeepPwHistory);
  StringX prefsStr(prefs->Store());
  auto *cmd = DBPrefsCommand::Create(&m_core, prefsStr);
  int ret;
  @synchronized(self)
  {
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"setDefaultKeepPwHistory, m_core.Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
}

- (uint) defaultPwHistoryEntries
{
#ifdef PWSPREFS_STATICS
  PWSprefs *prefs = PWSprefs::GetInstance();
#else
  std::auto_ptr<PWSprefs> prefs (new PWSprefs);
  @synchronized(self) {
    prefs->Load(m_core.GetDBPreferences());
  }
#endif
  unsigned val = prefs->GetPref(PWSprefs::NumPWHistoryDefault);
  return MIN(val, std::numeric_limits<u_int8_t>::max());
}

- (void)setDefaultPwHistoryEntries:(uint)defaultPwHistoryEntries
{
#ifdef PWSPREFS_STATICS
  PWSprefs *prefs = PWSprefs::GetInstance();
#else
  std::auto_ptr<PWSprefs> prefs (new PWSprefs);
  @synchronized(self) {
    prefs->Load(m_core.GetDBPreferences());
  }
#endif
  if(defaultPwHistoryEntries > std::numeric_limits<u_int8_t>::max())
    defaultPwHistoryEntries = std::numeric_limits<u_int8_t>::max();
  prefs->SetPref(PWSprefs::NumPWHistoryDefault, defaultPwHistoryEntries);
  StringX prefsStr(prefs->Store());
  auto *cmd = DBPrefsCommand::Create(&m_core, prefsStr);
  int ret;
  @synchronized(self)
  {
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"setDefaultKeepPwHistory, m_core.Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
}

- (void)enablePwHistoryOnAllEntries:(BOOL)enable withMaxEntries:(uint)maxEntries
{
  auto *cmd1 = UpdatePasswordHistoryCommand::Create(&m_core, enable ? -2 : -1, maxEntries);
  auto *cmd2 = UpdatePasswordHistoryCommand::Create(&m_core, -3, maxEntries);
  auto *cmd = MultiCommands::Create(&m_core);
  cmd->Add(cmd1);
  cmd->Add(cmd2);
  
  int ret;
  @synchronized(self)
  {
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"enablePwHistoryOnAllEntries, m_core. Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
}


-(Entry *)baseEntryForEntry:(Entry *)entry
{
  const CItemData *item = entry.cItemData;  
  CItemData::EntryType type = item->GetEntryType();
  if(type != CItemData::ET_ALIAS && type != CItemData::ET_SHORTCUT)
    return nil;
  const CItemData *baseItem = m_core.GetBaseEntry(item);
  auto it = m_entriesDict.find(baseItem);
  return it != m_entriesDict.end() ? it->second : nil;
}

- (NSData *)safeUUID
{
  NSMutableData *ret = [NSMutableData dataWithLength:sizeof(uuid_array_t)];
  uuid_array_t &uuid = *(uuid_array_t *)ret.mutableBytes;
  m_core.GetFileUUID().GetARep(uuid);
  return ret;
}

- (void)setSafeUUID:(NSData *)safeUUID
{
  NSAssert(safeUUID.length == sizeof(uuid_array_t), @"Wrong size");
  auto &uuid = *(const uuid_array_t *)safeUUID.bytes;
  m_core.SetFileUUID(uuid);
}

- (void) newRandomSafeUUID
{
  PWSrand *random = PWSrand::GetInstance();
  uuid_array_t uuid;
  random->GetRandomData(&uuid, sizeof(uuid));
  m_core.SetFileUUID(uuid);
}

- (void)addPwPolicy:(const PWPolicy &)policy named:(NSString *)name
{
  [self willChangeValueForKey:@"namedPwPoliciesNames"];
  auto cmd = DBPolicyNamesCommand::Create(&m_core, name.wstring, policy);
  
  int ret;
  @synchronized(self) {
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"addPwPolicy, m_core.Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
  [self didChangeValueForKey:@"namedPwPoliciesNames"];
}

- (BOOL)deletePwPolicyNamed:(NSString *)policyName
{
  PWPolicy policy;
  auto sPolicyName = policyName.wstring;
  int ret;
  @synchronized(self) {
    PSWDPolicyMap policies(m_core.GetPasswordPolicies());
    auto it = policies.find(sPolicyName);
    bool exists = it != policies.end();
    NSAssert(exists && policy.usecount == 0, @"Must be unused to delete");
    if(!exists || policy.usecount > 0)
      return NO;
    [self willChangeValueForKey:@"namedPwPoliciesNames"];
    policies.erase(it);
    auto cmd = DBPolicyNamesCommand::Create(&m_core, policies, DBPolicyNamesCommand::NP_REPLACEALL);
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"deletePwPolicyNamed, m_core.Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
  [self didChangeValueForKey:@"namedPwPoliciesNames"];
  
  return YES;
}

- (void)updatePwPolicy:(const PWPolicy &)policy named:(NSString *)policyName
{
  auto sPolicyName = policyName.wstring;
  int ret;
  @synchronized(self) {
    PSWDPolicyMap policies(m_core.GetPasswordPolicies());
    auto it = policies.find(sPolicyName);
    bool exists = it != policies.end();
    NSAssert(exists, @"Must exists to be updated");
    if(!exists)
      return;
    it->second = policy;
    auto cmd = DBPolicyNamesCommand::Create(&m_core, policies, DBPolicyNamesCommand::NP_REPLACEALL);
    ret = m_core.Execute(cmd);
  }
  DebugLog(@"updatePwPolicy, m_core.Execute = %d", ret);
  (void)ret;
  
  [self writeSafe];
}

#pragma mark - Password verification utility

+ (void)internalVerifyPassword:(NSString *)password ofFilePath:(NSString *)filePath withCallback:(void (^)(BOOL ok, NSError *error))callback
{
  PWSfile::VERSION version = PWSfile::UNKNOWN_VERSION;
  int ret = PWSfile::CheckPasskey(filePath.wstring, password.wstring, version);
  NSError *error = nil;
  BOOL ok = setStatus(ret, &error);
  dispatch_async(dispatch_get_main_queue(), ^{
    callback(ok, error);
  });
}

+ (void)verifyPassword:(NSString *)password ofFilePath:(NSString *)filePath withCallback:(void (^)(BOOL ok, NSError *error))callback
{
  callback = [callback copy];
  dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
    [self internalVerifyPassword:password ofFilePath:filePath withCallback:callback];
  });
}

#pragma mark - Merging

// This function ignores group changes on entries
- (void)mergeChangesFromOtherFile:(SafeFile *)otherFile writeNow:(BOOL)writeNow;
{
  [self beginMultipleChanges];
  
  NSMutableSet *myUUIDs = [NSMutableSet setWithCapacity:m_core.GetNumEntries()];
  
  // Look for new and modified entries
  [m_entries enumerateKeysAndObjectsUsingBlock:^(NSString *myGroupName, NSMutableArray *myGroup, BOOL *stop) {
    NSMutableArray *otherSameGroup = otherFile->m_entries[myGroupName];
    for(Entry *myEntry in [myGroup copy])
    {
      NSData *myUUID = myEntry.uuid;
      NSAssert(![myUUIDs containsObject:myUUID], @"Duplicate UUIDs");
      [myUUIDs addObject:myEntry.uuid];
      
      __block Entry *otherEntry;
      if(otherSameGroup)
        otherEntry = [otherSameGroup objectWithProperty:@"uuid" equalTo:myUUID];
      
      if(!otherEntry)
      {
        // Search in all groups
        [otherFile->m_entries enumerateKeysAndObjectsUsingBlock:^(NSString *otherGroupName, NSMutableArray *otherGroup, BOOL *stop) {
          if(otherGroup == otherSameGroup) // Skip
            return;
          otherEntry = [otherSameGroup objectWithProperty:@"uuid" equalTo:myUUID];
          *stop = otherEntry != nil;
        }];
      }
      
      if(!otherEntry)
        continue;
    
      // Compare
      BOOL changed = ![myEntry.title isEqualToString:otherEntry.title] ||
                      ![myEntry.username isEqualToString:otherEntry.username] ||
                      ![myEntry.url isEqualToString:otherEntry.url] ||
                      ![myEntry.email isEqualToString:otherEntry.email] ||
                      ![myEntry.notes isEqualToString:otherEntry.notes] ||
                      ![myEntry.pwPolicyName isEqualToString:otherEntry.pwPolicyName] ||
                       !myEntry.hasCustomPasswordPolicy != !otherEntry.hasCustomPasswordPolicy;
      
      if(!changed && myEntry.hasCustomPasswordPolicy)
      {
        // Compare pw policy
        PWPolicy myPol, otherPol;
        [myEntry getPasswordPolicy:myPol];
        [otherEntry getPasswordPolicy:otherPol];
        changed = myPol == otherPol;
      }
    
      if(!changed)
      {
        // Lastly, verify password
        if([myEntry.password isEqualToString:otherEntry.password])
          continue;
        time_t myPMTime, otherPMTime;
        myEntry.cItemData->GetPMTime(myPMTime);
        otherEntry.cItemData->GetPMTime(otherPMTime);
        
        if(!myPMTime || !otherPMTime)
        {
          changed = YES;
        }
        else if(myPMTime != otherPMTime)
        {
          Entry *newMyEntry = [[Entry alloc]initWithEntry:myEntry file:self];
          
          if(myPMTime < otherPMTime) // other is newer, replace mine
            newMyEntry.password = otherEntry.password;
          else                       // other is older, add to history
            [newMyEntry addPasswordToHistory:otherEntry.password withPMtime:otherPMTime];

          [self editEntry:myEntry newEntry:newMyEntry];
        }
      }
                 
      if(changed)
      {
        // See if entries have MTimes set
        NSDate *myMTime = myEntry.mTime;
        NSDate *otherMTime = otherEntry.mTime;
        
        if(myMTime && otherMTime)
        {
          NSComparisonResult order = [myMTime compare:otherMTime];
          if(order == NSOrderedAscending) // other is newer
          {
            Entry *newEntry = [[Entry alloc]initWithEntry:otherEntry file:self];
            [self editEntry:myEntry newEntry:newEntry];
          }
        }
        else
        {
          Entry *newEntry = [[Entry alloc]initWithEntry:otherEntry file:self];
          [newEntry generateNewUUID];
          newEntry.title = [newEntry.title stringByAppendingFormat:@" [conflicted %@]", [NSDateFormatter localizedStringFromDate:[NSDate date] dateStyle:NSDateFormatterShortStyle timeStyle:NSDateFormatterShortStyle]];
          [self addEntry:newEntry];
        }
      }
    }
  }];
  
  // Look for added entries
  [otherFile->m_entries enumerateKeysAndObjectsUsingBlock:^(NSString *otherGroupName, NSMutableArray *otherGroup, BOOL *stop) {
    NSMutableArray *mySameGroup = m_entries[otherGroupName];
    for(Entry *otherEntry in otherGroup)
    {
      NSData *otherUUID = otherEntry.uuid;
      if([myUUIDs containsObject:otherUUID])
        continue;
      __block NSUInteger myIdx = [mySameGroup indexOfObjectWithProperty:@"uuid" equalTo:otherUUID];
      if(myIdx == NSNotFound)
      {
        // Search in other groups
        [m_entries enumerateKeysAndObjectsUsingBlock:^(NSString *myGroupName, NSMutableArray *myGroup, BOOL *stop) {
          myIdx = [myGroup indexOfObjectWithProperty:@"uuid" equalTo:otherUUID];
          *stop = myIdx != NSNotFound;
        }];
      }
      if(myIdx == NSNotFound)
      {
        // Entry not found anywhere, create it
        if(!mySameGroup)
        {
          mySameGroup = [NSMutableArray arrayWithCapacity:8];
          m_entries[otherGroupName] = mySameGroup;
          m_sortedGroups = nil;
        }
        Entry *newEntry = [[Entry alloc]initWithEntry:otherEntry file:self];
        [self addEntry:newEntry];
      }
    }
  }];
               
  [self endMultipeChanges:writeNow];
}

#pragma mark - Import

- (void) importTextFile:(NSString *)textFilePath separator:(unichar)separator delimiter:(unichar)delimiter completion:(SafeFileImportCompletionBlock)block
{
  block = [block copy];
  
  dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
    stringT strErrors;
    int nImported, nSkipped, nErrors, nRenamed, nNoPolicy;
    CReport rpt;
    Command *cmd = nullptr;
    
    int res = m_core.ImportPlaintextFile(L"", [textFilePath wstring], separator, delimiter, false, strErrors, nImported, nSkipped, nErrors, nRenamed, nNoPolicy, rpt, cmd);
    
    NSString *report = [NSString stringWithwstring:rpt.GetString()];
    
    NSString *desc, *reason;
    
    switch (res)
    {
      case PWScore::CANT_OPEN_FILE:
        desc = NSLocalizedString(@"File Read Error", nil);
        reason = NSLocalizedString(@"Could not open file for reading.", nil);
        delete cmd;
        break;
      case PWScore::INVALID_FORMAT:
        desc = NSLocalizedString(@"File Read Error", nil);
        reason = NSLocalizedString(@"Invalid format.", nil);
        delete cmd;
        break;
      case PWScore::FAILURE:
        desc = NSLocalizedString(@"Import Text failed", nil);
        reason = [NSString stringWithWstring:strErrors];
        delete cmd;
        break;
      case PWScore::OK_WITH_ERRORS:
        desc = NSLocalizedString(@"Records Skipped", nil);
        reason = [NSString stringWithWstring:strErrors];
        // deliberate fallthru
      case PWScore::SUCCESS:
        // deliberate fallthru
      default:
      {
        if (cmd != NULL)
        {
          @synchronized(self) {
            m_core.Execute(cmd);
          }
          [self makeModel];
        }

        dispatch_async(dispatch_get_main_queue(), ^{
          NSError *error;
          if(desc.length > 0)
            error = [NSError errorWithDomain:kPasswordSafeErrorDomain code:res userInfo:@{NSLocalizedDescriptionKey : desc, NSLocalizedFailureReasonErrorKey : reason}];
          block(YES, report, nil);
        });
    
        return;
      }
    } // switch
    
    NSError *error = [NSError errorWithDomain:kPasswordSafeErrorDomain code:res userInfo:@{NSLocalizedDescriptionKey : desc, NSLocalizedFailureReasonErrorKey : reason}];
    dispatch_async(dispatch_get_main_queue(), ^{
      block(NO, report, error);
    });
  });
}

#pragma mark - Export

- (void)exportTextFile:(NSString *)textFilePath delimiter:(unichar)delimiter completion:(SafeFileExportCompletionBlock)block
{
  block = [block copy];
  
  dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
    CItemData::FieldBits fields;
    fields.set(CItemData::NAME);
    fields.set(CItemData::GROUP);
    fields.set(CItemData::TITLE);
    fields.set(CItemData::USER);
    fields.set(CItemData::NOTES);
    fields.set(CItemData::PASSWORD);
    fields.set(CItemData::CTIME);
    fields.set(CItemData::PMTIME);
    fields.set(CItemData::RMTIME);
    fields.set(CItemData::URL);
    fields.set(CItemData::PWHIST);
    fields.set(CItemData::POLICY);
    fields.set(CItemData::EMAIL);
    fields.set(CItemData::SYMBOLS);
    fields.set(CItemData::POLICYNAME);
    
    int numExported, res;
    @synchronized(self) {
      res = m_core.WritePlaintextFile(textFilePath.wstring, fields, stringT(), 0, 0, delimiter, numExported);
    }
    NSString *desc, *reason;
    
    switch(res)
    {
      case PWScore::NO_ENTRIES_EXPORTED:
        desc = NSLocalizedString(@"Nothing Exported", @"File export error");
        reason = NSLocalizedString(@"There was nothing to export", @"File export error");
        break;
        
      case PWScore::CANT_OPEN_FILE:
        desc = NSLocalizedString(@"File Write Error", nil);
        reason = NSLocalizedString(@"Could not open file for writing.", nil);
        break;
        
      case PWScore::SUCCESS:
        break;
        
      default:
        desc = NSLocalizedString(@"Unknown Error", nil);
        reason = [NSString stringWithFormat:NSLocalizedString(@"Unknown error %d.", nil), res];
        break;
    }
    
    NSError *error;
    if(desc.length > 0)
      error = [NSError errorWithDomain:kPasswordSafeErrorDomain code:res userInfo:@{NSLocalizedDescriptionKey : desc, NSLocalizedFailureReasonErrorKey : reason}];
    dispatch_async(dispatch_get_main_queue(), ^{
      block(error == nil, error);
    });
  });
}

@end
