//
//  UIViewController+OpenSafe.m
//  pwSafe
//
//  Created by Jorge Peixoto Vasquez on 04/07/12.
//  Copyright (c) 2012 app77. All rights reserved.
//

#import "UIViewController+OpenSafe.h"
#import "SafeCtxStorage.h"
#import "SafeCtx.h"
#import "OpenSafeViewController.h"
#import "DismissableNavigationController.h"
#ifdef PWSAFE_WITH_ICLOUD
#import "FileViewControllerSync.h"
#import "SafeFile_iCloud.h"
#define FileViewController FileViewControllerSync
#endif
#import "SafeFile.h"
#import "FileViewController.h"
#import "MBProgressHUD.h"
#import <objc/runtime.h>

static NSString *kOpenedSafeFilePath = @"OpenedSafe";

@implementation UIViewController (OpenSafe)

#pragma mark - Properties

#define IMPLEMENT_ASSOCIATED_PROP(type, name, setter, copy_) \
static const void * const name##key = &name##key; \
- (type)name { \
  return objc_getAssociatedObject(self, name##key); \
} \
- (void)setter:(type)value { \
  objc_setAssociatedObject(self, name##key, value, copy_ ? OBJC_ASSOCIATION_COPY_NONATOMIC : OBJC_ASSOCIATION_RETAIN_NONATOMIC); \
}

IMPLEMENT_ASSOCIATED_PROP(SafeCtx *, currentSafeCtx, setCurrentSafeCtx, NO)
IMPLEMENT_ASSOCIATED_PROP(SafeFile *, currentOpenedFile, setCurrentOpenedFile, NO)
IMPLEMENT_ASSOCIATED_PROP(OpenSafeCallback, openSafeCallback, setOpenSafeCallback, YES)
IMPLEMENT_ASSOCIATED_PROP(NSString *, initialGroup, setInitialGroup, YES)
IMPLEMENT_ASSOCIATED_PROP(NSString *, initialSearchTerm, setInitialSearchTerm, YES)

#pragma mark - Methods

- (void) askForPasswordAndOpenSafe: (NSString *)filePath groupName:(NSString *)groupName searchTerm:(NSString *)search
{
  if(self.modalViewController)
  {
    A77Log(@"Did not ask for password because there was another modal VC: %@", self.modalViewController.description);
    return;
  }
 
  self.initialGroup = groupName;
  self.initialSearchTerm = search;
  
  OpenSafeViewController *openSafeVC = [[OpenSafeViewController alloc]initWithSafePath:filePath target:self onDone:@selector(onOpenSafeComplete:)];
  DismissableNavigationController *nav = [[DismissableNavigationController alloc]initWithRootViewController:openSafeVC];
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
    nav.modalPresentationStyle = UIModalPresentationFormSheet;
  nav.modalTransitionStyle = UIModalTransitionStyleCoverVertical;
  
  [self presentModalViewController:nav animated:YES];
}

- (void)openSafeAndAskForPasswordIfNeeded:(NSString *)filePath state:(NSMutableDictionary *)state
{
  [self openSafeAndAskForPasswordIfNeeded:filePath groupName:nil searchTerm:nil state:state];
}

- (void)openSafeAndAskForPasswordIfNeeded:(NSString *)filePath groupName:(NSString *)groupName searchTerm:(NSString *)search state:(NSMutableDictionary *)state
{ 
  SafeCtx *ctx = [[SafeCtxStorage sharedStorage]safeCtxForPath:filePath];
  if(ctx.expired || !ctx.password || !ctx.password.length)
  {
    [self askForPasswordAndOpenSafe:filePath groupName:groupName searchTerm:search];
    return;
  }
  
  [self openSafeForCtx:ctx state:state callback:^(SafeFile *safeFile, NSError *error)  {
    if(!safeFile)
    {
      [self askForPasswordAndOpenSafe:filePath groupName:groupName searchTerm:search];
      return;
    }
    FileViewController *fileViewController = [[FileViewController alloc] initWithSafeCtx:ctx safeFile:safeFile state:state];
    fileViewController.initialGroup = groupName;
    fileViewController.initialSearchTerm = search;
    NSAssert(!self.currentSafeCtx, @"No safe should be opened");
    self.currentSafeCtx = ctx;
    self.currentOpenedFile = safeFile;
    [self.navigationController pushViewController:fileViewController animated:YES];
    state[kOpenedSafeFilePath] = filePath;
  }];
}

- (void)openSafeForCtx:(SafeCtx *)safeCtx state:(NSMutableDictionary *)state callback:(OpenSafeCallback)callback 
{
  if(self.openSafeCallback)
  {
#ifdef PWSAFE_VAULT
    NSDictionary *errorDict = @{NSLocalizedDescriptionKey : @"Vault open already in progress"};
#else
    NSDictionary *errorDict = @{NSLocalizedDescriptionKey : @"Safe open already in progress"};
#endif
    
    NSError *error = [NSError errorWithDomain:@"Password Safe" code:-1 userInfo:errorDict];
    DebugLog(@"OpenSafe denied, another open already in progress");
    
    OpenSafeCallback callbackCopy = [callback copy];
    
    dispatch_async(dispatch_get_main_queue(), ^{
      callbackCopy(nil, error);
    });
    
    return;
  }
  self.openSafeCallback = callback;

  dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
    [self createSafeFileForCtx:safeCtx state:state];
  });
}

- (void)createSafeFileForCtx:(SafeCtx *)safeCtx state:(NSMutableDictionary *)state
{
  SafeFile *file;
#ifndef PWSAFE_WITH_ICLOUD
  file = [[SafeFile alloc]initWithFilePath:safeCtx.safePath password:safeCtx.password];
#else
  file = [[SafeFile_iCloud alloc]initWithFilePath:safeCtx.safePath password:safeCtx.password];
#endif
  
#if TARGET_IPHONE_SIMULATOR
  sleep(1);
#endif
  
  NSError *error = nil;
  
  BOOL ok = file && [file loadSafeError:&error];
  
  if(ok)
  {
#ifdef PWSAFE_WITH_ICLOUD
    if([file isKindOfClass:[SafeFile_iCloud class]])
    {
      SafeFile_iCloud *iCloudFile = (SafeFile_iCloud *)file;
      DebugLog(@"Registering presenter");
      [NSFileCoordinator addFilePresenter:iCloudFile];
    }
#endif
  }
  
  
  dispatch_async(dispatch_get_main_queue(), ^{
    NSAssert(self.openSafeCallback, @"Must have a callback");
    
    OpenSafeCallback callback = self.openSafeCallback;
    self.openSafeCallback = nil;
    
    callback(ok ? file : nil, error);
    
    if(ok)
      state[kOpenedSafeFilePath] = safeCtx.safePath;
  });
}

- (void)closeSafeFile:(SafeFile *)file
{
#ifdef PWSAFE_WITH_ICLOUD
  if([file isKindOfClass:[SafeFile_iCloud class]])
  {
    SafeFile_iCloud *iCloudFile = (SafeFile_iCloud *)file;
    DebugLog(@"Unregistering presenter");
    [NSFileCoordinator removeFilePresenter:iCloudFile];
  }  
#endif
}

- (void)loadSavedState_OpenSafe:(NSMutableDictionary *)state
{
  NSString *filePath = state[kOpenedSafeFilePath];
  if(filePath)
  {
    SafeCtx *ctx = [[SafeCtxStorage sharedStorage]safeCtxForPath:filePath];
    if(ctx && !ctx.expired)
    {
      [self openSafeForCtx:ctx state:state callback:^(SafeFile *safeFile, NSError *error) {
        if(!safeFile)
          return;
        FileViewController *fileViewController = [[FileViewController alloc] initWithSafeCtx:ctx safeFile:safeFile state:state];
        DebugLog(@"Restoring safeCtx %@", filePath);
        NSAssert(!self.currentSafeCtx && !self.currentOpenedFile, @"No safe should be opened");
        self.currentSafeCtx = ctx;
        self.currentOpenedFile = safeFile;
        
        [self.navigationController pushViewController:fileViewController animated:NO];
        [fileViewController loadSavedState];
      }];
    }
  }
}

- (void)clearState:(NSMutableDictionary *)state
{
  [state removeObjectForKey:kOpenedSafeFilePath];
}


- (void)hideHud:(id)iHud
{
  MBProgressHUD *hud = (MBProgressHUD *)iHud;
  [hud hide:YES];
}

@end


