//
//  iPasswordSafeAppDelegate.m
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 3/4/11.
//  Copyright 2011 app77. All rights reserved.
//

#include <sys/types.h>
#include <sys/sysctl.h>
#include <mach/mach_time.h>

#if defined(PWSAFE_WITH_DROPBOX) || defined(PWSAFE_WITH_ICLOUD)
# import "RootViewControllerSync.h"
# define RootViewController RootViewControllerSync
# import <DropboxSDK/DropboxSDK.h>
# import "SyncEnabled.h"
#else
# import "RootViewController.h"
#endif

#import "iPasswordSafeAppDelegate.h"
#import "SafeCtxStorage.h"
#import "SafeCtx.h"
#import "DetailViewController.h"
#import "Appirater.h"
#import "MasterPasswordViewController.h"
#import "FileSystemFolders.h"
#import "UIViewController+OpenSafe.h"
#import "InactiveScreenCover.h"
#import "SSKeychain.h"
#import "URLParser.h"
#import "GTMStringEncoding.h"
#import "MBProgressHUD.h"
#import "MobiusSDK.h"
#import "Countly.h"
#import "MobiusAdVc.h"
#import "pwSafeTheme.h"

#import <Parse/Parse.h>

@interface iPasswordSafeAppDelegate() <UIApplicationDelegate, UIActionSheetDelegate, DBSessionDelegate, UISplitViewControllerDelegate, MobiusSDKDelegate>
{
  UIWindow *m_window;
  UINavigationController *m_navigationController;
  NSMutableDictionary *m_state;
  RootViewController *m_rootViewController;
  NSDate *m_whenResignedActive;
  uint64_t m_whenResignedActiveSecs;
  
  NSString *m_receivedFileNewPath;
  NSString *m_receivedFilePath;
  
  void (^m_afterMasterOk)();
}

@end 

@implementation iPasswordSafeAppDelegate

@synthesize window = m_window;

#pragma mark - Master password tasks

- (void)runNowOrAfterMasterPassword:(void (^)())block
{
  if(![MasterPasswordViewController isEnabled])
  {
    block();
    return;
  }
  
  if(!m_afterMasterOk)
  {
    m_afterMasterOk = [block copy];
    return;
  }
  
  void (^newBlock)() = [block copy];
  void (^oldBlock)() = m_afterMasterOk;
  m_afterMasterOk = [^()
  {
    oldBlock();
    newBlock();
  } copy];
}

#pragma mark - Time

static uint64_t getUptimeInSeconds()
{
  static const uint64_t kOneBillion = 1000 * 1000 * 1000;
  static mach_timebase_info_data_t s_timebase_info;
  
  if (s_timebase_info.denom == 0) {
    (void) mach_timebase_info(&s_timebase_info);
  }
  
  // mach_absolute_time() returns billionth of seconds,
  // so divide by one billion to get seconds
  return (uint64_t)((mach_absolute_time() * s_timebase_info.numer) / (kOneBillion * s_timebase_info.denom));
}

- (BOOL)isClockGood
{
  uint64_t nowSecs = getUptimeInSeconds();
  
  if(nowSecs < m_whenResignedActiveSecs)
    return NO;
  
  if(!m_whenResignedActive)
    return YES;
  
  uint64_t deltaSecs = nowSecs - m_whenResignedActiveSecs;
  NSTimeInterval  delta = [[NSDate date]timeIntervalSinceDate:m_whenResignedActive];
  double diff = delta - deltaSecs;
  return diff > -30.0 && delta > 0;
}

#pragma mark -
#pragma mark Application lifecycle

- (NSString *)stateFileName
{
  NSArray *cachesDirs = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
  NSString *cacheDir = cachesDirs[0];

  return [cacheDir stringByAppendingPathComponent:@"state.dat"];
}

-(void)popToRootViewController
{
  // Close the current state
  UIViewController *currentVC = m_navigationController.viewControllers.lastObject;
  if(currentVC != m_rootViewController)
  {
    [currentVC dismissModalViewControllerAnimated:YES];
    [m_navigationController popToRootViewControllerAnimated:YES];
  }
}

-(void)finishHandleReceivedFile:(NSString *)path newPath:(NSString *)newFilePath
{
  [self popToRootViewController];
  [m_state removeAllObjects];
  // Close the safe
  [[[SafeCtxStorage sharedStorage]safeCtxForPath:newFilePath]expireNow];
  
  // Move the file to the Documents dir
  NSFileManager *fileManager = [NSFileManager defaultManager];
  [fileManager removeItemAtPath:newFilePath error:nil];
  [fileManager moveItemAtPath:path toPath:newFilePath error:nil];
  // Remove the subdir
  NSString *subdir = [path stringByDeletingLastPathComponent];
  [fileManager removeItemAtPath:subdir error:nil];
  // Reload
  [m_rootViewController refresh];
  // Open new safe
  [m_rootViewController openSafeAndAskForPasswordIfNeeded:newFilePath state:m_state];
}

-(void)actionSheet:(UIActionSheet *)actionSheet didDismissWithButtonIndex:(NSInteger)buttonIndex
{
  if(buttonIndex == 1)
  {
    NSFileManager *fileManager = [NSFileManager defaultManager];
    
    // Remove the received file and subdir
    [fileManager removeItemAtPath:m_receivedFilePath error:nil];
    NSString *subdir = [m_receivedFilePath stringByDeletingLastPathComponent];
    [fileManager removeItemAtPath:subdir error:nil];    
  }
  else
  {
    NSAssert(!buttonIndex, @"Don't know about other buttons");
    [self finishHandleReceivedFile:m_receivedFilePath newPath:m_receivedFileNewPath];
  }
  
  m_receivedFilePath = nil;
  m_receivedFileNewPath = nil;
}

- (void)handleReceivedFile:(NSString *)path
{
  NSArray *docDirs = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
  NSString *docDir = docDirs[0];
  NSFileManager *fileManager = [NSFileManager defaultManager];
  
  NSString *fileName = [path lastPathComponent];
  NSString *newFilePath = [docDir stringByAppendingPathComponent:fileName];
  NSString *fileSimpleName = [fileName stringByDeletingPathExtension];
  if([newFilePath isEqualToString:path]) // already in Documents dir
    return;
  
  if(![fileManager fileExistsAtPath:newFilePath])
  {
    [self finishHandleReceivedFile:path newPath:newFilePath];
    return;
  }
    
  UIActionSheet *actionSheet = [[UIActionSheet alloc]initWithTitle:[NSString stringWithFormat:@"Replace %@", fileSimpleName] 
                                                           delegate:self 
                                                  cancelButtonTitle:@"Cancel" 
                                             destructiveButtonTitle:@"Replace"
                                                  otherButtonTitles:nil];
  
  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
  {
    [actionSheet showInView:m_window.rootViewController.view];
  }
  else
  {
    UINavigationController *nav = (UINavigationController *)m_window.rootViewController;
    [actionSheet showFromToolbar:nav.toolbar];
  }
  
  m_receivedFilePath = path;
  m_receivedFileNewPath = newFilePath;

  return;
}

static NSString * const kPWSafeScheme = @"com-app77-pwsafe";
static NSString * const kOpenURLHost = @"open";
static NSString * const kSearchParam = @"search";

static BOOL openURL(NSString *urlString)
{
  if(!urlString)
    return NO;
  
  NSURL *url = [NSURL URLWithString:urlString];
  UIApplication *app = [UIApplication sharedApplication];
  if([app canOpenURL:url])
  {
    [app openURL:url];
    return YES;
  }
  else
  {
    NSLog(@"Unable to open url %@", urlString);
    return NO;
  }
}

- (BOOL)application:(UIApplication *)application openURL:(NSURL *)url sourceApplication:(NSString *)sourceApplication annotation:(id)annotation
{
  A77Log(@"Handling URL of type: %@", url.scheme);
  
  if(!url)
    return NO;
  
  if(url.isFileURL)
  {
    [self runNowOrAfterMasterPassword:^{
      [self handleReceivedFile:url.path];
    }];
    
    return YES;    
  }

  if([[MobiusSDK sharedMobius]handleURL:url])
     return YES;
  
  if([url.scheme isEqualToString:kPWSafeScheme])
  {
    if([url.host isEqualToString:kOpenURLHost])
    {
      NSArray *pathComponents = url.pathComponents;
      
      if(pathComponents.count < 2)
        return NO;
      
      NSString *safeName = pathComponents[1];
      NSString *groupName = pathComponents.count > 2 ? pathComponents[2] : nil;

      URLParser *parser = [[URLParser alloc]initWithURLString:url.absoluteString];
      NSString *searchStr = parser[kSearchParam];
      
      if(safeName)
        [m_rootViewController openSafeWithName:safeName groupName:groupName searching:searchStr animated:NO];
      return YES;
    }
  }
  
#ifdef PWSAFE_WITH_DROPBOX
  if ([[DBSession sharedSession] handleOpenURL:url]) 
  {
    if ([[DBSession sharedSession] isLinked]) 
      [m_rootViewController dropboxLinkComplete]; 
    return YES;
  }  
#endif
  
  return NO;
}

- (BOOL)application:(UIApplication *)application didFinishLaunchingWithOptions:(NSDictionary *)launchOptions 
{   
  [Crashlytics startWithAPIKey:@"ea01597f64ab75175cece72db2c3c5f73f6d4866" afterDelay:3.0];

  [MobiusSDK appDidFinishLaunchWithLicenseKey:@"iopT54D5XhL4b5H031pEDpfY66ZGEb0UVGYodczWHXDy73XvsjVrBsSZ_Oc="
                                    urlScheme:@"com-app77-pwsafe-me"
                                     delegate:self];

  [Parse setApplicationId:@"XNBCXefjnnZH7QO9kr61tCKds7ixoB5AO2eo8pzt"
                clientKey:@"N6KEmDtD4VN7b9tqk6ggP1ZPOYYS2AdBzQfLNjnF"];
  
  NSString* consumerKey = @"7of4lwhmz4qpso3";
  NSString* consumerSecret = @"cwh00u9w9bakwdd";
  
  // Look below where the DBSession is created to understand how to use DBSession in your app
  DBSession* session = [[DBSession alloc]  initWithAppKey:consumerKey appSecret:consumerSecret root:kDBRootDropbox];
                        
  session.delegate = self; // DBSessionDelegate methods allow you to handle re-authenticating
  [DBSession setSharedSession:session];
  
  NSString *stateFilePath = self.stateFileName;
  
  
  // Upgrade de criptografia
  NSFileManager *fileManager = [NSFileManager defaultManager];
  [fileManager setAttributes:@{NSFileProtectionKey: NSFileProtectionComplete} ofItemAtPath:stateFilePath error:nil];
  
  // Change Dropbox files path
  BOOL isDir;
  if([fileManager fileExistsAtPath:oldDropboxDir() isDirectory:&isDir] && isDir)
  {
    BOOL ok = [fileManager moveItemAtPath:oldDropboxDir() toPath:dropboxDir() error:nil];
    (void)ok;
    NSAssert(ok, @"Must succeed");
  }
  
  self.window = [[UIWindow alloc]initWithFrame:[[UIScreen mainScreen] bounds]];
  self.window.screen = [UIScreen mainScreen];
  self.window.backgroundColor = iOS6() ? kThemeDarkBgColor : [UIColor blackColor];

  // Setup theme
  pwSafeThemeSetup(self.window);
  
  // Load the state
  m_state = [NSKeyedUnarchiver unarchiveObjectWithFile:stateFilePath];
  if(!m_state)
    m_state = [[NSMutableDictionary alloc]init];

  m_rootViewController = [[RootViewController alloc] initWithState:m_state];
  m_navigationController = [[UINavigationController alloc]initWithRootViewController:m_rootViewController];

  if(UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad) 
  {
    UISplitViewController *splitView = [[UISplitViewController alloc]init];
    splitView.viewControllers = @[m_navigationController, [DetailViewController sharedDetailViewControllerWithState:m_state]];
    splitView.delegate = self;
    self.window.rootViewController = splitView;
  }
  else
  {
    self.window.rootViewController = m_navigationController;
  }
  
  [InactiveScreenCover sharedCover].disableHideScreenOnInactive = NO;
  [InactiveScreenCover sharedCover].imageName = @"ElefanteTransparente";
  [InactiveScreenCover sharedCover].imageMargin = UIEdgeInsetsMake(15, 15, 15, 15);
  
  [self.window makeKeyAndVisible];

  NSURL *url = (NSURL *)[launchOptions valueForKey:UIApplicationLaunchOptionsURLKey];

  // Do not handle that here, handle in application:openURL:sourceApplication:annotation:
  if(!url || ![url isFileURL])
  {
    [m_rootViewController loadSavedState];
    [Appirater appLaunched:YES];
  }
  else
  {
    [Appirater appLaunched:NO];
  }
  
  dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_BACKGROUND, 0), ^{
    // Register installs & version upgrades
    NSString *account = @"pwSafeService";
    NSString *service = @"LastVersionLaunched";
    NSString *curVersion = [[NSBundle mainBundle]objectForInfoDictionaryKey:@"CFBundleShortVersionString"];
    
    NSString *lastVersion = [SSKeychain passwordForService:service account:account];
    if(!lastVersion || ![lastVersion isEqualToString:curVersion])
    {
      NSError *error = nil;
      BOOL ok = [SSKeychain setPassword:curVersion forService:service account:account error:&error];
      if(!ok)
      {
        A77Log(@"Error saving latest version to keychain: %@", error.localizedDescription);
      }
      else
      {
        if(lastVersion)
          A77Log(@"Upgrade from %@ to %@", lastVersion, curVersion);
        else
          A77Log(@"New install of version %@", curVersion);
      }
    }  
  });
  
  [[Countly sharedInstance]start:@"0e58478aeb0f06795f8ded237f1c50b5b2ba7f49" withHost:@"https://cloud.count.ly"];
  
  [[NSUbiquitousKeyValueStore defaultStore]synchronize];
  
  return YES;
}

- (BOOL)splitViewController:(UISplitViewController *)svc shouldHideViewController:(UIViewController *)vc inOrientation:(UIInterfaceOrientation)orientation
{
  return NO;
}

- (void)applicationWillResignActive:(UIApplication *)application 
{
  DebugLog(@"%s will mark current safe as expired", __FUNCTION__);
  
  [[SafeCtxStorage sharedStorage].openedSafeCtx sleep];
  
  m_whenResignedActive = [NSDate date];
  m_whenResignedActiveSecs = getUptimeInSeconds();
  
  [m_rootViewController appWillResignActive];
}

- (void) saveState 
{
  NSString *stateFilePath = self.stateFileName;
  
  [NSKeyedArchiver archiveRootObject:m_state toFile:stateFilePath];
  [[NSFileManager defaultManager]setAttributes:@{NSFileProtectionKey: NSFileProtectionComplete} ofItemAtPath:stateFilePath error:nil];

  [[SafeCtxStorage sharedStorage].openedSafeCtx sleep];
  [[SafeCtxStorage sharedStorage] save];
}

- (void)applicationDidEnterBackground:(UIApplication *)application 
{
  [self saveState];

  
  /*
     Use this method to release shared resources, save user data, invalidate timers, and store enough application state information to restore your application to its current state in case it is terminated later. 
     If your application supports background execution, called instead of applicationWillTerminate: when the user quits.
     */
}

- (void)applicationWillEnterForeground:(UIApplication *)application 
{
  /*
   Called as part of  transition from the background to the inactive state: here you can undo many of the changes made on entering the background.
   */
  [[NSUserDefaults standardUserDefaults]synchronize];
  [self runNowOrAfterMasterPassword:^{
    if(![m_rootViewController showWelcomeIfNeeded])
      [m_rootViewController showMacAddIfNeeded];
  }];
  
  [Appirater appEnteredForeground:YES];
  
  [[NSUbiquitousKeyValueStore defaultStore]synchronize];
}

- (void)applicationDidBecomeActive:(UIApplication *)application 
{
  BOOL clockGood = [self isClockGood];
  if(!clockGood)
  {
    for(SafeCtx *ctx in [[SafeCtxStorage sharedStorage]allSafeCtxs])
    {
      [ctx expireNow];
      ctx.password = nil;
    }
  }
  SafeCtx *safe = [SafeCtxStorage sharedStorage].openedSafeCtx;
  BOOL currentSafeExpired = !safe || safe.expired;
  DebugLog(@"%s: expired = %s", __FUNCTION__, currentSafeExpired ? "TRUE" : "FALSE");
  if (!clockGood || currentSafeExpired) 
  {
    safe.opened = NO;
    [self popToRootViewController];
    [m_state removeAllObjects];
  }
  
  BOOL checkMaster;
  if(!clockGood || !m_whenResignedActive)
  {
    checkMaster = YES;
  }
  else
  {
    NSTimeInterval  delta = [[NSDate date]timeIntervalSinceDate:m_whenResignedActive];    
    m_whenResignedActive = nil;
    
    NSNumber *masterTimeoutMinutes = [[NSUserDefaults standardUserDefaults]objectForKey:kMasterPasswordExpireMinutesKey];
    NSTimeInterval max = 60 * (masterTimeoutMinutes ? [masterTimeoutMinutes integerValue] : 5);
    if(max >= 0)
      checkMaster = delta > max;
    else
      checkMaster = NO;
  }

  if(checkMaster)
  {
    dispatch_async(dispatch_get_main_queue(), ^() {
      [MasterPasswordViewController verifyInViewController:nil allowCancel:NO callback:^(BOOL ok) 
      {
        NSAssert(ok, @"Must suceed");
        // Reload encrypted passwords
        if(clockGood)
          [[SafeCtxStorage sharedStorage] decryptPasswordsIfNeeded];        
        
        [m_rootViewController refresh];
        
        dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, .5 * NSEC_PER_SEC);
        dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
          if(m_afterMasterOk)
          {
            m_afterMasterOk();
            m_afterMasterOk = nil;
          }
        });
      }];
    });
  }
  
  [m_rootViewController refresh];
  [m_rootViewController appDidBecomeActive];
}


- (void)applicationWillTerminate:(UIApplication *)application 
{
  [[SafeCtxStorage sharedStorage].openedSafeCtx sleep];
  [self saveState];
}

#pragma mark - DropBox

- (void)sessionDidReceiveAuthorizationFailure:(DBSession *)session userId:(NSString *)userId
{
  // Nothing to do
}

#pragma mark - Mobius

- (void)mobiusAccount:(NSString *)accountID authenticatedWithSecret:(NSString *)secret
{
  [m_rootViewController openSafeWithUUID:accountID password:secret animated:NO];
  A77Log(@"Authenticated by Mobius Engine");
}

- (void)mobiusAccount:(NSString *)accountID failedAuthenticationWithError:(NSError *)error
{
  SafeCtx *safeCtx = [[SafeCtxStorage sharedStorage]safeCtxForUUID:accountID];
  NSString *safeName = safeCtx ? safeCtx.safeName : accountID;
  UIAlertView *alertView = [[UIAlertView alloc]initWithTitle:@"Authentication failed"
                                                    message:[NSString stringWithFormat:@"Mobius Engine failed to authenticate safe %@. Error: %@.", safeName, error.userInfo[MobiusSDKErrorDescription]]
                                                   delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil];
  [alertView show];
  A77Log(@"Mobius Engine failed to authenticate: %@", error);
}

- (void)mobiusAccount:(NSString *)accountID registeredWithResult:(BOOL)ok error:(NSError *)error
{
  UIAlertView *alertView;
  if(ok)
  {
    A77Log(@"Mobius Engine registered");
    alertView = [[UIAlertView alloc]initWithTitle:@"Registered"
                                          message:@"Successfully registered with Mobius Engine. From now on, you will be able to open your safe with Mobius instead of typing your password."
                                         delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil];
  }
  else
  {
    SafeCtx *safeCtx = [[SafeCtxStorage sharedStorage]safeCtxForUUID:accountID];
    NSString *safeName = safeCtx ? safeCtx.safeName : accountID;
    alertView = [[UIAlertView alloc]initWithTitle:@"Authentication failed"
                                          message:[NSString stringWithFormat:@"Mobius Engine failed to register safe %@. Error: %@.", safeName, error.userInfo[MobiusSDKErrorDescription]]
                                         delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil];

    A77Log(@"Mobius Engine failed to register: %@", error);
  }
  [alertView show];
}

- (void)mobiusEngineInstalledByReferral:(BOOL)referral onDate:(NSDate *)date onDeviceWithId:(NSString *)mobiusDeviceID
{
#if (TARGET_IPHONE_SIMULATOR) || defined(DEBUG_MODE)
  return;
#else
  NSDate *myOwnReferral = [MobiusAdVc firstDateMobiusAdViewed];
 
  if(!referral && !myOwnReferral)
    return;

  NSDate *firstReferral = date ? date : [NSDate date];
  if(myOwnReferral && [myOwnReferral compare:firstReferral] == NSOrderedAscending)
    firstReferral = myOwnReferral;
  
  PFObject *ref = [PFObject objectWithClassName:@"MobiusReferral"];
  ref[@"firstReferral"] = firstReferral;
  ref[@"deviceId"] = mobiusDeviceID;
  [ref saveEventually];
#endif
}

@end

