//
//  multibyte.c
//  iPasswordSafe
//
//  Created by Jorge Peixoto Vasquez on 4/21/11.
//  Copyright 2011 app77. All rights reserved.
//

#import <Foundation/Foundation.h>
#include <wchar.h>
#include <vector>

#if TARGET_RT_BIG_ENDIAN
static const NSStringEncoding kEncoding = NSUTF32BigEndianStringEncoding;
#else
static const NSStringEncoding kEncoding = NSUTF32LittleEndianStringEncoding;
#endif

extern "C" size_t wcsrtombs(char * __restrict out, const wchar_t ** __restrict p_in, size_t maxLen,
                  mbstate_t * __restrict st)
{
  const wchar_t  *in = *p_in;
  size_t inLen = wcslen(in);
  NSString *str = [[NSString alloc]initWithBytesNoCopy:const_cast<wchar_t *>(in) length:inLen*sizeof(wchar_t) encoding:kEncoding freeWhenDone:NO];
  if(!str)
    return -1;
  NSData *acp = [str dataUsingEncoding:NSWindowsCP1252StringEncoding];
  NSUInteger len = [acp length];
  if(!out)
    return len;
  size_t bytesToCopy = len > maxLen ? maxLen : len;
  memcpy(out, [acp bytes], bytesToCopy);
  if(maxLen > bytesToCopy)
    out[bytesToCopy] = 0;
  return bytesToCopy;  
}

extern "C" size_t wcstombs(char * __restrict out, const wchar_t * __restrict in, size_t maxLen)
{
  size_t inLen = wcslen(in);
  NSString *str = [[NSString alloc]initWithBytesNoCopy:const_cast<wchar_t *>(in) length:inLen*sizeof(wchar_t) encoding:kEncoding freeWhenDone:NO];
  if(!str)
    return -1;
  NSData *utf8 = [str dataUsingEncoding:NSUTF8StringEncoding];
  NSUInteger len = [utf8 length];
  if(!out)
    return len;
  size_t bytesToCopy = len > maxLen ? maxLen : len;
  memcpy(out, [utf8 bytes], bytesToCopy);
  if(maxLen > bytesToCopy)
    out[bytesToCopy] = 0;
  return bytesToCopy;
}

extern "C" size_t mbstowcs(wchar_t * __restrict out, const char * __restrict in, size_t maxLen)
{
  NSString *str = @(in);
  if(!str)
    return -1;
  NSData *wStr = [str dataUsingEncoding:kEncoding];
  NSUInteger len = [wStr length] / sizeof(wchar_t);
  if(!out)
    return len;
  size_t nToCopy = len > maxLen ? maxLen : len;
  memcpy(out, [wStr bytes], nToCopy * sizeof(wchar_t));
  if(maxLen > nToCopy)
    out[nToCopy] = 0;
  
  return nToCopy;
}

#include <xlocale.h>

#if TARGET_RT_64_BIT
#define SW_PTR_LEN 25
#define SW_SHIFT 60
static NSString * const kPtrFormat = @"_ãíóüé_0x%016qx";
#else
#define SW_PTR_LEN 17
#define SW_SHIFT 28
static NSString * const kPtrFormat = @"_ãíóüé_0x%08x";
#endif
static NSString * const kPtrMarker = @"_ãíóüé_0x";
#define SW_PTR_MARKER_SKIP_LEN 9

static NSString *ns_swprintf(const wchar_t * __restrict fmt, __darwin_va_list va) 
{
  size_t fmtLen = wcslen(fmt);
  NSMutableString *nsFormat = [[NSMutableString alloc]initWithBytes:fmt length:fmtLen*sizeof(wchar_t) encoding:kEncoding];
  
  // Convert %S -> %p
  NSUInteger nsFormatLen = nsFormat.length;
  NSUInteger idx = 0;
  while((idx = [nsFormat rangeOfString:@"%S" options:0 range:NSMakeRange(idx, nsFormatLen - idx)].location) != NSNotFound)
  {
    [nsFormat replaceCharactersInRange:NSMakeRange(idx, 2) withString:kPtrFormat];
    idx += kPtrFormat.length;
    nsFormatLen += kPtrFormat.length - 2;
  }
  
  // Print
  NSMutableString *nsOut = [[NSMutableString alloc]initWithFormat:nsFormat arguments:va];
  
  // Change the address printed by %p into strings
  
  idx = 0;
  NSUInteger nsOutLen = nsOut.length;
  while((idx = [nsOut rangeOfString:kPtrMarker options:0 range:NSMakeRange(idx, nsOutLen - idx)].location) != NSNotFound)
  {
    // Parse address
    NSCAssert(idx + SW_PTR_LEN <= nsOut.length, @"Must not overflow");
    const char *singleByteStr = nullptr;
    NSUInteger max = idx + SW_PTR_LEN;
    NSUInteger shift = SW_SHIFT;
    for(idx += SW_PTR_MARKER_SKIP_LEN; idx < max; idx++, shift -= 4)
    {
      size_t val = [nsOut characterAtIndex:idx];
      NSCAssert((val >= '0' && val <= '9') || (val >= 'a' || val <= 'f'), @"Must be hex");
      size_t shiftedNibble;
      if(val >= '0' && val <= '9')
        shiftedNibble = (val - '0') << shift;
      else
        shiftedNibble = (val + 10 - 'a') << shift;
      
      singleByteStr += shiftedNibble;
    }
    NSCAssert(shift == -4, @"Shift must have finished too");
    idx -= SW_PTR_LEN;
    
    // Substitute string
    const wchar_t *str = reinterpret_cast<const wchar_t *>(singleByteStr);
    size_t strLen = wcslen(str);
    NSString *nsStr = [[NSString alloc]initWithBytes:const_cast<wchar_t *>(str) length:strLen*sizeof(wchar_t) encoding:kEncoding];
    NSUInteger nsStrLen = nsStr.length;
    if(nsStrLen <= SW_PTR_LEN)
    {
      if(nsStrLen > 0)
        [nsOut replaceCharactersInRange:NSMakeRange(idx, nsStrLen) withString:nsStr];
      NSUInteger delta = SW_PTR_LEN - nsStrLen;
      if(delta > 0)
        [nsOut deleteCharactersInRange:NSMakeRange(idx + nsStrLen, delta)];
    }
    else
    {
      [nsOut replaceCharactersInRange:NSMakeRange(idx, SW_PTR_LEN) withString:[nsStr substringToIndex:SW_PTR_LEN]];
      NSUInteger targetIdx = idx + SW_PTR_LEN;
      NSString *targetStr = [nsStr substringFromIndex:SW_PTR_LEN];
      if(targetIdx >= nsOut.length)
        [nsOut appendString:targetStr];
      else
        [nsOut insertString:targetStr atIndex:targetIdx];
    }
    idx += nsStrLen;
    nsOutLen += nsStrLen - SW_PTR_LEN;
  }
  return nsOut;
}

extern "C" int vswprintf(wchar_t * __restrict out, size_t len, const wchar_t * __restrict fmt, __darwin_va_list va)
{
  NSString *nsOut = ns_swprintf(fmt, va);

  // Just convert into wchar_t
  NSData *outBytes = [nsOut dataUsingEncoding:kEncoding allowLossyConversion:YES];
  NSUInteger outBytesLen = outBytes.length;
  NSCAssert(outBytesLen % sizeof(wchar_t) == 0, @"Must be a multiple");
  NSUInteger outWchars = outBytesLen / sizeof(wchar_t);
  if(outWchars >= len)
    return -1;
  memcpy(out, outBytes.bytes, outBytesLen);
  out[outWchars] = 0;
  return static_cast<int>(outWchars);
}

extern "C" int vaswprintf(wchar_t * __restrict * __restrict p_out, const wchar_t * __restrict fmt, __darwin_va_list va)
{
  auto &out = *p_out; // Make things easier
  
  NSString *nsOut = ns_swprintf(fmt, va);
  
  // Just convert into wchar_t
  NSData *outBytes = [nsOut dataUsingEncoding:kEncoding allowLossyConversion:YES];
  NSUInteger outBytesLen = outBytes.length;
  NSUInteger outWchars = outBytesLen / sizeof(wchar_t);
  NSCAssert(outBytesLen % sizeof(wchar_t) == 0, @"Must be a multiple");
  
  out = (wchar_t *)malloc(outBytesLen + sizeof(wchar_t));
  if(!out)
    return -1;
  memcpy(out, outBytes.bytes, outBytesLen);
  out[outWchars] = 0;
  return static_cast<int>(outWchars);
}

extern "C" int ios_iswalpha(wint_t _wc)
{
  auto ret = [[NSCharacterSet letterCharacterSet]longCharacterIsMember:_wc];
  return ret;
}

extern "C" int ios_iswalnum(wint_t _wc)
{
  auto ret = [[NSCharacterSet alphanumericCharacterSet]longCharacterIsMember:_wc];
  return ret;
}

extern "C" int ios_iswdigit(wint_t _wc)
{
  auto ret = [[NSCharacterSet decimalDigitCharacterSet]longCharacterIsMember:_wc];
  return ret;
}

extern "C" wint_t ios_towupper(wint_t _wc)
{
  if(![[NSCharacterSet lowercaseLetterCharacterSet]longCharacterIsMember:_wc])
    return _wc;

  wchar_t wc = _wc;
  NSString *upperStr = [[[NSString alloc]initWithBytes:&wc length:sizeof(wchar_t) encoding:kEncoding]uppercaseString];
  wint_t ret = *static_cast<const wchar_t *>([upperStr dataUsingEncoding:kEncoding].bytes);
  return ret;
}

extern "C" wint_t ios_towlower(wint_t _wc)
{
  if(![[NSCharacterSet uppercaseLetterCharacterSet]longCharacterIsMember:_wc])
    return _wc;
  
  wchar_t wc = _wc;
  NSString *lowerStr = [[[NSString alloc]initWithBytes:&wc length:sizeof(wchar_t) encoding:kEncoding]lowercaseString];
  wint_t ret = *static_cast<const wchar_t *>([lowerStr dataUsingEncoding:kEncoding].bytes);
  return ret;
}

extern "C" int ios_iswlower(wint_t _wc)
{
  auto ret = [[NSCharacterSet lowercaseLetterCharacterSet]longCharacterIsMember:_wc];
  return ret;
}

extern "C" int ios_iswupper(wint_t _wc)
{
  auto ret = [[NSCharacterSet uppercaseLetterCharacterSet]longCharacterIsMember:_wc];
  return ret;
}


